/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef SYNAPS_ARITHM_ZP_H
#define SYNAPS_ARITHM_ZP_H
//--------------------------------------------------------------------
#include <iostream>
#include <cstdlib>
#include <synaps/init.h>
#include <synaps/arithm/let.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
/** \file                                                           */
/**
   Class used to represent modular numbers. The internal storage is
   of type @T@. The computations are performed modulo @p@.
   The number @p@ should be a prime number, otherwise
   inversion and exponentiation may not be valid.
 **/
template <unsigned int p, class T=int>
class Z
{
private:
  T rep_;

public:
  Z(): rep_(0) {}
  Z(int n): rep_(n%p){ if(rep_<0) rep_+=p; }
  Z(const char* s);
  Z(const Z<p,T> & z): rep_(z.rep_) {}

  T rep() const {return rep_;}
  Z<p,T>&  operator= (const Z<p,T>  z)
    {rep_=z.rep_; return *this;}
  Z<p,T>&  operator+=(const Z<p,T>  z)
    {rep_ += z.rep_; if(rep_>=(int)p) rep_-=p; return *this;}
  Z<p,T>&  operator-=(const Z<p,T>  z)
    {rep_ -= z.rep_; if(rep_<0) rep_+=p; return *this;}
  Z<p,T>&  operator*=(const Z<p,T>  z)
    {(rep_ *= z.rep_)%= p; if(rep_<0) rep_+=p; return *this;}
  Z<p,T>&  operator/=(const Z<p,T> &  z);
  bool   operator==(const Z<p,T> & z) const {return (rep_==z.rep_);}
  bool   operator!=(const Z<p,T> & z) const {return !(*this==z);}

  friend inline
  Z<p,T> operator -(const Z<p,T> & a) {return Z<p,T>(-a.rep_);}
  friend inline
  Z<p,T> operator+(const Z<p,T>  & a, const  Z<p,T> & b){return (Z<p,T>(a.rep_+b.rep_));}
  friend inline
  Z<p,T> operator-(const Z<p,T> & a, const  Z<p,T> & b){return (Z<p,T>(a.rep_-b.rep_));}
  friend inline
  Z<p,T> operator*(const Z<p,T> & a, const  Z<p,T> & b){return (Z<p,T>(a.rep_*b.rep_));}
  friend inline
  Z<p,T> operator/(const Z<p,T> & a, const  Z<p,T> & b){return (Z<p,T>(a)/=b);}

  friend
  std::ostream &  operator<<(std::ostream&  os, const Z<p,T> & z) {return (os << z.rep_);}
  friend
  std::istream & operator>>(std::istream&  is, Z<p,T> & z) {return (is >> z.rep_);}

};


//----------------------------------------------------------------------
/// Conversion of a string to the cooresponding modular number.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Z<p,T>::Z(const char* s) :rep_(0)
{
  int i=0; while(s[i]!='\0') i++;
  unsigned b=1;
  while(i>0){
    rep_+=(s[i-1]-'0')*b; rep_%=p;
    b*=10; b%=p;
    i--;
  }
}
//----------------------------------------------------------------------
/// Division operator, based on Euclid's algorithm.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Z<p,T> & Z<p,T>::operator/=(const Z<p,T> & z)
{
  T r=z.rep_, b=z.rep_, a=p,u1=rep_,u2=rep_, q, s=1;
  if(z==0){ std::cerr<< "xxxx division by zero"<<std::endl;}
  rep_=0;
  if(a<0) {s=-1; a*=s;}
  while(r!=0)
    {
      q = a/b;
      r = a-q*b; a=b; b=r;
      u2= rep_-q*u1; rep_=u1; u1=u2;
    }
  rep_ =(rep_ %p + p)%p;  rep_ *=s;
  return *this;
}
//----------------------------------------------------------------------
/// Fast binary exponentiation.
//----------------------------------------------------------------------
template <unsigned int p, class T>
Z<p,T> operator^(const Z<p,T> a, const  int n)
{
    int m;
    Z<p,T> r=1, e;
    if(n <0){e=1/a;m=-n;}else {e=a;m=n;}
    while(m !=0) {
      if(m%2){
	m -=1; m/=2; r*=e; e*=e;
      }else{
	m/=2; e*=e;
      }
    }
    return r;
}
//--------------------------------------------------------------------
template <unsigned int p, class T>
bool with_plus_sign(const Z<p,T>& z)
{ 
  return (z.rep()>0);
}
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
#endif // SYNAPS_ARITHM_ZP_H

