/********************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 *   $Id: diadic.h,v 1.1 2005/07/11 10:39:49 mourrain Exp $
 *   
 ********************************************************************/
#ifndef SYNAPS_ARITHM_DIADIC_H
#define SYNAPS_ARITHM_DIADIC_H
//--------------------------------------------------------------------
#include <cassert>
#include <synaps/init.h>
#include <synaps/base/type.h>
#include <synaps/base/comparison.h>
#include <synaps/arithm/let.h>
#include <synaps/arithm/pow.h>
#include <synaps/arithm/gmp.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
template <class NT_>
NT_ shift_2(const NT_& z, int d)
{
  if(d>0)
    return z*pow(NT_(2),d);
  else
    return z/pow(NT_(2),-d);
}

//--------------------------------------------------------------------
template< typename  NT_ > class diadic;
namespace let 
{
   template<typename NT_> void assign(diadic<NT_> & x, int i);
}
//--------------------------------------------------------------------
/** This class provides arithmetic operations on numbers 
    which are an integer of type @NT_@, 
    times a power of 2, which is stored by as its exponent.
**/
template < typename  NT_ >
class diadic
{
  public:
  typedef NT_      NT;
  
  diadic() : num_(0), exp_(0) { }

  diadic(const NT& n,int d=0) : num_(n), exp_(d) {normalize();}

  template < typename  RT>
  diadic(const RT& n,int d=0) 
  { 
    let::assign(*this,n); 
    exp_+=d;
  }

  diadic<NT>& operator+= (const diadic<NT>& r);
  diadic<NT>& operator-= (const diadic<NT>& r);
  diadic<NT>& operator*= (const diadic<NT>& r);
  diadic<NT>& operator/= (const diadic<NT>& r);
  diadic<NT>& operator/= (int r);
  diadic<NT>& operator+= (const NT& r);
  diadic<NT>& operator-= (const NT& r);
  diadic<NT>& operator*= (const NT& r);
  diadic<NT>& operator/= (const NT& r);
  
  diadic<NT>& normalize();
  
  const NT&   numerator()   const { return num_; }
        NT    denominator() const { return shift_2(1,exp_); }
        int   valuation ()  const { return exp_; }
  
 public:
  NT   num_;
  int  exp_;
};
//------------------------------------------------------------------------
template < typename NT >  inline
diadic<NT>& diadic<NT>::normalize()
{
 if( num_==0)
    exp_=0;
  else 
    while( (num_%2) ==0 ) 
      {
	exp_++;
	num_/=2;
      }
  return *this;
}
//--------------------------------------------------------------------
// Arithmetic operators
//--------------------------------------------------------------------
/// Inplace addition  of diadic numbers.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator+= (const diadic<NT>& r)
{
  if(r.exp_<exp_)
    {
      int d=exp_-r.exp_;
      num_ =  shift_2(num_ ,d);
      num_ += r.num_;
      exp_ = r.exp_;
    }
  else if(r.exp_>exp_)
    {
      num_  += shift_2(r.num_,r.exp_-exp_);
    }
  else
    {
      num_ += r.num_;
      normalize();
    }
  return *this;
}
//------------------------------------------------------------------------
/// Inplace substraction  of diadic numbers.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator-= (const diadic<NT>& r)
{
  if(r.exp_<exp_)
    {
      num_  = shift_2(num_,exp_-r.exp_);
      num_ -= r.num_;
      exp_  = r.exp_;
    }
  else if(r.exp_>exp_)
    {
      num_ -=  shift_2(r.num_ ,r.exp_-exp_);
    }
  else
    {
      num_ -= r.num_;
      normalize();
    }
  return *this;
}
//------------------------------------------------------------------------
/// Inplace multiplication of diadic numbers.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator*= (const diadic<NT>& r)
{
  num_ *= r.num_;
  exp_ += r.exp_;
  return *this;
}
//------------------------------------------------------------------------
/** Inplace division of diadic numbers. The numerator is divided inplace
    by the other denominators and the exponents are substracted.
    This corresponds to a euclidean division.
**/
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator/= (const diadic<NT>& r)
{
  assert( r.num_ != 0 );
  exp_ -= r.exp_;
  num_ /= r.num_;
  normalize();
  return *this;
}
//------------------------------------------------------------------------
/** Inplace division of a diadic number by an machine integer @n@. 
    The number @n@ is supposed to be a power of two.
**/
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator/= (const NT& r)
{
    assert( r != 0 );
    NT d=r;
    while( (d%2) ==0 ) 
      {
	exp_--;
	d/=2;
      }
    num_/=d;
    return *this;
}
//------------------------------------------------------------------------
/** Inplace division of a diadic number by an machine integer @n@. 
    The number @n@ is supposed to be a power of two.
**/
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator/= (int r)
{
    assert( r != 0 );
    int d=r;
    while( d%2==0 ) { exp_--; d/=2; }
    num_/=d;
    return *this;
}
//------------------------------------------------------------------------
/// Inplace addition of a diadic number with an integer.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator+= (const NT& r)
{
  if(exp_<0)
    num_ += shift_2(r,exp_);
  else if(exp_>0)
    {
      num_ = shift_2(num_,exp_);
      num_+=r;
    }
  else
    num_+=r;
  normalize();
  return *this;
}
//------------------------------------------------------------------------
/// Inplace substraction of a diadic number with an integer.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator-= (const NT& r)
{
  if(exp_<0)
    num_ -= shift_2(r,exp_);
  else if(exp_>0)
    {
      num_ = shift_2(num_,exp_);
      num_-= r;
    }
  else
    num_-= r;
  normalize();
  return *this;
}
//------------------------------------------------------------------------
/// Inplace multiplication of a diadic number with an integer.
template < typename NT >  inline
diadic<NT>&
diadic<NT>::operator*= (const NT& r)
{
  num_ *= r;
  normalize();
  return *this;
}
//------------------------------------------------------------------------
// Help function. Do not use it :-)
template < typename NT >  inline
int
quotient_cmp(const diadic<NT>& a, const diadic<NT>& b)
{
  int asign = sign(a.num_);
  int bsign = sign(b.num_);
  
  if ((asign == 0) && (bsign == 0)) return EQUAL;
  if (asign < bsign) return SMALLER;
  if (asign > bsign) return LARGER;

  // now a and b have the same sign
  NT ta=a.num_, tb=b.num_;
  int d=a.exp_-b.exp_;

  if(d>0)
    ta = shift_2(ta,d);
  else
    tb = shift_2(tb,-d);

  if(ta<tb)
    return SMALLER;
  else if(tb<ta)
    return LARGER;
  else 
    return EQUAL;
}
//--------------------------------------------------------------------
template < typename NT >
inline
int
compare(const diadic<NT>& x, const diadic<NT>& y)
{ return quotient_cmp(x, y); }
//--------------------------------------------------------------------
template < typename NT >
std::ostream&
operator<<(std::ostream& s, const diadic<NT>& r)
{
  s << r.numerator();
  if(r.valuation()) s<< "*2^(" << r.valuation()<<")";
  return s;
}
//--------------------------------------------------------------------
// Arithmetic with diadics
//--------------------------------------------------------------------
/// Addition of diadic numbers.
template < typename NT >  inline
diadic<NT>
operator+(const diadic<NT>& x, const diadic<NT>& y)
{
  diadic<NT> z = x;
  return z += y;
}
//--------------------------------------------------------------------
/// Substraction of diadic numbers.
template < typename NT >  inline
diadic<NT>
operator-(const diadic<NT>& x, const diadic<NT>& y)
{ return (diadic<NT>(x) -= y); }

//--------------------------------------------------------------------
/// Multiplication of diadic numbers.
template < typename NT >  inline
diadic<NT>
operator*(const diadic<NT>& x, const diadic<NT>& y)
{
  diadic<NT> z = x;
  return z *= y;
}
//--------------------------------------------------------------------
/// Division of diadic numbers.
template < typename NT >  inline
diadic<NT>
operator/(const diadic<NT>& x, const diadic<NT>& y)
{
  diadic<NT> z = x;
  return z /= y;
}
//--------------------------------------------------------------------
/// Division of a diadic number by a power of 2.
template < typename NT >  inline
diadic<NT>
operator/(const diadic<NT>& x, int d)
{
  diadic<NT> z(x);
  return z /= d;
}
//--------------------------------------------------------------------
/// Negation opertor.
template < typename NT >  inline
diadic<NT>
operator-(const diadic<NT>& x)
{ 
  return diadic<NT>(-x.num_,x.exp_); 
}
//--------------------------------------------------------------------
/// Addition of an integer and a diadic number.
template < typename NT >  inline
diadic<NT>
operator+(const NT& x, const diadic<NT>& y)
{
  diadic<NT> z(x);
  return z += y;
}
//--------------------------------------------------------------------
/// Addition of a diadic number and an integer.
template < typename NT >  inline
diadic<NT>
operator+(const diadic<NT>& x, const NT& y)
{
  diadic<NT> z = x;
  return z += y;
}
//--------------------------------------------------------------------
/// 
template < typename NT >  inline
diadic<NT>
operator-(const NT& x, const diadic<NT>& y)
{
  diadic<NT> z(x);
  return z -= y;
}
//--------------------------------------------------------------------
/// 
template < typename NT >  inline
diadic<NT>
operator-(const diadic<NT>& x, const NT& y)
{
  diadic<NT> z = x;
  return z -= y;
}
//--------------------------------------------------------------------
/// 
template < typename NT >  inline
diadic<NT>
operator*(const NT& x, const diadic<NT>& y)
{
  diadic<NT> z(x);
  return z *= y;
}
//--------------------------------------------------------------------
template < typename NT >  inline
diadic<NT>
operator*(const diadic<NT>& x, const NT& y)
{
  diadic<NT> z = x;
  return z *= y;
}

//--------------------------------------------------------------------
#ifdef SYNAPS_WITH_GMP
QQ operator*(const diadic<ZZ>& x, const QQ& y)
{
  return (y*x.numerator()/shift_2(QQ(1),x.valuation()));
}
//--------------------------------------------------------------------
QQ& operator*=(QQ& y, const diadic<ZZ>& x)
{
  y *=x.numerator();
  y /=shift_2(QQ(1),x.valuation());
  return y;
}
#endif //SYNAPS_WITH_GMP
//--------------------------------------------------------------------
/// 
template < typename NT >  inline
diadic<NT>
operator/(const NT& x, const diadic<NT>& y)
{
  diadic<NT> z(x) ;
  return z /= y;
}
//--------------------------------------------------------------------
template < typename NT >  inline
diadic<NT>
operator/(const diadic<NT>& x, const NT& y)
{
  diadic<NT> z = x;
  return z /= y;
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
int
sign(const diadic<NT>& x) 
{ 
  return  sign(x.num_); 
}
//--------------------------------------------------------------------
// Booleans operations
//--------------------------------------------------------------------
/// 
template < typename NT >  inline
bool
operator==(const diadic<NT>& x, const diadic<NT>& y)
{ 
  return shift_2(x.num_ ,y.exp_) == shift_2(y.num_,x.exp_); 
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool operator==(const diadic<NT>& x, const NT& y)
{ return shift_2(y, x.exp_) == x.num_; }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator==(const NT& x, const diadic<NT>& y)
{ return y == x; }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator!=(const diadic<NT>& x, const diadic<NT>& y)
{ return ! (x == y); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator!=(const diadic<NT>& x, const NT& y)
{ return ! (x == y); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator!=(const NT& x, const diadic<NT>& y)
{ return ! (x == y); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator<(const diadic<NT>& x, const diadic<NT>& y)
{
  return quotient_cmp(x,y) == SMALLER;
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator<(const diadic<NT>& x, const NT& y)
{
  return quotient_cmp(x,diadic<NT>(y)) == SMALLER;
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator<(const NT& x, const diadic<NT>& y)
{
  return quotient_cmp(diadic<NT>(x),y) == SMALLER;
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>(const diadic<NT>& x, const diadic<NT>& y)
{ 
  return(y < x); 
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>(const diadic<NT>& x, const NT& y)
{ 
  return y < x; 
}
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>(const NT& x, const diadic<NT>& y)
{ return y < x; }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator<=(const diadic<NT>& x, const diadic<NT>& y)
{ return ! (y < x); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator<=(const diadic<NT>& x, const NT& y)
{ return ! (y < x); }
//------------------------------------------------------------------------------------------
template < typename NT >  inline
bool
operator<=(const NT& x, const diadic<NT>& y)
{ return ! (y < x); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>=(const diadic<NT>& x, const diadic<NT>& y)
{ return ! (x < y); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>=(const diadic<NT>& x, const NT& y)
{ return ! (x < y); }
//--------------------------------------------------------------------
///
template < typename NT >  inline
bool
operator>=(const NT& x, const diadic<NT>& y)
{ return ! (x < y); }
//-------------------------------------------------------------------
/// Global functions for the denominator.
template < typename NT >  inline
NT denominator(const diadic<NT>& q)
{ return q.denominator(); }
//--------------------------------------------------------------------
/// Global functions for the numerator.
template < typename NT >  inline
const NT&
numerator(const diadic<NT>& q)
{ 
  return q.numerator(); 
}
//--------------------------------------------------------------------
template<class NT,class X>
bool operator< (const diadic<NT>& d, const X& q)
{
  NT r;
  if(d.valuation()>=0)
    {
      r= shift_2(d.numerator(),d.valuation());
      return (r<q);
    }
  else
    {
      r=NT(q);
      r= shift_2(r,-d.valuation());
      return (d.numerator()<r);
    }
}
//--------------------------------------------------------------------
template<class NT,class X>
bool operator> (const diadic<NT>& d, const X& q)
{
  NT r;
  if(d.valuation()>=0)
    {
      r= shift_2(d.numerator(),d.valuation());
      return (r>q);
    }
  else
    {
      r=NT(q);
      r= shift_2(r,-d.valuation());
      return (d.numerator()>r);
    }
}
//--------------------------------------------------------------------
template<class NT,class X>
bool operator> (const X& q, const diadic<NT>& d)
{
  return (d<q);
}
//--------------------------------------------------------------------
template<class NT,class X>
bool operator< (const X& q, const diadic<NT>& d)
{
  return (d>q);
}
//--------------------------------------------------------------------
//typedef QQ;
//typedef ZZ;
//--------------------------------------------------------------------
namespace let {
  ///
  template < typename NT >
  double
    convert(const diadic<NT>& q, type::As<double> )
  {
    return (convert(q.numerator(),type::As<double>())/shift_2(1.0,q.valuation()));
  }
  ///
 //  template < typename T, typename S >
//   void assign(T & a, const diadic<S>& x)
//   {
//     a = x.numerator()/x.denominator();
//   }
//   ///
//   template < typename T >
//   void assign(diadic<T>& a, const T& x)
//   {
// 	a = diadic<T>(x);
//   }

#ifdef SYNAPS_WITH_GMP 
  void assign(QQ& r, const diadic<ZZ> & x)
  {
    assign(r,x.numerator());
    for(unsigned i=0;i<x.exp_;i++) r/=2;
  }

  template<> 
  void assign(diadic<ZZ> & x, int n)
  {
    assign(x.num_,n);
    x.exp_=0;
    x.normalize();
  }
#endif //SYNAPS_WITH_GMP

}
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
#endif // SYNAPS_ARITHM_DIADIC_H

