/********************************************************************
 *   This file is part of the source code of the SYNAPS kernel.
 *   Author(s): B. Mourrain, GALAAD, INRIA
 ********************************************************************/
#ifndef SYNAPS_ARITHM_SYMBOLIC_H
#define SYNAPS_ARITHM_SYMBOLIC_H
/********************************************************************/
#include <synaps/init.h>
#include <assert.h>
#include <vector>
/********************************************************************/
__BEGIN_NAMESPACE_SYNAPS
//====================================================================
/// Namespace for symbolic expressions.
/** The symbolic expressions are representated by unevaluated trees, 
    with arbitrary nodes.
**/
namespace symbolic {
  template<class X> bool equal(const X& s1, const X& s2) {return(s1==s2);}
  template<class C> struct val;
  template<class C> struct val_rep;

  template<char X,class C=int> struct op;
  template<char X,class C=int> struct op_rep;

  //--------------------------------------------------------------------
  struct Exp;
  //--------------------------------------------------------------------
  struct Exp_rep
  {
  public:
    Exp_rep() {}
    virtual ~Exp_rep()  {};
    virtual char type() const {return '0';};
    virtual Exp operator[] (unsigned ) const;
    virtual bool operator==(Exp_rep* x) const {return true; }
    virtual void print(std::ostream& os)    {os<<"()"; }
    virtual void insert (const Exp & x)    {}
    virtual unsigned nops() const {return 0;}
  };
  //--------------------------------------------------------------------
  /// Generic expression.
  /** Class for expression represented by an unevaluated tree.
   **/
  struct Exp
  {
    //protected:
    Exp_rep* rep_;

  public:
    Exp(): rep_(new Exp_rep) {}
    Exp(Exp_rep* ptr) : rep_(ptr) {}

    Exp_rep* operator -> () {return rep_;}
    Exp_rep* rep         () {return rep_;}
    Exp_rep* rep         ()  const {return rep_;}
    char     type () const  {return rep_->type();}

    Exp      operator[](unsigned i) const {return rep_->operator[](i);}

    bool operator==(const Exp & x) const {return rep_->operator==(x.rep());}
    unsigned nops() const { rep_->nops(); }

    void insert(const Exp & x){ rep_->insert(x); }

    void print(std::ostream& out) const { rep_->print(out); }
    void print(std::ostream& out)       { rep_->print(out); }
  };
  //--------------------------------------------------------------------
  Exp Exp_rep::operator[](unsigned i) const {return Exp();}
  //--------------------------------------------------------------------
  std::ostream& operator << (std::ostream& os, Exp a)
  {
    a->print(os); return os;
  }
  //--------------------------------------------------------------------
  template<class C>  struct val_rep : public Exp_rep
  {
  protected:
    static const char t_='v';
    C  rep_;
  public:
    val_rep() {}
    val_rep(const C& c):rep_(c) {};
    virtual ~val_rep<C>() {};
    virtual char type() const {return 'v';}
    virtual bool operator== (Exp_rep* a) const
    {
      if(a->type()!='v')
	return false;
      else
	{
	  val_rep<C>* v= dynamic_cast<val_rep<C>*>(a);
	  return (equal(rep_,v->rep_));
	}
    }
    //    virtual void print(std::ostream & os) const { os<<rep_; }
    virtual void print(std::ostream & os)       { os<<rep_; }
  };
  //--------------------------------------------------------------------
  /// Class for expressions associated with values.
  template<class C> struct val
  {
    //protected:
    val_rep<C>* rep_;

  public:
    val<C>(): rep_(NULL) {}
    val(val_rep<C>* ptr) : rep_(ptr) {}
    val(const C& x) : rep_(new val_rep<C>(x)) {}

    val_rep<C>* operator -> () {return rep_;}
    val_rep<C>* rep         () {return rep_;}
    inline operator Exp () { return Exp(rep_); }

    //    void print(std::ostream& out) const { rep_->print(out); }
    void print(std::ostream& out)       { rep_->print(out); }
  };

  template<class C>
  std::ostream& operator << (std::ostream& os, const val<C> & a)
  {
    a->print(os); return os;
  }
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'+',C> : public Exp_rep
  {
  protected:
    static const char t_='+';
    std::vector<C> rep_;
  public:
    friend class op<'+',C>;

    op_rep(): rep_() {};
    op_rep(const C& a, const C& b): rep_(2) { rep_[0]=a;rep_[1]=b; }
    virtual char type() const {return '+';}

    Exp  operator[](unsigned i) const { assert(i<nops()); return rep_[i]; }
    unsigned nops() const { return rep_.size(); }

    void insert(const Exp & x) { rep_.push_back(x); }

    void print(std::ostream& os) 
    {
      os<<"(+"; 
      for(unsigned i=0;i<rep_.size();i++) 
	{os <<" "; rep_[i].print(os); }
      os<<")";
    }
  }; 
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'-',C> : public Exp_rep
  {
  protected:
    static const char t_='-';
    std::vector<C> rep_;
  public:
    friend class op<'-',C>;

    op_rep(): rep_() {};
    op_rep(const C& a, const C& b): rep_(2) { rep_[0]=a;rep_[1]=b; }
    virtual char type() const {return '-';}

    Exp  operator[](unsigned i) const { assert(i<nops()); return rep_[i]; }
    unsigned nops() const { return rep_.size(); }

    void insert(const Exp & x) { rep_.push_back(x); }

    void print(std::ostream& os) 
    {
      os<<"(-"; 
      for(unsigned i=0;i<rep_.size();i++) 
	{os <<" "; rep_[i].print(os); }
      os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'*',C> : public Exp_rep
  {
  protected:
    static const char t_='*';
    std::vector<C> rep_;
  public:
    friend class op<'*',C>;
    op_rep(): rep_() {};
    op_rep(const C& a, const C& b): rep_(2) { rep_[0]=a;rep_[1]=b; }
    Exp  operator[](unsigned i) const { return rep_[i]; }
    virtual char type() const {return '*';}
    
    unsigned nops() const { return rep_.size(); }
    void insert(const Exp & x) { rep_.push_back(x); }

    void print(std::ostream& os) 
    {
      os<<"(*"; 
      for(unsigned i=0;i<rep_.size();i++) 
	{os <<" "; rep_[i].print(os); }
      os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'/',C> : public Exp_rep
  {
  protected:
    static const char t_='/';
    std::vector<C> rep_;
  public:
    friend class op<'/',C>;
    op_rep(): rep_() {};
    op_rep(const C& a, const C& b): rep_(2) { rep_[0]=a;rep_[1]=b; }

    virtual char type() const {return '/';}
    
    unsigned nops() const { return rep_.size(); }
    Exp  operator[](unsigned i) const { return rep_[i]; }

    void print(std::ostream& os) 
    {
      os<<"(/"; 
      for(unsigned i=0;i<rep_.size();i++) 
	{os <<" "; rep_[i].print(os); }
      os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'^',C> : public Exp_rep
  {
  protected:
    static const char t_='^';
    C op1_, op2_;
  public:
    friend class op<'^',C>;
    op_rep(const C& a, const C& b): op1_(a), op2_(b) {}
    Exp  operator[](unsigned i) const 
       {if (i==0) return op1_; else return op2_;}
    
    virtual char type() const {return '^';}
    void print(std::ostream& os) {
      os<<"(^ "; op1_.print(os); os<<" "; op2_.print(os); os<<")";
    }
  };

  //--------------------------------------------------------------------
  template<class C>
  struct op_rep<'l',C> : public Exp_rep
  {
  protected:
    static const char t_='l';
    C op1_, op2_;
  public:
    friend class op<'l',C>;
    op_rep(const C& a, const C& b): op1_(a), op2_(b) {}
    Exp  operator[](unsigned i) const 
       {if (i==0) return op1_; else return op2_;}

    char type() const {return 'l';}
    void print(std::ostream& os) {
      os<<"(-> "; op1_.print(os); os<<" "; op2_.print(os); os<<")";
    }
  };
  //--------------------------------------------------------------------
  template<char O, class X>
  struct op
  {
  protected:
    op_rep<O,X>* rep_;
  public:
    op<O,X>(): rep_(new op_rep<O,X>()) {}
    op<O,X>(const X & a, const X& b): rep_(new op_rep<O,X>(a,b)) {}
    inline operator Exp () { return Exp(rep_); }
  };
  //--------------------------------------------------------------------
  /// Return a symbolic tree corresponding to a sum.
  Exp operator+  (const Exp & a, const Exp & b)  
  {
    Exp r  = op<'+',Exp>();
    if(a.type()=='+')
      {
	for(unsigned i=0;i<a.nops();i++) r.insert(a[i]);
      }
    else
      r.insert(a);
    if(b.type()=='+')
      for(unsigned i=0;i<b.nops();i++) r.insert(b[i]);
    else  
      r.insert(b);
    return r;
  }
  //--------------------------------------------------------------------
  Exp operator+ (const Exp & a, int b)  
  {
    Exp i=val<int>(b);
    return a+i;
  }
  //--------------------------------------------------------------------
  Exp operator+ (int a, const Exp & b)  
  {
    Exp i=val<int>(a); 
    return i+b;
  }
  //--------------------------------------------------------------------
  /// Return a symbolic tree corresponding to a substraction.
  Exp operator-  (const Exp & a, const Exp & b)  {return op<'-',Exp>(a,b);}
  //--------------------------------------------------------------------
  /// Return a symbolic tree corresponding to a multiplication
  Exp operator*  (Exp a, Exp b)
  {
    Exp r  = op<'*',Exp>();
    if(a.type()=='*')
      {
	for(unsigned i=0;i<a.nops();i++) r.insert(a[i]);
      }
    else
      r.insert(a);
    if(b.type()=='*')
      for(unsigned i=0;i<b.nops();i++) r.insert(b[i]);
    else  
      r.insert(b);
    return r;
  }
  //--------------------------------------------------------------------
  /// Return a symbolic tree corresponding to a division.
  Exp operator/  (Exp a, Exp b)  {return op<'/',Exp>(a,b);}
  /// Return a symbolic tree corresponding to a power.
  Exp operator^  (Exp a, Exp b)  {return op<'^',Exp>(a,b);}
  /// Return a symbolic tree corresponding to a function.
  Exp lambda     (Exp a, Exp b)  {return op<'l',Exp>(a,b);}
  //======================================================================
  /// Alias for a symbol.
  typedef val<char*> symbol;
  //--------------------------------------------------------------------
  Exp make(char c, Exp a, Exp b)
  {
    if(c=='+')
	return op<'+',Exp>(a,b);
    else if(c=='-')
      {
	return op<'-',Exp>(a,b);
      }
    else if(c=='*')
      {
	return op<'*',Exp>(a,b);
      }
    else if(c=='/')
      {
	return op<'/',Exp>(a,b);
      }
    else if(c=='l')
      {
	return op<'l',Exp>(a,b);
      }
    assert("op not defined"==0);
    //    return op<'',Exp>(a,b);
  }  

  Exp subs(Exp  r, Exp b)
  {
    if(r.type() != 'l')
      return Exp();
    else
      {
	Exp var=r[0];
	Exp val=r[1];
	if(b.type()=='v')
	  if(b==var)
	    return val;
	  else
	    return b;
	else 
	  {
	    return make(b.type(),subs(r,b[0]),subs(r,b[1]));
	  }
      }
    return Exp();
  }
      
  //--------------------------------------------------------------------


  bool equal(char* s1, char* s2) {return (strcmp(s1,s2)==0);}
  //--------------------------------------------------------------------
} //generic
//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif // SYNAPS_ARITHM_SYMBOLIC_H

