/*********************************************************************
*      This file is part of the source code of SYNAPS kernel.        *
*   Author(s): B. Mourrain, GALAAD, INRIA                            *
**********************************************************************
History: 
$Id: Svd.h,v 1.1 2005/07/11 08:33:34 mourrain Exp $
**********************************************************************/
#ifndef SYNAPS_LINALG_SVD_H
#define SYNAPS_LINALG_SVD_H
#include <synaps/init.h>
#include <synaps/linalg/VectDse.h>
#include <synaps/linalg/MATRIX.m>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//====================================================================
/** \file Svd.h                                                    
    \brief Generic Singular Value Decomposition.
**/
//--------------------------------------------------------------------
/// Singular Value Decomposition
template <class M> inline 
VectDse<typename realof<typename M::value_type>::T> Svd(M & A)
{
  using namespace MATRIX;
  typedef VectDse<typename realof<typename M::value_type>::T> VT;
  VT x(std::min(A.nbcol(),A.nbrow()));
  svd(x.rep(),A.rep());  
  return x;
}
//--------------------------------------------------------------------
/// Singular Value decomposition, with left factor.
template <class M> inline 
VectDse<typename realof<typename M::value_type>::T> Svd(M & A, M & U)
{
  typedef VectDse<typename realof<typename M::value_type>::T> VT;
  unsigned int l = std::min(A.nbcol(),A.nbrow());
  VT x(l);
  U.rep().resize(A.nbrow(),A.nbrow());
  using namespace MATRIX;  svd(x.rep(),A.rep(), U.rep());
  return x;
}
//--------------------------------------------------------------------
/** Singular Value decomposition, with left and right orthogonal factor.
 We have @A= U\, Diag(S) \, V@. 
 **/ 
template <class M> inline
VectDse<typename realof<typename M::value_type>::T> Svd(M & A, M & U, M & V)
{
  typedef VectDse<typename realof<typename M::value_type>::T> VT;
  unsigned int  m = std::max(A.nbcol(),A.nbrow());
  VT x(m);
  U.rep().resize(A.nbrow(),A.nbrow());
  V.rep().resize(A.nbcol(),A.nbcol());
  using namespace MATRIX; svd(x.rep(),A.rep(), U.rep(), V.rep());
  return x;
}

//--------------------------------------------------------------------
/** 
  Singular Value decomposition, with left and orthogonal right factor.
  We have @A= U \, V@.
**/
template <class M> inline 
VectDse<typename realof<typename M::value_type>::T> Svdx(M & A, M & U, M & V)
{
  typedef VectDse<realof<typename M::value_type> > VT;
  unsigned int l = std::max(A.nbcol(),A.nbrow());
  VT x(l);
  U.rep().resize(A.nbrow(),A.nbrow());
  V.rep().resize(A.nbcol(),A.nbcol());
   using namespace MATRIX; svdx(x.rep(), A.rep(), U.rep(), V.rep());
  return x;
}

//----------------------------------------------------------------------
/** Condition number based on Svd. It returns the largest singular value, 
    divided by the smallest.
 **/
template <class M> 
inline typename M::value_type Cond(M & A)
{
  typedef typename linalg::rep1d<typename realof<typename M::value_type>::T> VT;
  unsigned int l = std::min(A.nbcol(),A.nbrow());
  VT S(l);
  using namespace MATRIX;  svd(S,A.rep());
  return (S[0]/S[l-1]);
}
//====================================================================
__END_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
#endif // SYNAPS_LINALG_SVD_H


