/*********************************************************************
*      This file is part of the source code of SYNAPS library        *
*   Author(s): L. Deschamps                                          *
*              G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef synaps_linalg_TENSOR3D_H
#define synaps_linalg_TENSOR3D_H

/*********************************************************************/

#ifdef HAVE_CONFIG_H
#include <synaps/init.h>
#endif 

#include <iostream>
#include <assert.h>

#include <synaps/linalg/MethodName.h>

__BEGIN_NAMESPACE_SYNAPS

/*********************************************************************/
/*! Implementation of useful functions for the matrix-3d. */
namespace TENSOR3D {
 
  typedef unsigned size_type;
  
  /** Output function for general 3D-matrices. It is called 
      by the output operator @<<@ of the classes of 3D-matrices. **/
  template < class R >
  inline std::ostream & print( std::ostream & os, const R & m ) {
    typedef typename R::size_type  size_type;
    
    for ( size_type k = 0 ; k < m.nbhgt() ; k++ ) {
      for ( size_type i = 0 ; i < m.nbrow(); ++i ){
	os << "[ ";
	for ( size_type j = 0 ; j < m.nbcol() ; j++ )
	  os<< m( i, j, k ) << " ";
	os << "]"<< std::endl;
      }
      os << std::endl;
    }
    return os;
  }

 /** Output function for general 3D-matrices. It is called 
      by the output operator @<<@ of the classes of 3D-matrices. **/
  template < class R >
  inline std::ostream & printsign( std::ostream & os, const R & m ) {
    typedef typename R::size_type  size_type;
    
    for ( size_type k = 0 ; k < m.nbhgt() ; k++ ) {
      for ( size_type i = 0 ; i < m.nbrow(); ++i ){	
        for( size_type u=0;u<k;u++) os<<" ";
	os << "[";
	for ( size_type j = 0 ; j < m.nbcol() ; j++ )
	  if(m(i,j,k)<0) os<< "-"; else if (m(i,j,k)>0) os<< "+"; else os << "0"; 
	os << "]"<< std::endl;
      }
      os << std::endl;
    }
    return os;
  }

  
  /** Input function for general 3D-matrices. The input format is of 
      the form @m n c00 c0 ...@ where @m@ is the number of rows, @n@
      is the number of columns @o@ is the number of "heigth" and the
      coefficients are read row by row and height by height. **/

  template< class R >  inline
  std::istream & read( std::istream & is, R & M ){
    typedef typename R::size_type size_type;
    size_type m, n, o;
    is >> m >> n >> o;
    M.rep().resize( m, n );
 
    for ( size_type k = 0; k < o ; k++ )
      for ( size_type i = 0 ; i < m ; i++ )
	for (size_type j = 0 ; j < n ; j++ ) {
	  is >> M(i,j);
	}
    return( is );
  }

  /** Initialise all the entries of a with the value c.  **/

  template< class R, class C > inline
  void init( R & a, const C & c) { 
    typedef typename R::size_type size_type;

    for ( size_type i = 0 ; i < a.nbrow() ; i++ ) 
      for ( size_type j = 0 ; j < a.nbcol(); j++ )
	for ( size_type k = 0 ; k < a.nbhgt() ; k++)
	  a( i, j, k ) = c;
  }

  /** Assignement of a dense matrix to another one, assuming they have the
      same size, but not necessarily the same type. **/

  template < class R1, class R2 >
  inline void assign( R1 & r, const R2 & m ) {
    typedef typename R1::size_type  size_type;
    
    for ( size_type i = 0 ; i < m.nbrow(); ++i )
      for (size_type j = 0 ; j < m.nbcol(); j++ )
	for( size_type k = 0 ; k < m.nbhgt() ; k++ )
	  r( i, j, k ) = m( i, j, k );
  }
  
  
} // TENSOR3D

/*********************************************************************/

__END_NAMESPACE_SYNAPS

#endif // synaps_linalg_TENSOR3D_H
