#ifndef SYNAPS_SOLVE_BEZIER2D_H
#define SYNAPS_SOLVE_BEZIER2D_H


#include <synaps/init.h>
#include <synaps/base/Seq.h>
#include <synaps/linalg.h>
#include <synaps/mpoly.h>
#include <synaps/solve/Bezier1d.h>


__BEGIN_NAMESPACE_SYNAPS


/// Used for storing the Bernstein coefficients and the refinements made
typedef  MatrDse<array::rep2d<long double> > MAT;



/// Structure used to keep track of the coordinates of the box or matrix we are working on  and finally they are passed back as the solution set.
struct box2d
{

  long double x0, x1, y0, y1;
  typedef box2d self_t;

  box2d (): x0(0),x1(1),y0(0),y1(1) {}
  box2d(const box2d & b1): x0(b1.x0),x1(b1.x1),y0(b1.y0),y1(b1.y1) {}

  self_t operator=(const box2d & b1)
    {
      x0=b1.x0; x1=b1.x1; y0=b1.y0; y1=b1.y1;
      return *this;
    }
};

std::ostream& operator<<(std::ostream & os, const box2d & b)
{
  os<<"[["<<b.x0<<","<<b.x1<<"],["<<b.y0<<","<<b.y1<<"]]";
  return os;
}


/* Functions used to change the appropriate co-ordinates of the box for 
different splits of the matrix, depending on whether it is for the first or 
the second newly generated matrix, that is has to change the coordinates for.
PARAMETERS: @b@ :box whose coordinates have to be changed. 
*/
void rsplit1box2d(box2d &b)
{
b.x1=(b.x0+b.x1)/2;
}


void rsplit2box2d(box2d &b)
{
b.x0=(b.x0+b.x1)/2;
}

void csplit1box2d(box2d &b)
{
b.y1=(b.y0+b.y1)/2;
}

void csplit2box2d(box2d &b)
{
b.y0=(b.y0+b.y1)/2;
}


/** Function that checks the dimensions of the box from its coordinates , 
compares it with the precision constant @e@ and then returns the appropriate 
flag  value:  3 if the two dimensions are  $< e$, 
2 if the $y$-dimension  is $<e$, 1 if the $x$-dimension is $< e$ and $0$ otherwise.
**/
int chkbox2d(box2d b,double e)
{
  int ret=0;
  if((b.x1-b.x0)<e) ret+=1;
  if((b.y1-b.y0)<e) ret+=2;
  return(ret);
}

/** Isolate the root of the polynomial
represented by the sequence of coefficient @l@ on the intervale @[c0,c1]@ within the precision @eps@.
**/
template<class T>
long double isolate(const T &l, long double c0, long double c1,
		    const double eps)
{
  if((c1-c0)<eps)
    {return (c1+c0)/2;}
  else if (*l.begin()==0) return c0;
  else 
    {
      array::rep1d<double> l0(l.size()), l1(l);
      subdivise(l0,l1);
      if(Variation(l0)==1) 
	return isolate(l0,c0,(c0+c1)/2,eps);
      else if(Variation(l1)==1)
	return isolate(l1,(c0+c1)/2,c1,eps);
      else if(*l1.begin()==0) return (c0+c1)/2;
      else 
	{cout<<"Pb in the isolation"<<endl; return (c0+c1)/2;}
  }
}


/** Functional object used in the splitting process.  The operator
@SPL::operator(mat,b)@ is used to known if the representation has to be
split. @ma@ is the matrix of control coefficients, @b@ is the box. The
value @prec@ is the precision of approximation.
**/
struct SPL
{
  long double prec;
  SPL(): prec(0.0001)  {}
  SPL(long double eps): prec(eps) {}
  SPL(const SPL & spl):prec(spl.prec) {}
      
  template<class T> int operator()(const T & ma, const box2d & b) const;
};


/** Operator used in the recursive splitting algorithm. Returns a value @spl@
 which  precises if a splitting step is necessary or the type of the arc, with
 the  following convention:
\latexonly
\begin{center}
\begin{tabular}{l|lll|l}
-2 & split by column     \\
-1 & split by row \\
0  & remove the box && 1  & arc through S \\
2  & arc through N  && 3  & arc through SN \\
4  & arc through E  && 5  & arc through SE \\
6  & arc through NE && 7  & arc through ESN \\
8  & arc through W  && 9  & arc through SW \\
10 & arc through NW && 11 & arc through WSN \\
12 & arc through EW && 13 & arc through EWS \\
14 & arc through EWN && 15 & arc through EWSN 
\end{tabular}\end{center}
\enlatexonly
**/
template<class T>
int SPL::operator()(const T & ma,const box2d & b) const  
{
 
 int split=0, arc=0;
 int nr=ma.nbrow()-1;
 int nc=ma.nbcol()-1;
 int rowvar0,rowvar1,colvar1,colvar0;


 MAT::row_type rt;
 MAT::col_type ct;

 MAT::row_type r0=Row(ma,0);
 MAT::row_type r1=Row(ma,nr);  
 MAT::col_type c0=Col(ma,0);
 MAT::col_type c1=Col(ma,nc);

 rowvar0=Variation(r0);
 rowvar1=Variation(r1);
 colvar0=Variation(c0);
 colvar1=Variation(c1);
 // cout <<"-> "<<rowvar0<<" "<<rowvar1<<" "<<colvar0<<" "<<colvar1<<" : ";

 //Spliting information; Arc type.
 split=0; arc=0;


 //Shape size information.
 int chk=chkbox2d(b,prec); 
 //cout <<chk<<" ";

 // small boxin both direction
 if(chk==3){
   if(rowvar0) arc+=1; 
   if(rowvar1) arc+=2;
   if(colvar0) arc+=4;
   if(colvar1) arc+=8; 
   if(arc==13 || arc==14)  
     return 12;
   else if (arc==7 || arc==11) 
     return 3;
   else if(arc)
     return 15;
   else 
     return 0;
 } else if (chk==1) {
   if(rowvar0) arc+=1; 
   if(rowvar1) arc+=2;
   if(colvar0==1) arc+=4; else if(colvar0>1) return -2;
   if(colvar1==1) arc+=8; else if(colvar1>1) return -2;
   if(arc==7 || arc==11 || arc==13 || arc==14 || arc==15) return -2;
   if(arc) return arc;
 } else if(chk==2) {
   if(rowvar0==1) arc+=1; else if(rowvar0>1) return -1;
   if(rowvar1==1) arc+=2; else if(rowvar1>1) return -1;
   if(colvar0) arc+=4;
   if(colvar1) arc+=8;
   if(arc==7 || arc==11 || arc==13 || arc==14 || arc==15) return -1;
   if(arc) return arc;
 } else {
   for(int i=0;i<=nr;i++)
     {
       rt=Row(ma,i);
       if (Variation(rt)>1) return -1;
     }
   for(int i=0;i<=nc;i++)
     {
       ct=Col(ma,i);
       if (Variation(ct)>1) return -2;
     }
   if(rowvar0==1) arc+=1; else if(rowvar0>1) return -1;
   if(rowvar1==1) arc+=2; else if(rowvar1>1) return -1;
   if(colvar0==1) arc+=4; else if(colvar0>1) return -2;
   if(colvar1==1) arc+=8; else if(colvar1>1) return -2;

   if(arc==7)  return -1;
   else if(arc==11) return -1;
   else if(arc==13) return -2;
   else if(arc==14) return -2;
   else if(arc==15) return -2;   
   else if(arc==0)  return 0;
   return arc;
 }
 return 0;
}

 


/** Core function of the program. It splits the box according to the
value returned by the function object @spl@ 
and apply recursively the subroutine to the two subrepresentations.
The two @int@ parameters @s1@ and @s2@ (which should be $\pm 1$)
specify which  symmetry we apply to the coordinates of the solutions.
The solutions found in this process are insert at the end of @sol@.
**/
template<class T,class S>
void subdivision(vector<box2d>& sol,
		  const T& b, const box2d & box,
		  int s1, int s2, 
		  const S & spl)
{
  box2d box0(box), box1(box);
  
  int d1= b.nbrow();
  int d2= b.nbcol();
 
  MAT::row_type r0=Row(b,0), r1=Row(b,(d1-1));  
  MAT::col_type c0=Col(b,0), c1=Col(b,(d2-1));  

  int oui=spl(b,box);
  //  cout<<oui<<"\n"<<endl;

  if(oui==-1)
    {
      MAT b0(d1,d2);
      MAT b1(b.rep());
      
      for(int i=0;i<d1;i++)
       {
	 MAT::row_type r0=Row(b0,i), r1=Row(b1,i);
	 subdivise(r0,r1);
       }
     
      rsplit1box2d(box0);
      rsplit2box2d(box1);
      subdivision(sol,b0,box0,s1,s2,spl);
      subdivision(sol,b1,box1,s1,s2,spl);
    }
  else if(oui==-2)
    {
      MAT b0(d1,d2);
      MAT b1(b.rep());
      for(int i=0;i<d2;i++)
	{
	  MAT::col_type c0=Col(b0,i), c1=Col(b1,i);
	  subdivise(c0,c1);
	}
     csplit1box2d(box0);
     csplit2box2d(box1);
     
     subdivision(sol,b0,box0,s1,s2,spl);
     subdivision(sol,b1,box1,s1,s2,spl); 
    }  
  else if(oui>0) {
    
    box2d sbox(box);
    
    int c=0;
    // Arc through S
    if(oui%2)
      {
	array::rep1d<double> l0(d2);
	for(int i=0;i<d2;++i) {l0[i]=r0[i];}
	sbox.x0= isolate(l0,box.x0,box.x1,spl.prec);
	sbox.y0= box.y0;
	c++;
      }
    oui/=2;
    // Arc through N
    if(oui%2)
      {
	array::rep1d<double> l0(d2);
	for(int i=0;i<d2;++i) {l0[i]=r1[i];}
	if(c){
	  sbox.x1= isolate(l0,box.x0,box.x1,spl.prec);
	  sbox.y1= box.y1;
	} else {
	  sbox.x0= isolate(l0,box.x0,box.x1,spl.prec);
	  sbox.y0= box.y1;
	}
	c++;
      }
    oui/=2;
    // Arc through E
    if(oui%2)
      {
	array::rep1d<double> l0(d1);
	for(int i=0;i<d1;++i) {l0[i]=c0[i];}
	if(c){
	  sbox.x1= box.x0;
	  sbox.y1= isolate(l0,box.y0,box.y1,spl.prec);
	} else {
	  sbox.x0= box.x0;
	  sbox.y0= isolate(l0,box.y0,box.y1,spl.prec);
	}
	c++;
      }
    oui /=2;
    // Arc through 0
    if(oui%2)
      {
	array::rep1d<double> l0(d1);
	for(int i=0;i<d1;++i) {l0[i]=c1[i];}
	if(c){
	  sbox.x1= box.x1;
	  sbox.y1= isolate(l0,box.y0,box.y1,spl.prec);
	} else {
	  sbox.x0= box.x1;
	  sbox.y0= isolate(l0,box.y0,box.y1,spl.prec);
	}
	c++;
      }

    if(c>1) {
      sbox.x0=sbox.x0/(1-sbox.x0);
      sbox.x1=sbox.x1/(1-sbox.x1);
      sbox.y0=sbox.y0/(1-sbox.y0);
      sbox.y1=sbox.y1/(1-sbox.y1);
      sbox.x0*=s1; sbox.x1*=s1;
      sbox.y0*=s2; sbox.y1*=s2;
      //      cout <<".... "<<sbox<<endl;
      sol.push_back(sbox);
    }
  }
}
 

//----------------------------------------------------------------------
template<class MT>
void shrink(MT & ma, box2d & b, long double rg1, long double rg2)
{
  for(unsigned i=0;i<ma.nbrow();i++)
    for(unsigned j=0; j<ma.nbcol()-1;j++)
      for(int k= ma.nbcol()-1;k>(int)j;k--)
	{
	  ma(i,k)= rg1*ma(i,k)+(1-rg1)*ma(i,k-1);
	}
  for(unsigned i=0;i<ma.nbcol();i++)
    for(unsigned j=0; j<ma.nbrow()-1;j++)
      for(int k= ma.nbrow()-1;k>(int)j;k--)
	{
	  ma(k,i)= rg1*ma(k,i)+(1-rg1)*ma(k-1,i);
	}
  b.x1=rg1; b.y1=rg2;
}

//----------------------------------------------------------------------
/** Function which outputs the list of segments representing the curve
$p(x_0,x_1)=0$ associated with the coefficient matrix @MT@,
 within a precision @eps@ in the box of $x$-dimension
@rg1@ and $y$-dimension @rg2@. 
The functionnal object @split@ (operator ()) is used to specify how to split the box.
**/
template<class MT, class SPLIT>
Seq<box2d> Draw(const MT & p,
		const SPLIT& split,
		long double rg1, long double rg2)
{
  
  //cout <<"matrix\n"<<b<<endl;
  //initialised with [0,1]x[0,1]

  int d1=p.nbrow();
  int d2=p.nbcol();

  Seq<box2d> sol;//= new vector<box2d>;

  box2d box1, box2, box3, box4;
  MT bt1(p), bt2(p), bt3(p), bt4(p);

  //cout <<"                                 [0,+oo[x[0,+oo["<<endl;
  shrink(bt1,box1,rg1,rg2);  cout <<bt1<<endl;
  subdivision(sol.rep(),bt1,box1,1,1,split);

  //changing x to -x
  //cout <<"                                 ]-oo, 0]x[0,+oo["<<endl;
  for (int i=0;i<d1;i++)
    for (int j=1;j<d2;j+=2) {bt3(i,j)*=-1;} 
  shrink(bt3,box3,rg1,rg2);  cout <<bt2<<endl;
  subdivision(sol.rep(),bt3,box3,-1,1,split);
  
  //changing y to -y 
  //cout <<"                                 [0,+oo[x]-oo,0]"<<endl;
  for (int i=1;i<d1;i+=2)
    for (int j=0;j<d2;j++) {bt2(i,j)*=-1;} 
  shrink(bt2,box2,rg1,rg2);  cout <<bt3<<endl;
  subdivision(sol.rep(),bt2,box2,1,-1,split); 
  
  //changing x to -x , y to -y
  //cout <<"                                 ]-oo,0]x]-oo,0]"<<endl;
   for (int i=0;i<d1;i++)
    for (int j=1;j<d2;j+=2) {bt4(i,j)*=-1;} 
   for (int i=1;i<d1;i+=2)
     for (int j=0;j<d2;j++) {bt4(i,j)*=-1;} 
  shrink(bt4,box4,rg1,rg2);  cout <<bt4<<endl;
  subdivision(sol.rep(),bt4,box4,-1,-1,split);

 return sol;
}

//----------------------------------------------------------------------
template<class R,class O>
const MAT build_matrix(const MPol<R,O> & p)
{
  int d0= Degree(p,0)+1, d1=Degree(p,1)+1;
  MAT  A(d1,d0);
  MATRIX::init(A,0);
  for(typename MPol<R,O>::const_iterator it=p.begin();it!=p.end();it++)
    {
      A((*it)[1],(*it)[0])= it->coeff();
    }
  for (int k=0;k<d0;k++)    
   for (int j=0;j<d1;j++)
   { 
    A(k,j)/= (binomial(k,(d0-1))*binomial(j,(d1-1)));  
   }

  return A;
}
//----------------------------------------------------------------------
/** Function which outputs the list of segments representing the curve
$p(x0,x1)=0$ within the precision @spl.prec@ in the box of $x$-dimension
@rg1@ and $y$-dimension @rg2@. The functionnal object @split@ (indeed
the operator SPL::operator()() ) is used to specify how to split the box.   
**/
template<class R, class O, class SPLIT>
Seq<box2d> Draw(const MPol<R,O> & p,
		const SPLIT& spl,
		long double rg1=0.93, long double rg2=0.93)
{
  MAT A = build_matrix(p);
  return Draw(A,spl,rg1,rg2);
}


__END_NAMESPACE_SYNAPS


//----------------------------------------------------------------------
#endif // SYNAPS_SOLVE_BEZIER2D_H

