/********************************************************************
 *   This file is part of the source code of the SYNAPS library.
 *   Author(s): J.P. Pavone, GALAAD, INRIA
 *   $Id: sbdslv.h,v 1.1 2005/07/11 10:03:51 jppavone Exp $
 ********************************************************************/
/** \file  sbdslv.h                                                 **/
#ifndef SYNAPS_SOLVE_SBDSLV_H
#define SYNAPS_SOLVE_SBDSLV_H
/********************************************************************/
#include <set>
#include<vector>
#include <synaps/init.h>
#include <synaps/base/xml.h>
#include <synaps/mpol.h>
#include <synaps/linalg/VectDse.h>
#include <synaps/linalg/MatrDse.h>
#include <synaps/msolve/sbdslv/method.hpp>
#include <synaps/msolve/sbdslv/support.hpp>
#include <synaps/base/chrono.h>
#include <synaps/mpol/MPolDse.h>
//====================================================================
__BEGIN_NAMESPACE_SYNAPS
//====================================================================
namespace sbdslv
{
#define SBDSLV_RD    0
#define SBDSLV_RDL   1
#define SBDSLV_RDS   2
#define SBDSLV_RDRDL 3
#define SBDSLV_RDRDLRDS 4
#define SBDSLV_RDRDS    5
#define SBDSLV_RDLRDS   6
#define SBDSLV_RDSRDL   7
#ifndef SBDSLV_MTH
#define SBDSLV_MTH SBDSLV_RDRDL
#endif
  
  template< class system, int mth > /* selection de methodes */
  struct select_mth
  { typedef sbdslv::method<system,strgy::newton<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
  struct select_mth<system,SBDSLV_RD>
  { typedef sbdslv::method<system,strgy::simple<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
  struct select_mth<system,SBDSLV_RDL>
  { typedef sbdslv::method<system,strgy::newton<system>,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system>
  struct select_mth<system,SBDSLV_RDRDL>
  { typedef sbdslv::method<system,strgy::newton< system, strgy::simple<system> >,rdslv::parallel,sbdrl::parametric> 
    result_t; };
  template< class system >
  struct select_mth<system,SBDSLV_RDRDLRDS>
  { typedef sbdslv::method<system,strgy::cvmatrix< system, strgy::newton<system,strgy::simple<system> > >,rdslv::parallel,sbdrl::parametric> result_t; };
  template< class system > 
    struct select_mth<system,SBDSLV_RDS>
  { typedef sbdslv::method<system,strgy::cvmatrix<system>,rdslv::parallel,sbdrl::parametric> result_t; };

  template< class system > 
    struct select_mth<system,SBDSLV_RDRDS>
  { typedef sbdslv::method<system,strgy::cvmatrix<system, strgy::simple<system> >,rdslv::parallel,sbdrl::parametric> result_t; };
  
  template< class system > 
    struct select_mth<system,SBDSLV_RDLRDS>
  { typedef sbdslv::method<system,strgy::cvmatrix<system, strgy::newton<system> >,rdslv::parallel,sbdrl::parametric> result_t; };

  template< class system > 
    struct select_mth<system,SBDSLV_RDSRDL>
  { typedef sbdslv::method<system,strgy::cvmatrix<system, strgy::newton<system> >,rdslv::parallel,sbdrl::parametric> result_t; };
  
};

/** Class which specifies a subdivision solver for multivariate polynomials.
   The class @T@ is the type of coefficients for the polynomials.
   The type used to store the domain is @DOMAIN@.
   The preconditioning and reduction and subdivision strategy is let 
   by the integer @mth@.
   \param SBDSLV_RD is performing simple reductions.
   \param SBDSLV_RDL is performing reductions and a local preconditioning.
   \param SBDSLV_RDS is performing reductions and a global preconditioning.
   \param SBDSLV_RDRDL is performing two steps of reductions and a local preconditioning. 
   \param SBDSLV_RDRDLRDS is performing two steps of reductions and a local preconditioning and again a reduction.
**/
template< class T, int mth = SBDSLV_RDRDL>
struct SBDSLV
{
  typedef T coeff_t;
  T prec;
  /// Construction of the solver for the precision \c precision.
  SBDSLV( const T& precision = (T)1e-6 ): prec(precision){};
};

namespace SolveSBDSLV
{
  template< class T, class eenv_t, class MPOL >
  void fill_data( T * data, eenv_t * env, const MPOL& p, binomials<T>& benv  )
  {
    typename MPOL::const_iterator mn; int i,a;
    std::fill(data,data+env->sz(),0); 
    mpolfill(data,p,*env);
    convertm2b (data,*env,benv);
  };
  template< class monom >
  void scan_monom_env ( std::set< int >& env, const monom& m )
  { for ( unsigned i = 0; i < (m.rep()).size(); i++ ) if ( (m.rep())[i] != 0 ) env.insert(i); };
}



/** This is the general interface function for solving multivariate polynomial
systems. 
 \param l is a sequence of polynomials (with @begin()@ and @end()@ member functions)
 \param mth specifies the strategies and the domain for the resolution.
 \param verbose set to @true@ output complete information on the resolution.
**/
template< class MPC, class C, int MTH, class Domains >
Seq<VectDse<C> >  solve( const MPC& l, const SBDSLV< C, MTH >& mth, 
			 const Domains & dom, 
			 bool verbose = false  )
{
  typedef C T;
  typedef typename MPC::value_type          MPoly;
  typedef MPoly                             mpoly_t;
  typedef MPC                               plist_t;
  typedef typename plist_t::const_iterator  pptr_t;
  typedef typename MPoly::const_iterator    mptr_t;
  typedef typename MPoly::monom_t           monom_t;
  typedef SBDSLV< C, MTH >         mth_t;
  typedef int            sz_t;
  sz_t nvars, neqs, * nvrs, ** vrs, ** szs; 
  pptr_t cpl;
  int e;
  nvars = 0;
  // recuperation de l'environement global
  for ( neqs = 0, cpl = l.begin(); cpl != l.end(); cpl++, neqs++  )
    for ( mptr_t mn = cpl->begin(); mn != cpl->end(); mn++ )
      nvars = std::max(nvars,mn->nvars()+1);
  // recuperation de l'environement de chaque equation
  std::set<int> * seenvs = new std::set< int > [ neqs ];
  for ( e = 0, cpl = l.begin(); e < neqs; e++, cpl++  )
    for ( mptr_t mn = cpl->begin(); mn != cpl->end(); mn++ ) 
      SolveSBDSLV::scan_monom_env(seenvs[e],*mn);
  
  //  assert(sizeof(sz_t)==sizeof(sz_t*));
  
  sz_t  s = 0;
  for ( e = 0; e < neqs; s += seenvs[e].size(), e++ );
  nvrs = new sz_t [neqs+2*s];
  vrs  = new sz_t*[2*neqs];
  szs  = vrs + neqs;
  sz_t * base = nvrs + neqs;
  std::set<int>::iterator seenvit;  
  for ( e = 0, cpl = l.begin(); e < neqs; base += 2*nvrs[e++], cpl++ )
    {
      nvrs[e] = seenvs[e].size();
      vrs [e] = base;
      szs [e] = base + nvrs[e];
      int k = 0;
      for ( seenvit = seenvs[e].begin(); k < nvrs[e]; seenvit++, k++  ) 
	{
	  vrs[e][k]  = *seenvit;
	  szs[e][k]  = MPOLDST::degree( *cpl, *seenvit ) + 1;
	};
    };
  
  std::vector<C> result;
  C * dmns = 0;
  if ( dom.size() >= 2*nvars )
    {
      using namespace let;
      dmns = new C[ 2*nvars ];
      for ( int i = 0; i < 2*nvars; i ++ ) assign(dmns[i],dom[i]);
    };
  
  chrono tm;
  std::vector<T> tmp;
  
  {  /* remplissage et appel */
    using namespace sbdslv;
    binomials<T> bznv;
    typedef  sbdslv::system<T> system_t;
    system_t sys( neqs, nvars, nvrs, (const int**)vrs, (const int**)szs, (T)(mth.prec) );
    for ( e = 0, cpl = l.begin(); e < neqs; e ++, cpl++  )
      SolveSBDSLV::fill_data( sys.data(e), sys.env(e), *cpl, bznv );
    sbdslv::system_ctrl< std::vector< T > >  intf(tmp);
    tm.start();
    typename sbdslv::select_mth<system_t,MTH>::result_t __mth;
    __mth.launch(&sys,intf,dmns);
    tm.stop();
    if ( verbose )
      {
	std::cout << " iterations  = " << __mth.m_niter << std::endl;
	std::cout << "subdivisions = "<< __mth.m_nsbd  << std::endl;
      };
  };
  
  if ( !verbose && dmns ) delete[] dmns;
  delete[] seenvs;
  delete[] nvrs;
  delete[] vrs;
  if ( verbose )
    {
      std::cout << "<domain>";
      if  ( dmns )
	{
	  std::cout << std::endl;
	  for ( int i = 0; i < nvars; i ++ )
	    std::cout << "\t" 
		      << "x" << i//monom_t::var_of_index(i) 
		      << " = [ " << dmns[2*i] 
		      << "," 
		      << dmns[2*i+1] 
		      << " ]";
	}
      else std::cout << "[0,1]^" << nvars;
      std::cout << "</domain>\n";
      //	std::cout << "\treductions   = " << bsys.nrest() << std::endl;
      //	std::cout << "\titerations   = " << bsys.niter() << std::endl;
      std::cout << "<solver_time>"; tm.write(std::cout); 
      std::cout <<"</solver_time>"<< std::endl;	
      if ( dmns ) delete[] dmns;
    };
  
  int nsols = tmp.size() / (2*nvars);
  T boxes[nsols*2*nvars];
  std::copy(&tmp[0], &tmp[0]+2*nvars*nsols,boxes);
  int nsols2 = sbdslv::clean_result( &tmp[0], nvars, nsols,(T)1e-3 );
  if ( verbose ) { 
    xml::print(std::cout,nsols,"nb_box")<< std::endl;
    for ( unsigned i = 0; i < nsols; i ++ )
      {
	std::cout << "\t[ ";
	for ( unsigned j = 0; j < nvars; j ++ )
	  std::cout << (boxes[2*(i*nvars+j)]+boxes[2*(i*nvars+j)+1])/2.0 << " ";
	std::cout << "]";
	std::cout << std::endl;
      };
    xml::print(std::cout,nsols2,"nb_sol")<< std::endl; 
    //    if ( nsols ) vctops::print(&tmp[0],nvars*nsols );
  };


  tmp.resize( nsols2 * nvars );
  //  cout <<"-----"<<endl;
  VectDse<T> vsol(nvars, AsSize());
  Seq<VectDse<T> > res;
  for(unsigned i=0;i<nsols2;++i)
    {
      for(unsigned j=0; j<nvars; j++)
	vsol[j]=tmp[i*nvars+j];
      res.push_back(vsol);
    }
  return res;
};

//----------------------------------------------------------------------
template< class MPC, class C, int MTH>
Seq<VectDse<C> >  solve( const MPC& l, const SBDSLV< C, MTH >& mth, 
			 bool verbose = false  )
{
  return solve(l,mth,VectDse<C>(),verbose);
}
//----------------------------------------------------------------------

template<class C, int METHOD>
struct SbdSolver
{
  typedef sbdslv::system<C> system_t;
  system_t  * sys;
  

  
  template< class MPDI >
  SbdSolver( MPDI  first, MPDI last, const C& eps )
  {

    int    ne = 0;
    MPDI eptr;
    for ( eptr = first; eptr != last; eptr ++ ) ne++;
    int nvrs[ne];
    const int *szs[ne];
    int e;
    int mxvr = 0;
    std::set<int> gvrs;
    for ( e = 0, eptr = first; eptr != last; eptr ++, e++  ) { 
      nvrs[e] = eptr->nvr(); 
      szs[e]  = eptr->szs();
      const int * tmp = eptr->vrs();
      for ( int v = 0; v < eptr->nvr(); v ++ ) gvrs.insert(tmp[v]);
    };
    int mxv = *(gvrs.rend());
    int lnames[ mxv+1 ];
    int nv = 0;
    for ( std::set<int>::const_iterator it = gvrs.begin(); it != gvrs.end(); lnames[*it++] = nv++ );
    int * vrdat = new int[ nv * ne ];
    int * vrptr = vrdat;
    const int *  vrs[ne];
    e = 0;
    for ( eptr = first; eptr != last; vrptr+= eptr->nvr(), e++, eptr ++ ) {
      vrs[e] = vrptr;
      const int * tmp = eptr->vrs();
      for ( int v = 0; v < eptr->nvr();  v ++ )
	vrptr[v] = lnames[ tmp[v] ];	
    };
    //    std::cout << ne << " " << nv << std::endl;
    assert(ne>=nv);
    sys = new sbdslv::system<C>( ne, nv, nvrs, (const int**)vrs, (const int**)szs, eps );
    delete[] vrdat;
  };
  
  void add( const mpoldse::bernstein<C>& data, int e )
  {
    sys->set_bernstein2( data.begin(), e );
  };
  
  void add( const mpoldse::monomial<C>& data, int e ) 
  {
    sys->set_monomial2( data.begin(), e );
  };
  
  template<class MPDI>
  void solve( std::vector<C> & R, int& ns, int& nv, MPDI first, MPDI last )
  {
    R.resize(0);
    typedef SBDSLV< C, METHOD > MTH;
    typename sbdslv::select_mth<system_t,METHOD>::result_t __mth;
    int e = 0;
    for ( MPDI eptr = first; eptr != last; add( eptr->rep(), e++  ), eptr++ );
    //std::vector<C> tmp;
    sbdslv::system_ctrl< std::vector< C > >  intf(R);
    int nvars = sys->nvr();
    __mth.launch(sys,intf,0);

    int nsols = R.size() / (2*nvars);
    int nsols2 = sbdslv::clean_result( &R[0], nvars, nsols,(C)1e-3 );
    R.resize(nsols2) ;
    ns = nsols2;
    nv = nvars;
  };
  
  ~SbdSolver() { delete sys; };
};


//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif //SYNAPS_SOLVE_SBDSLV_H
