/* SISL - SINTEF Spline Library version 4.4.                              */
/* Definition and interrogation of NURBS curves and surface.              */
/* Copyright (C) 1978-2005, SINTEF ICT, Applied Mathematics, Norway.      */

/* This program is free software; you can redistribute it and/or          */
/* modify it under the terms of the GNU General Public License            */
/* as published by the Free Software Foundation version 2 of the License. */

/* This program is distributed in the hope that it will be useful,        */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/* GNU General Public License for more details.                           */

/* You should have received a copy of the GNU General Public License      */
/* along with this program; if not, write to the Free Software            */
/* Foundation, Inc.,                                                      */
/* 59 Temple Place - Suite 330,                                           */
/* Boston, MA  02111-1307, USA.                                           */

/* Contact information: e-mail: tor.dokken@sintef.no                      */
/* SINTEF ICT, Department of Applied Mathematics,                         */
/* P.O. Box 124 Blindern,                                                 */
/* 0314 Oslo, Norway.                                                     */

/* SISL commercial licenses are available for:                            */
/* - Building commercial software.                                        */
/* - Building software whose source code you wish to keep private.        */


#include "copyright.h"



#define S2543

#include "sislP.h"

#if defined(SISLNEEDPROTOTYPES)
void 
s2543(SISLSurf *surf, int ider, double derive[], double normal[], double *k1,
      double *k2, double d1[], double d2[], int *jstat)
#else
 void s2543(surf, ider, derive, normal, k1, k2, d1, d2, jstat)
      SISLSurf *surf;
      int ider;
      double derive[];
      double normal[];
      double *k1;
      double *k2;
      double d1[];
      double d2[];
      int *jstat;
#endif
{
   double denom;        
   double a, b, c;      
   double sqrt_arg;     
   double ratio;        
   double length;       
   double fundform[6];  
   double transform[4]; 
   double Su[3];        
   double Sv[3];        

   if (surf->idim == 1 || surf->idim == 3) 
   {
      
      
      
      if (surf->idim == 1)
      {
	 Su[0] = 1.;
	 Su[1] = 0.;
	 Su[2] = derive[1];
	 Sv[0] = 0.;
	 Sv[1] = 1.;
	 Sv[2] = derive[2];
      }
      else
      {
	 Su[0] = derive[3];
	 Su[1] = derive[4];
	 Su[2] = derive[5];
	 Sv[0] = derive[6];
	 Sv[1] = derive[7];
	 Sv[2] = derive[8];
      }
      
      
      
      s2513(surf, ider, 2, 1, derive, normal, fundform, jstat);
      if (*jstat < 0) goto error;
      
      
      
      denom = fundform[0]*fundform[2] - fundform[1]*fundform[1];
      
      transform[0] = (fundform[1]*fundform[4] - fundform[2]*fundform[3])/
	 denom;
      transform[1] = (fundform[1]*fundform[5] - fundform[2]*fundform[4])/
	 denom;
      transform[2] = (fundform[1]*fundform[3] - fundform[0]*fundform[4])/
	 denom;
      transform[3] = (fundform[1]*fundform[4] - fundform[0]*fundform[5])/
	 denom;
      
      
      
      a = 1.;
      b = transform[0] + transform[3];
      c = transform[0]*transform[3] - transform[1]*transform[2];
      
      sqrt_arg = b*b - 4.*a*c;
      if (sqrt_arg < REL_PAR_RES)
	 goto war100;
      
      *k1 = (- b + sqrt(sqrt_arg))/(2.*a);
      *k2 = (- b - sqrt(sqrt_arg))/(2.*a);
      
      
      
      
      
      if (fabs(transform[0] + *k1) < REL_PAR_RES && 
	  fabs(transform[1]) < REL_PAR_RES)
      {
	 
	 
	 
	 length = 1./sqrt(Su[0]*Su[0] + Su[1]*Su[1] + Su[2]*Su[2]);
	 
	 d1[0] = length;
	 d1[1] = 0.;
      }
      else if (fabs(transform[3] + *k1) < REL_PAR_RES && 
	       fabs(transform[2]) < REL_PAR_RES)
      {
	 
	 
	 
	 length = 1./sqrt(Sv[0]*Sv[0] + Sv[1]*Sv[1] + Sv[2]*Sv[2]);
	 
	 d1[0] = 0.;
	 d1[1] = length;
      }
      else if (fabs(transform[0] + *k1) < fabs(transform[1]))
      {
	 ratio = (transform[0] + *k1)/transform[1];
	 length = 1./sqrt((Su[0] - ratio*Sv[0])*(Su[0] - ratio*Sv[0]) +
			  (Su[1] - ratio*Sv[1])*(Su[1] - ratio*Sv[1]) +
			  (Su[2] - ratio*Sv[2])*(Su[2] - ratio*Sv[2]));
	 
	 d1[0] = length;
	 d1[1] = -ratio*length;
      }
      else
      {
	 ratio = transform[1]/(transform[0] + *k1);
	 length = 1./sqrt((Sv[0] - ratio*Su[0])*(Sv[0] - ratio*Su[0]) +
			  (Sv[1] - ratio*Su[1])*(Sv[1] - ratio*Su[1]) +
			  (Sv[2] - ratio*Su[2])*(Sv[2] - ratio*Su[2]));
	 
	 d1[0] = -ratio*length;
	 d1[1] = length;
      }
      
      
      
      if (fabs(transform[0] + *k2) < REL_PAR_RES && 
	  fabs(transform[1]) < REL_PAR_RES)
      {
	 
	 
	 
	 length = 1./sqrt(Su[0]*Su[0] + Su[1]*Su[1] + Su[2]*Su[2]);
	 
	 d2[0] = length;
	 d2[1] = 0.;
      }
      else if (fabs(transform[3] + *k2) < REL_PAR_RES && 
	       fabs(transform[2]) < REL_PAR_RES)
      {
	 
	 
	 
	 length = 1./sqrt(Sv[0]*Sv[0] + Sv[1]*Sv[1] + Sv[2]*Sv[2]);
	 
	 d2[0] = 0.;
	 d2[1] = length;	 
      }
      else if (fabs(transform[0] + *k2) < fabs(transform[1]))
      {
	 ratio = (transform[0] + *k2)/transform[1];
	 length = 1./sqrt((Su[0] - ratio*Sv[0])*(Su[0] - ratio*Sv[0]) +
			  (Su[1] - ratio*Sv[1])*(Su[1] - ratio*Sv[1]) +
			  (Su[2] - ratio*Sv[2])*(Su[2] - ratio*Sv[2]));
	 
	 d2[0] = length;
	 d2[1] = -ratio*length;
      }
      else
      {
	 ratio = transform[1]/(transform[0] + *k2);
	 length = 1./sqrt((Sv[0] - ratio*Su[0])*(Sv[0] - ratio*Su[0]) +
			  (Sv[1] - ratio*Su[1])*(Sv[1] - ratio*Su[1]) +
			  (Sv[2] - ratio*Su[2])*(Sv[2] - ratio*Su[2]));
	 
	 d2[0] = -ratio*length;
	 d2[1] = length;
      }
      
   }
   else if (surf->idim == 2) 
   {
      
      
      *k1 = 0.0;
      *k2 = 0.0;
      d1[0] = 1.0;
      d1[1] = 0.0;
      d2[0] = 0.0;
      d2[1] = 1.0;
   }
   else 
   {
      goto err105;
   }

   
   
   
   *jstat = 0;
   goto out;
   
   
   
   war100:
      if (fabs(sqrt_arg) < REL_PAR_RES)
      {
	 *k1 = - b/(2.*a);
	 *k2 = *k1;
      }
      else
      {
	 *k1 = 0.0;
	 *k2 = 0.0;
      }
   d1[0] = 1.0;
   d1[1] = 0.0;
   d2[0] = 0.0;
   d2[1] = 1.0;
   goto out;
   
   
   err105:
      *jstat = -105;
   s6err("s2543",*jstat,0);
   goto out;
   
   error:
      s6err("s2543",*jstat,0);
   goto out;
   
   out:
      
      return;

}

