#ifndef axel_mesh_BEZIER_H
#define axel_mesh_BEZIER_H

#include <synaps/init.h>
#include <assert.h>

__BEGIN_NAMESPACE_SYNAPS

namespace BEZIER {
  
  // One-variable derivative
  template<class T>
  void  diff(T* b, int degree )
  {
    // This function computes the derivative of the polynomial
    // represented in a Bernstein basis by the coefficients of b.
    // This computation is done in place.
    // T is a one-dimensional container of some sort, providing the
    // operator []
    // We assume that b[0] holds the coefficient of B^degree_0
    // Likewise, b[0] will hold the coefficient of B^{degree-1}_0
    // b[degree] is set to 0.
    
    // NOTE: degree is the last position in the container, not its size

    for (int i = 0;i < degree;i++)
      b[i] = degree*(b[i+1] - b[i]);
   
    b[degree] = 0;
  }


  // Two-variables derivative
  template<class T>
  void diff(T & b,unsigned variable,
	    int degree0,int degree1)
  {
    // This function computes the partial derivative of the polynomial
    // represented by b in a Bernstein basis. The variable must be 0 or 1
    // since we use a two-dimensional container. We also use the degree
    // of the polynomial in the variable we are differentiting with 
    // respect to.
    // T is a two-dimensional container, which has the operator (,)
    // We asume that b(i,j) holds the coefficient of
    // B^degree0_i(x0)*B^degree1_j(x1)
    // We use the analogous representation for the output
    // The entries which dissappear because of the reduction of the
    // degree are set to 0.
    // The conputation is done in place

    switch (variable) {
    case 0:
      // We work with the first variable
      for (int j = 0;j <= degree1;j++) { // For each j
	for (int i = 0;i < degree0;i++)  // We process the i's
	  b(i,j) = degree0*(b(i+1,j) - b(i,j));
	b(degree0,j) = 0;
      }
      break;
    case 1:
      // We work with the second variable
      for (int i = 0; i <= degree0;i++) { // For each i
	for (int j = 0;j < degree1;j++) // We process the j's
	  b(i,j) = degree1*(b(i,j+1) - b(i,j));
	b(i,degree1) = 0;
      }
      break;
    default:
      std::cerr << "Wrong variable number: " << variable 
		<<  ". should be 0 or 1" << std::endl;
      assert(variable == 0 || variable == 1);
    }
    
  }
  
  // Three-variables derivative
  // This should be more generic...  
  template<class T>
  void diff(T & b,unsigned variable,
	    int degree0,int degree1,int degree2)
  {
    // This function computes the partial derivative of the polynomial
    // represented by b in a Bernstein basis. The variable must be 0,1 or 2
    // since we use a three-dimensional container. We also use the degree
    // of the polynomial in the variable we are differentiting with 
    // respect to.
    // T is a three-dimensional container, which has the operator (,,)
    // We asume that b(i,j,k) holds the coefficient of
    // B^degree1_i(x1)*B^degree0_j(x0)*B^degree2_k(x2)
    // We use the analogous representation for the output
    // The entries which dissappear because of the reduction of the
    // degree are set to 0.

    T newb;

    switch (variable) {
    case 0: // Variable x0 is represented by the j's in the tensor
      newb.resize(degree1 + 1,degree0,degree2 + 1);

      for (int i = 0;i <= degree1;i++)
	for (int k = 0;k <= degree2;k++)
	  for (int j = 0;j < degree0;j++)
	    newb(i,j,k) = degree0 * (b(i,j+1,k) - b(i,j,k));

      break;
    case 1: // Variable x1 is represented by the i's in the tensor
      newb.resize(degree1,degree0 + 1,degree2 + 1);

      for (int j = 0;j <= degree0;j++)
	for (int k = 0;k <= degree2;k++) 
	  for (int i = 0;i < degree1;i++) 
	    newb(i,j,k) = degree1 * (b(i+1,j,k) - b(i,j,k));

      break;
    case 2: // Variable x2 is represented by the k's in the tensor
      newb.resize(degree1 + 1,degree0 + 1,degree2);
      for (int i = 0;i <= degree1;i++) // For each i
	for (int j = 0;j <= degree0;j++) // For each j
	  for (int k = 0;k < degree2;k++) // We process the k's
	    newb(i,j,k) = degree2*(b(i,j,k+1) - b(i,j,k));
      break;
    default:
      std::cerr << "Wrong variable number " << variable 
		<<  ". should be 0, 1 or 2" << std::endl;
      assert(variable == 0 || variable == 1 || variable == 2 );
    }

    // We resize the original "matrix" and copy the new entries
    b.resize(newb.nbrow(),newb.nbcol(),newb.nbhgt());
    
    for (int i = 0;i < b.nbrow();i++)
      for (int j = 0;j < b.nbcol();j++)
	for (int k = 0;k < b.nbhgt();k++) {
	  b(i,j,k) = newb(i,j,k);
	}
  }
} // namespace BEZIER

__END_NAMESPACE_SYNAPS

#endif
