/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): L. Deschamps                                          *
               G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/

#ifndef axel_mesh_bezier3d_H
#define axel_mesh_bezier3d_H

/*********************************************************************/

#include <synaps/init.h>
#include <synaps/arithm/traits.h>
#include <synaps/arithm/functions.h>
#include <synaps/linalg/rep3d.h>
#include <synaps/linalg/rep1d.h>
#include <synaps/linalg/MethodName.h>
#include <synaps/linalg/MATRIX3D.h>

#include <vector>

/*********************************************************************/

__BEGIN_NAMESPACE_SYNAPS

/** Structure of the tensor of control, encapsulating the necessary 
    predicats on the variations of signs of this last. **/
namespace bezier 
{

  template< class MP = MPol<double>,
            class REP3D = linalg::rep3d< typename MP::coeff_t > >
  struct rep3d : REP3D {

    typedef MP                           polynomial_t;
    typedef typename MP::coeff_t         coeff_t;
    typedef rep3d< MP, REP3D >           self_t;
    typedef typename REP3D::size_type    size_type;
    typedef typename REP3D::rep1d        rep1d;

    /** Default constructor. **/
    rep3d< MP,REP3D >(): REP3D() {}
    
    /** Regular constructor. **/
    rep3d< MP,REP3D >( unsigned n, unsigned m, unsigned o ):
      REP3D( n, m, o ) {}

    /** Copy constructor. **/
    rep3d< MP,REP3D >( const rep3d & r ):
      REP3D( r ){}
    
    rep3d< MP,REP3D >( size_type l, coeff_t* t):
      REP3D( l, t ){}
    
    self_t & operator=( const self_t & r ){
      this->REP3D::operator=( r );
      return ( *this );
    }


  };
 
  /** Function returning -1 si x < 0 
                          0 si x = 0
                          1 si x > 0. **/
  template< class C > inline int sign( const C& x ) {
    return (x<0?-1:x>0?1:0);
  }


  /** Convert a polynomial expressed in the monomial basis into its
      representation into the Bernstein basis. **/
  template< class MP, class REP3D >
  void MonomialBasisToBezierBasis( rep3d< MP,REP3D > & coeff,
				   const MP & p ) {
    int d0 = degree( p, 0 ) + 1,
        d1 = degree( p, 1 ) + 1,
        d2 = degree( p, 2 ) + 1 ;
    
    // Coefficient du polynome multivariable. 
    rep3d< MP,REP3D > mPolCoeff;

    mPolCoeff.resize( d1, d0, d2 );
    coeff.resize( d1, d0, d2 );

    TENSOR3D::init( coeff, 0 );
    TENSOR3D::init( mPolCoeff, 0 );
        
    for( typename MP::const_iterator it = p.begin();
	 it != p.end() ; it++ ){
      mPolCoeff( (*it)[1], (*it)[0], (*it)[2] ) = it->coeff();
    }
    
    // Direction suivant les lignes.
    for ( int k = 0 ; k < d2 ; ++k)   
      for ( int i = 0; i < d1 ; ++i) {
	
	typename REP3D::rep1d bezierCoeff;
	typename REP3D::rep1d polCoeff;
	
	bezierCoeff.resize( d0 );
	polCoeff.resize( d0 );
	
	for ( int j = 0 ; j < d0 ; ++j )
	  polCoeff[j] = mPolCoeff( i, j, k );
	
	monomialToBezier( bezierCoeff, polCoeff );
	
	for ( int j = 0 ; j < d0 ; ++j )
	  coeff( i, j, k ) = bezierCoeff[j];
	
      }
    
    // Direction suivant les colonnes.
    for ( int k = 0 ; k < d2 ; ++k )
      for (int j = 0 ; j < d0 ; ++j ) {
	
	typename REP3D::rep1d bezierCoeff;
	typename REP3D::rep1d polCoeff;
	
	bezierCoeff.resize( d1 );
	polCoeff.resize( d1 );
	
	for ( int i = 0 ; i < d1 ; ++i )
	  polCoeff[i] = coeff( i, j, k );
	
	monomialToBezier( bezierCoeff, polCoeff );
	
	for ( int i = 0 ; i < d1 ; ++i )
	  coeff( i, j, k ) = bezierCoeff[i]; 
	
      }

    // Direction suivant la hauteur.
    for ( int j = 0 ; j < d0 ; ++j )    
      for (int i = 0 ; i < d1 ; ++i ) {
	
	typename REP3D::rep1d bezierCoeff;
	typename REP3D::rep1d polCoeff;
	
	bezierCoeff.resize( d2 );
	polCoeff.resize( d2 );
    
	for ( int k = 0 ; k < d2 ; ++k )
	  polCoeff[k] = coeff( i, j, k );
	
	monomialToBezier( bezierCoeff, polCoeff );
	
	for (int k = 0 ; k < d2 ; ++k )
	  coeff( i, j, k ) = bezierCoeff[k];
      }
  }


  /** Conversion of the monomial basis towards the bernstein basis. **/
  template< class rep1d >
  void monomialToBezier( rep1d &bz,
			 const rep1d &polynomCoeff ) {

    typedef typename rep1d::value_type coeff_t;

    VECTOR::init( bz, polynomCoeff[0] );
    
    coeff_t size = polynomCoeff.size();
    coeff_t denominator = 1, numerator = 1;
    
    for( unsigned j = 1; j < size; j++ ) {
      
      denominator *= ( size - j );
      denominator /= j;
      numerator = 1;
      bz[j] += polynomCoeff[j] / denominator;
      
      for( unsigned i = j + 1 ; i < size ; i++)  {
	
	numerator *= i;
	numerator /= ( i - j );
	bz[i] += numerator / denominator * polynomCoeff[j];
	
      }
    }
    
  }
  
  /** Convert a polynomial expressed in the Bernstein basis into its
      representation into the monomial basis. 
       NOT TESTED !!. **/
  template< class MP, class REP3D >
  void convertBezierToMonomial( const rep3d< MP, REP3D > &bezierCoeff,
				rep3d< MP, REP3D > &monomialCoeff) {

    typedef typename MP::coeff_t coeff_t;
    
    int nbrow = bezierCoeff.nbrow();
    int nbcol = bezierCoeff.nbcol();
    int nbhgt = bezierCoeff.nbhgt();

    monomialCoeff.resize( nbrow, nbcol, nbhgt );
    TENSOR3D::init( monomialCoeff, 0 );
    
    // Direction suivant les lignes.
    for ( int k = 0 ; k < nbhgt ; k++ )   
      for ( int i = 0 ; i < nbrow ; i++ ) {
	
	typename REP3D::rep1d bzCoeff;
	typename REP3D::rep1d monomCoeff;
	
	bzCoeff.resize( nbcol );
	monomCoeff.resize( nbcol );
	
	for ( int j = 0 ; j < nbcol ; j++ )
	  bzCoeff[j] = bezierCoeff( i, j, k );
	
	bezierToMonomial( monomCoeff, bzCoeff );
	
	for (int j = 0 ; j < nbcol  ; j++)
	  monomialCoeff( i, j, k ) = monomCoeff[j];
	
      }
    
    // Direction suivant les colonnes.
    for (int k = 0 ; k < nbhgt ; k++)    
      for (int j = 0 ; j < nbcol ; j++) {
	
	typename REP3D::rep1d bzCoeff;
	typename REP3D::rep1d monomCoeff;
	
	bzCoeff.resize( nbrow );
	monomCoeff.resize( nbrow );
	
	for (int i=0 ; i < nbrow ; i++)
	  bzCoeff[i] = bezierCoeff(i,j,k);
	
	bezierToMonomial( monomCoeff, bzCoeff);
	
	for (int i=0 ; i < nbrow ; i++)
	  monomialCoeff(i,j,k) = monomCoeff[i]; 
	
      }

    // Direction suivant la hauteur.
    for (int j = 0 ; j < nbcol ; j++)    
      for (int i = 0 ; i < nbrow ; i++) {
	
	typename REP3D::rep1d bzCoeff;
	typename REP3D::rep1d monomCoeff;
	
	bzCoeff.resize( nbhgt );
	monomCoeff.resize( nbhgt );
    
	for (int k = 0 ; k < nbhgt ; k++)
	  bzCoeff[k] = bezierCoeff(i,j,k);
	
	bezierToMonomial( monomCoeff, bzCoeff);
	
	for (int k = 0 ; k < nbhgt ; k++)
	  monomialCoeff(i,j,k) = monomCoeff[k]; 
	
      }   
  }


  /** Compute the coefficients of the polynomial <t>A</t> associated
      to the Bernstein basis on $[0,1]\times [0,1]\times [0,1]$,
      in the Bernstein basis corresponding to the boxes
      $[0,1/2]\times [0,1]\times [0,1]$ and 
      $[1/2,1]\times [0,1]\times [0,1]$. **/

  template< class MP, class REP3D >
  void deCasteljau( rep3d< MP,REP3D > & left, rep3d< MP,REP3D > & right, ByRow M , double t=0.5) {

    typedef typename MP::coeff_t coeff_t;

    int col=left.nbcol(),row=left.nbrow(),hgt=left.nbhgt();
    assert( col > 1 );
    coeff_t s(t),is(1-t);
    for ( int h = 0 ; h < hgt ; ++h )
      for ( unsigned i = 0 ; i < row ; ++i )
	{
	  for ( int j = 1 ; j < col ; ++j )
	    {
	      left( i, j-1, h ) = right( i, 0, h ); 
	      for ( int k = 0 ; k < col - j ; ++k ) 
		{
		  right( i, k, h ) = right( i,k,h )*s + right( i,k+1,h )*is;
		}
	    }
	  left( i, col-1, h ) = right( i, 0, h );
      }
  }  


  /** Compute the coefficients of the polynomial <t>A</t> associated
      to the Bernstein basis on $[0,1]\times [0,1]\times [0,1]$,
      in the Bernstein basis corresponding to the boxes
      $[0,1]\times [0,1/2]\times [0,1]$ and 
      $[0,1]\times [1/2,1]\times [0,1]$. **/

  template< class MP, class REP3D >
  void deCasteljau( rep3d< MP,REP3D > & left, rep3d< MP,REP3D > & right, ByCol M ,double t=0.5 ) {

    typedef typename MP::coeff_t coeff_t;

    int col=left.nbcol(),row=left.nbrow(),hgt=left.nbhgt();
    assert( row > 1 );
    coeff_t s(t),is(1-t);
    for ( int h = 0 ; h < hgt ; ++h )
      for ( unsigned i = 0 ; i < col ; ++i )
	{
	  for ( int j = 1 ; j < row ; ++j )
	    {
	      left( j-1, i, h ) = right( 0, i, h ); 
	      for ( int k = 0 ; k < row - j ; ++k ) 
		{
		  right( k, i, h ) =  right( k,i,h )*s + right( k+1,i,h )*is;
		}
	    }
	  left( row-1, i, h ) = right( 0, i, h );
      }
  }  


   /** Compute the coefficients of the polynomial <t>A</t> associated
      to the Bernstein basis on $[0,1]\times [0,1]\times [0,1]$,
      in the Bernstein basis corresponding to the boxes
      $[0,1]\times [0,1]\times [0,1/2]$ and 
      $[0,1]\times [0,1]\times [1/2,1]$. **/

  template< class MP, class REP3D >
  void deCasteljau( rep3d< MP,REP3D > & left, rep3d< MP,REP3D > & right, ByHgt M, double  t= 0.5) {

    typedef typename MP::coeff_t coeff_t;

    int col=left.nbcol(),row=left.nbrow(),hgt=left.nbhgt();
    assert( hgt > 1 );
    coeff_t s(t),is(1-t);
    for ( int h = 0 ; h < row ; ++h )
      for ( unsigned i = 0 ; i < col ; ++i )
	{
	  for ( int j = 1 ; j < hgt ; ++j )
	    {
	      left( h, i, j-1 ) = right( h, i, 0 ); 
	      for ( int k = 0 ; k < hgt - j ; ++k ) 
		{
		  right( h, i, k ) = right( h,i,k )*s + right( h,i,k+1 )*is;
		}
	    }
	  left( h, i, hgt-1 ) = right( h, i, 0 );
      }
  }  


  /** Calculate the derivee D of M compared to the ieme variable
      of the polynomial.
      NOT TESTED!!! **/
  template< class MP, class REP3D >
  void diff( rep3d< MP,REP3D > & M, rep3d< MP,REP3D > & D, int i) {
    
    int n = 1;
    int h, j, k;
    
    // Derivee par rapport a y.
    if (i == 1) {
      D.resize( M.nbrow() - 1, M.nbcol(), M.nbhgt() );
      for ( k=0 ; k < D.nbhgt() ; k++)
	for ( j=0 ; j < D.nbcol() ; j++) {
	  for ( h=0 ; h < D.nbrow() ; h++) {
	    D(h,j,k) = n * M(h+1,j,k) - M(h,j,k);
	    n++;
	  }
	  n = 1;
	}
      return;
    }
    
    // Derivee par rapport a x.
    if (i==0) {
      D.resize(M.nbrow(), M.nbcol()-1, M.nbhgt()) ;
      for ( k=0 ; k < D.nbhgt() ; k++)
	for ( h=0 ; h < D.nbrow() ; h++) {
	  for ( j=0 ; j < D.nbcol() ; j++) {
	    D(h,j,k) = n * M(h,j+1,k) - M(h,j,k);
	    n++;
	  }
	  n = 1;
	}
      return;
    }

    // Derivee par rapport a z.
    if (i==2) {
      D.resize(M.nbrow(), M.nbcol(), M.nbhgt()-1);
      for ( j=0 ; j < D.nbcol() ; j++) 
	for ( h=0 ; h < D.nbrow() ; h++) {
	  for ( k=0 ; k < D.nbhgt() ; k++) {
	    D(h,j,k) = n * M(h,j,k+1) - M(h,j,k);
	    n++;
	  }
	  n = 1;
	}
      return;
    }
    
  }


  /** Return true if there is at least a change of sign in all 
      the matrix-3d of the coefficients. 
      Return false if not. **/
  template < class MP, class REP3D >
  bool changeOfSign(const rep3d< MP,REP3D > &bezier) {
    
    int nbrow = bezier.nbrow(), nbcol = bezier.nbcol(), nbhgt = bezier.nbhgt();
    int size = nbrow * nbcol * nbhgt - 1;
    int i;
    
    for ( i = 0 ; i < size ; ++i )
      if (sign( bezier[ i ] ) * sign( bezier[ i+1 ] ) < 0 ) 
	      return true;
      
    return false;
 }


  /** Allows to know if the curve is coplanar at one of the face of
      the current cube. **/
  template < class MP, class REP3D > 
  bool isCoplanar(const rep3d< MP,REP3D > &bezier ) {
    
    int nbrow = bezier.nbrow();
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();
    int i, j, k;

    bool isCoplanarToX = true;
    bool isCoplanarToY = true;
    bool isCoplanarToZ = true;

    // Pour le plan xz = 0
    i = 0;
    for ( j = 0 ; j < nbcol ; j++ )
      for ( k = 0 ; k < nbhgt ; k++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToY = false;
	    
    // Pour le plan yz = 0
    j = 0;
    for ( i = 0 ; i < nbrow ; i++ )
      for ( k = 0 ; k < nbhgt ; k++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToX = false;
	    
    // Pour le plan xy = 0
    k = 0;
    for ( j = 0 ; j < nbcol ; j++ )
      for ( i = 0 ; i < nbrow ; i++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToZ = false;

     // Pour le plan xz = 1
    i = nbrow - 1;
    for ( j = 0 ; j < nbcol ; j++ )
      for ( k = 0 ; k < nbhgt ; k++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToY = false;
	    
    // Pour le plan yz = 1
    j = nbcol - 1;
    for ( i = 0 ; i < nbrow ; i++ )
      for ( k = 0 ; k < nbhgt ; k++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToX = false;
	    
    // Pour le plan xy = 1
    k = nbhgt - 1;
    for ( j = 0 ; j < nbcol ; j++ )
      for ( i = 0 ; i < nbrow ; i++ )
	if ( bezier( i, j, k ) != 0 ) 
	  isCoplanarToZ = false;
 
    return ( isCoplanarToX || isCoplanarToY || isCoplanarToZ );

  }
  
  /** Numbers of variation of sign in the edge number 1. **/
  template < class MP, class REP3D > 
  int varInEdge1(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int var = 0;
    
    for ( int i = 0 ; i < nbcol-1 ; i++ ) {
      if ( bezier( 0, i, 0 ) != 0 ) {
	if ( sign( bezier( 0, i, 0 ) ) * 
	     sign( bezier( 0, i+1, 0 ) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 2. **/
  template < class MP, class REP3D > 
  int varInEdge2(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbcol-1 ; i++ ) {
      if ( bezier( 0, i, nbhgt-1 ) != 0 ) {
	if ( sign( bezier(0, i, nbhgt-1) ) * 
	     sign( bezier(0, i+1, nbhgt-1) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 3. **/
  template < class MP, class REP3D > 
  int varInEdge3(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int var = 0;
    
    for ( int i = 0 ; i < nbcol-1 ; i++ ) { 
      if ( bezier( nbrow-1, i, 0 ) != 0 ) {
	if ( sign( bezier(nbrow-1, i, 0) ) *
	     sign( bezier(nbrow-1, i+1, 0) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 4. **/
  template < class MP, class REP3D > 
  int varInEdge4(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbcol-1 ; i++ ){
      if ( bezier(nbrow-1, i, nbhgt-1) != 0 ) {
	if ( sign( bezier(nbrow-1, i, nbhgt-1) ) * 
	     sign( bezier(nbrow-1, i+1, nbhgt-1) ) < 0 )
	  var++;
      }
    }
    return var;
  }

  /** Numbers of variation of sign in the edge number 5. **/
  template < class MP, class REP3D > 
  int varInEdge5(const rep3d< MP,REP3D > &bezier ) {
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbhgt-1 ; i++ ) {
      if ( bezier(0, 0, i) != 0 ) {
	if ( sign( bezier(0, 0, i) ) * 
	     sign( bezier(0, 0, i+1) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 6. **/
  template < class MP, class REP3D > 
  int varInEdge6(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();
    int var = 0;;
    
    for ( int i = 0 ; i < nbhgt-1 ; i++ ) {
      if ( bezier( 0, nbcol-1, i ) != 0 ) {
	if ( sign( bezier( 0, nbcol-1, i ) ) *
	     sign( bezier( 0, nbcol-1, i+1 ) ) < 0 )
	  var++;
      }
    }
    return var;
  }

  /** Numbers of variation of sign in the edge number 7. **/
  template < class MP, class REP3D > 
  int varInEdge7(const rep3d< MP,REP3D > &bezier ) {
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbhgt-1 ; i++ ) {
      if ( bezier( nbrow-1, 0, i ) != 0 ) {
	if ( sign( bezier( nbrow-1, 0, i ) ) *
	     sign( bezier( nbrow-1, 0, i+1 ) ) < 0 )
	  var++;
      }
    }
    return var;	
  }
  
  /** Numbers of variation of sign in the edge number 8. **/
  template < class MP, class REP3D > 
  int varInEdge8(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbhgt-1 ; i++ ) {
      if ( bezier(nbrow-1, nbcol-1, i) != 0 ) {
	if ( sign( bezier(nbrow-1, nbcol-1, i ) ) * 
	     sign( bezier(nbrow-1, nbcol-1, i+1 ) ) < 0 )
	  var++;
      }
    }
    return var;	
  }

  /** Numbers of variation of sign in the edge number 9. **/
  template < class MP, class REP3D > 
  int varInEdge9(const rep3d< MP,REP3D > &bezier ) {
    int nbrow = bezier.nbrow();
    int var = 0;
    
    for ( int i = 0 ; i < nbrow-1 ; i++ ) {
      if ( bezier( i, 0, 0 ) != 0 ) {
	if ( sign( bezier( i, 0, 0 ) ) * 
	     sign( bezier( i+1, 0, 0 ) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 10. **/
  template < class MP, class REP3D > 
  int varInEdge10(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int var = 0;
    
    for ( int i = 0 ; i < nbrow-1 ; i++ ) {
      if ( bezier( i, nbcol-1, 0 ) != 0 ) {
	if ( sign( bezier( i, nbcol-1, 0 ) ) * 
	     sign( bezier( i+1, nbcol-1, 0 ) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 11. **/
  template < class MP, class REP3D > 
  int varInEdge11(const rep3d< MP,REP3D > &bezier ) {
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbrow-1 ; i++ ) { 
      if ( bezier( i, 0, nbhgt-1 ) != 0 ) { 
	if ( sign( bezier( i, 0, nbhgt-1 ) ) * 
	     sign( bezier( i+1, 0, nbhgt-1 ) ) < 0 )
	  var++;
      }
    }
    return var;
  }
  
  /** Numbers of variation of sign in the edge number 12. **/
  template < class MP, class REP3D > 
  int varInEdge12(const rep3d< MP,REP3D > &bezier ) {
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int var = 0;
    
    for ( int i = 0 ; i < nbrow-1 ; i++ ) {
      if ( bezier( i, nbcol-1, nbhgt-1 ) != 0 ) {
	if ( sign( bezier( i, nbcol-1, nbhgt-1 ) ) *
	     sign( bezier( i+1, nbcol-1, nbhgt-1 ) ) < 0 ) 
	  var++;
      }
    }
    return var;
    
  }  

  /** Numbers maximum variation according to the columns on face 1. **/
  template< class MP, class REP3D >
  int varMaxByColInFace_1(const rep3d< MP,REP3D > &bezier){
    
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbcol ; i++ ) {
      
      for ( j = 0 ; j < nbrow - 1 ; j++ )

	if ( bezier( j, i, 0 ) != 0 ) {
	  if ( sign( bezier( j, i, 0 ) ) * 
	       sign( bezier( j + 1, i, 0 ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }
  
   /** Numbers maximum variation according to the rows on face 1. **/
  template< class MP, class REP3D >
  int varMaxByRowInFace_1(const rep3d< MP,REP3D > &bezier){
    
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbrow ; i++ ) {

      for ( j = 0 ; j < nbcol - 1 ; j++ )
	
	if (  bezier( i, j, 0 ) != 0 ) {
	  if ( sign( bezier( i, j, 0 ) ) * 
	       sign( bezier( i, j + 1, 0 ) ) < 0 )
	    nbVar++;
	}
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0; 
    }
    
    return nbMaxVar;
  }

  
  /** Numbers maximum variation according to the columns on face 3. **/
  template< class MP, class REP3D >
  int varMaxByColInFace_3(const rep3d< MP,REP3D > &bezier ){
   
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbcol ; i++ ) {

      for ( j = 0 ; j < nbrow - 1 ; j++ )
 	
	if ( bezier( j, i, nbhgt - 1 ) != 0 ) {
	  if ( sign( bezier( j, i, nbhgt - 1 ) ) *
	       sign( bezier( j + 1, i, nbhgt - 1 ) ) < 0 )
	    nbVar++;
	}
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }

    return nbMaxVar;
  }

  /** Numbers maximum variation according to the rows on face 3. **/
  template< class MP, class REP3D >
  int varMaxByRowInFace_3(const rep3d< MP,REP3D > &bezier ){
    
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbrow ; i++ ) {
      
      for ( j = 0 ; j < nbcol - 1 ; j++ )
	
	if ( bezier( i, j, nbhgt - 1 ) != 0 ) {
	  if ( sign( bezier( i, j, nbhgt - 1 ) ) * 
	       sign( bezier( i, j + 1, nbhgt - 1 ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }
  
  
  /** Numbers maximum variation according to the columns on face 2. **/
  template< class MP, class REP3D >
  int varMaxByColInFace_2(const rep3d< MP,REP3D > &bezier ){
 
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbcol ; i++ ) {
      
      for ( j = 0 ; j < nbhgt - 1 ; j++ )
	
	if ( bezier( 0, i, j ) != 0 ) {
	  if ( sign( bezier( 0, i, j ) ) *
	       sign( bezier( 0, i, j + 1 ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }

    return nbMaxVar;
  }

  
  /** Numbers maximum variation according to the depth on face 2. **/
  template< class MP, class REP3D >
  int varMaxByHgtInFace_2(const rep3d< MP,REP3D > &bezier ){
   
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbhgt ; i++ ) {

      for ( j = 0 ; j < nbcol - 1 ; j++ )
	
	if ( bezier( 0, j, i ) != 0 ) {
	  if ( sign( bezier( 0, j, i ) ) *
	       sign( bezier( 0, j + 1, i ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }
  
  /** Numbers maximum variation according to the columns on face 5. **/
  template< class MP, class REP3D >
  int varMaxByColInFace_5(const rep3d< MP,REP3D > & bezier ){

    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbcol ; i++) {

      for ( j = 0 ; j < nbhgt - 1 ; j++)

	if ( bezier( nbrow - 1, i, j ) != 0 ) {
	  if ( sign( bezier( nbrow - 1, i, j ) ) *
	       sign( bezier( nbrow - 1, i, j + 1 ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }

    return nbMaxVar;
  }

  
  /** Numbers maximum variation according to the depth on face 5. **/
  template< class MP, class REP3D >
  int varMaxByHgtInFace_5(const rep3d< MP,REP3D > &bezier ){

    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbhgt ; i++ ) {

      for ( j = 0 ; j < nbcol - 1 ; j++ )

	if ( bezier( nbrow - 1, j, i ) != 0 ) {
	  if ( sign( bezier( nbrow - 1, j, i ) ) *
	       sign( bezier( nbrow - 1, j + 1, i ) ) < 0 )
	  nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }

  /** Numbers maximum variation according to the rows on face 4. **/
  template<class MP, class REP3D>
  int varMaxByRowInFace_4(const rep3d< MP,REP3D > &bezier ){

    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbrow ; i++ ) {

      for ( j = 0 ; j < nbhgt - 1 ; j++ )

	if ( bezier( i, nbcol - 1, j ) != 0 ) {
	  if ( sign ( bezier( i, nbcol - 1, j ) ) *
	       sign ( bezier( i, nbcol - 1, j + 1 ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }

  
  /** Numbers maximum variation according to the depth on face 4. **/
  template< class MP, class REP3D >
  int varMaxByHgtInFace_4(const rep3d< MP,REP3D > &bezier ){

    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbhgt ; i++ ) {

      for ( j = 0 ; j < nbrow - 1; j++ )

	if  ( bezier( j, nbcol - 1, i ) != 0 ) {
	  if ( sign ( bezier( j, nbcol - 1, i ) ) *
	       sign ( bezier( j + 1, nbcol - 1, i ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }


  /** Numbers maximum variation according to the rows on face 6. **/
  template< class MP, class REP3D >
  int varMaxByRowInFace_6(const rep3d< MP,REP3D > &bezier ){

    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbrow ; i++ ) {

      for ( j = 0 ; j < nbhgt - 1 ; j++ )

	if ( bezier( i, 0, j ) != 0 ) {
	  if ( sign ( bezier( i, 0, j ) ) *
	       sign ( bezier( i, 0, j + 1 ) ) < 0 ) 
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }
  
  
  /** Numbers maximum variation according to the depth on face 6. **/
  template< class MP, class REP3D >
  int varMaxByHgtInFace_6(const rep3d< MP,REP3D > &bezier ){

    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int i,j;
    int nbMaxVar = 0;
    int nbVar = 0;
    
    for ( i = 0 ; i < nbhgt ; i++ ) {

      for ( j = 0 ; j < nbrow - 1 ; j++ )

	if ( bezier( j, 0, i ) != 0 ) {
	  if ( sign ( bezier( j, 0, i ) ) *
	       sign ( bezier( j + 1, 0, i ) ) < 0 )
	    nbVar++;
	}
      
      if ( nbVar > nbMaxVar )
	nbMaxVar = nbVar;
      nbVar = 0;
    }
    
    return nbMaxVar;
  }
  
  /** Return true if the number of variation according to the columns
      is a strict inferior has 2. **/
  template< class MP, class REP3D >
  bool lessThanTwoVarByRow(const rep3d< MP,REP3D > &bezier ) {
    
    int i, j, k;
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int nbVar = 0;
    int signVar = 0;
    int sl=sign(bezier(1,0,0)-bezier(0,0,0));
    for ( j = 0 ; j < nbcol ; j++) {
      nbVar = 0;
      for ( k = 0 ; k < nbhgt ; k++ ) {
	nbVar = 0;
	
	for ( i = 0 ; i < nbrow - 1; i++ ) {
	  
	  if ( sign( bezier( i, j, k ) ) *
	       sign( bezier( i + 1, j, k ) ) < 0 ) {
	    nbVar++;
	    if ( signVar == 0 ) 
	      signVar = sign( bezier( i + 1, j, k ) );
	    else 
	      if ( signVar !=  sign( bezier( i + 1, j, k ) ) )
		return false;
	  }
	  if ( nbVar > 1 ) return false;
	  if(sl*bezier(i,j,k)>sl*bezier(i+1,j,k)) return false;
	}
      }
    }
    return true;
  }

  /** Return true if the number of variation according to the rows
      is a strict inferior has 2. **/
  template< class MP, class REP3D >
  bool lessThanTwoVarByCol(const rep3d< MP,REP3D > &bezier ) {
    
    int i, j, k;
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int nbVar = 0;
    int signVar = 0;
    int sl=sign(bezier(0,1,0)-bezier(0,0,0));
    for ( i = 0 ; i < nbrow ; i++ ) {
      nbVar = 0;
      for ( k = 0 ; k < nbhgt ; k++ ) {
	nbVar = 0;
	for ( j = 0 ; j < nbcol - 1 ; j++ )  {
	  if ( sign( bezier( i, j, k ) ) * 
	       sign( bezier( i, j + 1, k ) ) < 0 ) {
	    nbVar++;
	    if ( signVar == 0 ) 
	      signVar = sign( bezier( i, j + 1, k ) );
	    else 
	      if ( signVar !=  sign( bezier( i, j + 1, k ) ) )
		return false;
	  }
	  if ( nbVar > 1 ) return false;
	  if(sl*bezier(i,j,k)>sl*bezier(i,j+1,k)) return false;
	}
      }
    }
    return true;
  }
  

  /** Return true if the number of variation according to the depth
      is a strict inferior has 2. **/
  template< class MP, class REP3D >
  bool lessThanTwoVarByHgt(const rep3d< MP,REP3D > &bezier ) {
    
    int i, j, k;
    int nbcol = bezier.nbcol();
    int nbrow = bezier.nbrow();
    int nbhgt = bezier.nbhgt();
    int nbVar = 0;
    int signVar = 0; 
    int sl = sign(bezier(0,0,1)-bezier(0,0,0));
    for ( j = 0 ; j < nbcol ; j++ ) {
      nbVar = 0;
      for ( i = 0 ; i < nbrow ; i++ ) {
	nbVar = 0;
	for ( k = 0 ; k < nbhgt - 1; k++ ) {
	  if ( sign( bezier( i, j, k ) ) * 
	       sign( bezier( i, j, k + 1 ) ) < 0 ) {
	    nbVar++;
	    if ( signVar == 0 ) 
	      signVar = sign( bezier( i, j, k + 1 ) );
	    else 
	      if ( signVar !=  sign( bezier( i, j, k + 1 ) ) )
		return false;
	  }
	  if ( nbVar > 1 ) return false;
	  if(sl*bezier(i,j,k)>sl*bezier(i,j,k+1)) return false;
	}
      }
    }
    return true;
  }

  /** Allows to find a diagonal whose ends are signs oppose.
      That allows deduct that the diagonal and surface are intersected
      in a point at least. **/
  template < class MP, class REP3D >
  int findDiagonal(const rep3d< MP,REP3D > &bezier ) {
    
    typedef typename MP::coeff_t coeff_t;

    int nbrow = bezier.nbrow();
    int nbcol = bezier.nbcol();
    int nbhgt = bezier.nbhgt();

    coeff_t Vertex_000 = bezier( 0, 0, 0 );
    coeff_t Vertex_001 = bezier( 0, 0, nbhgt-1 );
    coeff_t Vertex_010 = bezier( 0, nbcol-1, 0 );
    coeff_t Vertex_011 = bezier( 0, nbcol-1, nbhgt-1 );
    coeff_t Vertex_100 = bezier( nbrow-1, 0, 0 );
    coeff_t Vertex_101 = bezier( nbrow-1, 0, nbhgt-1 );
    coeff_t Vertex_110 = bezier( nbrow-1, nbcol-1, 0 );
    coeff_t Vertex_111 = bezier( nbrow-1, nbcol-1, nbhgt-1 );
       
    // Diagonale 1.  
    if ( ( ( Vertex_000 > 0 ) && ( Vertex_111 < 0 ) ) ||
	 ( ( Vertex_000 < 0 ) && ( Vertex_111 > 0 ) ) )
      return 1;
    
    // Diagonale 2.
    if ( ( ( Vertex_001 > 0 ) && ( Vertex_110 < 0 ) ) ||
	 ( ( Vertex_001 < 0 ) && ( Vertex_110 > 0 ) ) )
      return 2;
    
    // Diagonale 3.  
    if ( ( ( Vertex_010 > 0 ) && ( Vertex_101 < 0 ) ) ||
	 ( ( Vertex_010 < 0 ) && ( Vertex_101 > 0 ) ) )
      return 3;
  
    // Diagonale 4.
    if ( ( ( Vertex_100 > 0 ) && ( Vertex_011 < 0 ) ) ||
	 ( ( Vertex_100 < 0 ) && ( Vertex_011 > 0 ) ) )
      return 4;
    
    //    std::cerr << " Erreur : findDiagonal, pas de diagonale dont les"
    //	      << " extremites sont de signes opposes \n";
    //    TENSOR3D::print( std::cerr, bezier );
    return 0;
    
  }

} // bezier

/*********************************************************************/

__END_NAMESPACE_SYNAPS

#endif // axel_mesh_bezier3d_H
