/*********************************************************************
*      This file is part of the source code of AXEL library          *
*   Author(s): L. Deschamps                                          *
*              G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef axel_mesh_cell3d_H
#define axel_mesh_cell3d_H

/*********************************************************************/

#include <synaps/init.h>

#include <synaps/mpol.h>
#include <synaps/mpol/MPOLDST.m>
#include <synaps/arithm/sign.h>
#include <synaps/linalg/rep1d.h>
#include <synaps/linalg/rep3d.h>
#include <synaps/linalg/MethodName.h>

#include <synaps/topology/box3d.h>
#include <synaps/topology/point.h>
#include <synaps/topology/segment.h>
#include <synaps/topology/triangle.h>
#include <synaps/topology/mesh/bezier3d.h>
#include <synaps/topology/mesh/cube3d.h>
#include <synaps/topology/Octree.h>
#include <synaps/topology/mesh/scene.h>
#include <synaps/topology/mesh/subdivisor.h>

#include <assert.h>

/*********************************************************************/

__BEGIN_NAMESPACE_SYNAPS

using namespace OCTREE;

/** Structure containing a matrix-3d containing the coefficients of the
    polynomial in the base of Berstein, a cube and one limps containing 
    the coordonnees this one. **/
template < class MP = MPol<double>,
           class REP3D = linalg::rep3d< typename MP::coeff_t > >
struct cell3d {
  
  /** Type  definition. **/
  typedef MP                          polynomial_t;
  typedef typename MP::coeff_t        coeff_t;
  typedef REP3D                       rep_t;
  typedef topology::box3d< coeff_t >    Box;
  typedef bezier::rep3d< MP, REP3D >  Bezier;
  
  /** Precision on the calculation of the roots on the edges. **/
  static coeff_t epsilon;

  /** Coordinates of the box. **/
  Box box;

  /** Polynom's coefficients in Bernstein basis. **/
  Bezier bezier;

  /** Cube. **/
  topology::cube3d< MP,REP3D > cube;

  /** Default Constructor. **/
  cell3d() : box(), bezier() {}

  /** Copy constructor. **/
  cell3d( const cell3d< MP, REP3D > & bz3d ) :
    box( bz3d.box ),
    bezier( bz3d.bezier ) {}
  
  /** Regular constructor. **/
  cell3d( const Box &bo, const Bezier &be ) : 
    box( bo ),
    bezier( be ) {}
    
  /** Regular constructor. **/
//   cell3d( const char * polynome,
// 	    coeff_t x0, coeff_t x1, coeff_t y0, coeff_t y1, coeff_t z0, coeff_t z1);


  cell3d(const MP &, coeff_t x0, coeff_t x1, coeff_t y0,
	 coeff_t y1, coeff_t z0, coeff_t z1);

  
  /** Computation of the roots on the edges. **/
  void computeRootOnEdge_1(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_2(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_3(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_4(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_5(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_6(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_7(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_8(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_9(  scene<MP,REP3D> & scn );
  void computeRootOnEdge_10( scene<MP,REP3D> & scn );
  void computeRootOnEdge_11( scene<MP,REP3D> & scn );
  void computeRootOnEdge_12( scene<MP,REP3D> & scn );


  /** Propagation of the roots to the close edges. **/
  void propagationOfEdge_1(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_2(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_3(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_4(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_5(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_6(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_7(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_8(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_9(  Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_10( Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_11( Node< cell3d< MP, REP3D > > * node );
  void propagationOfEdge_12( Node< cell3d< MP, REP3D > > * node );


  void findVertexOnDiagonal_1( int depth, topology::point< coeff_t > &point );
  void findVertexOnDiagonal_2( int depth, topology::point< coeff_t > &point );
  void findVertexOnDiagonal_3( int depth, topology::point< coeff_t > &point );
  void findVertexOnDiagonal_4( int depth, topology::point< coeff_t > &point );

  bool buildMeshInCube(scene<MP,REP3D> & scn, Node< cell3d< MP,REP3D > > * node ); 

  int computeVertexOnDiagonal( topology::point< coeff_t > &point, int depthOfNode);

  void splitDiagonal( std::vector< cell3d< MP,REP3D > >  &listbezier3d );

  void makeOnlyOneTriangle(scene<MP,REP3D> & scn);

  void setCubeCorners( void );

};

/** Regular constructor.
    polynomial is character strings describing the polynomial, the 
    interval of study is [ x0, x1 ] [ y0, y1 ] [ z0, z1 ]. 
    red, green, blue, are the intensities of the colors of the future
    mesh has represente the transparency of the mesh the last 4 parameters
    lie between 0.  and 1.  
    min and max are the two thresholds being useful has the subdivision. **/
template < class MP, class REP3D >
cell3d< MP, REP3D >::cell3d( const MP & polynom,
		      coeff_t x0, coeff_t x1, coeff_t y0,
		      coeff_t y1, coeff_t z0, coeff_t z1)
{  
  /** Carry out a change of variable on the equation of the polynomial. **/
  MP polynom2 = MPOLDST::shift( polynom, x0, y0, z0 );
  
  /** The new polynomial.  **/ 
  MP newPolynom;
  
  newPolynom = MPOLDST::scale( polynom2, coeff_t( x1-x0 ),
			  coeff_t( y1-y0 ), coeff_t( z1-z0 ) );
  
  /** Recuperate coefficients in the base of Bernstein.  
      The coefficients are milked on [ 0,1 ] X [ 0,1 ] X [ 0,1 ]. **/
  MonomialBasisToBezierBasis( bezier, newPolynom );

  box.x1 = coeff_t( 1 );
  box.y1 = coeff_t( 1 );
  box.z1 = coeff_t( 1 );
  
  box.x0 = coeff_t( 0 );
  box.y0 = coeff_t( 0 );
  box.z0 = coeff_t( 0 );

  /** Initialization of the signs of the corners of the cube. **/
  setCubeCorners();  
}

/** Initialization of the precision in the calculation of the roots. **/
template < class MP, class REP3D >
typename MP::coeff_t cell3d< MP, REP3D >::epsilon = 1.e-3;

/** Insertion in the list of node. **/
template < class MP, class REP3D >
static void insert(scene<MP,REP3D> & scn,
		   Node< cell3d< MP, REP3D > > *node ) 
{
  scn.listOfNode.push_back( node );
}

/** Initialization of the sign of the corners of the cube.
    That is useful has determiner the orientation of a triangle during
    its construction. **/
template< class MP, class REP3D >
void cell3d< MP, REP3D >::setCubeCorners()
{
  int _1r = bezier.nbrow() - 1;
  int _1c = bezier.nbcol() - 1;
  int _1h = bezier.nbhgt() - 1;
  
  if ( bezier( 0,   0,   0)  > 0) cube._000 = 1; else cube._000 = -1;
  if ( bezier( 0,   0,  _1h) > 0) cube._001 = 1; else cube._001 = -1;
  if ( bezier( 0,  _1c,  0)  > 0) cube._010 = 1; else cube._010 = -1;
  if ( bezier( 0,  _1c, _1h) > 0) cube._011 = 1; else cube._011 = -1;
  if ( bezier(_1r,  0,   0)  > 0) cube._100 = 1; else cube._100 = -1;
  if ( bezier(_1r,  0,  _1h) > 0) cube._101 = 1; else cube._101 = -1;
  if ( bezier(_1r, _1c,  0)  > 0) cube._110 = 1; else cube._110 = -1;
  if ( bezier(_1r, _1c, _1h) > 0) cube._111 = 1; else cube._111 = -1;
}

//----------------------------------------------------------------------

/** Useful function for the search for top on a diagonal. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::splitDiagonal( std::vector< cell3d< MP, REP3D > >  &listbezier3d ) 
{
  bool minRow = false, minCol = false, minHgt = false;
  
  // Si c'est une courbe ou une droite ...
  if ( bezier.nbrow() <= 1 ) minRow = true;
  if ( bezier.nbcol() <= 1 ) minCol = true;
  if ( bezier.nbhgt() <= 1 ) minHgt = true;

  // Exploitation de 2 listes pour permettre un partitionnement intelligent
  std::vector< cell3d< MP, REP3D > > listA, listB;
  std::vector< cell3d< MP, REP3D > > *mustBeDivide, *divided, *swap;
  cell3d< MP, REP3D > b0, b1;
  
  mustBeDivide = &listA;
  divided = &listB;
  
  if ( !minRow || !minCol || !minHgt ) 
    mustBeDivide->push_back( *this );
 
  // Division suivant les lignes
  if ( !minRow ) {
   
    for ( int i = 0 ; i < mustBeDivide->size() ; ++i ) {
      b0 = (*mustBeDivide)[i];
      b1 = b0;
      
      double t=0.5;
      bezier::deCasteljau( b0.bezier, b1.bezier, ByRow(), t);
      b0.box.x1 =  b0.box.x0*t + b0.box.x1*(1-t);
      b1.box.x0 =  b0.box.x1;
      
      divided->push_back( b0 );
      divided->push_back( b1 );
    }

    mustBeDivide->clear();
    swap = mustBeDivide;
    mustBeDivide = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  
  // Division suivant les colonnes
  if ( !minCol ) {
    
    for ( int i = 0 ; i < mustBeDivide->size() ; ++i ) {
      b0 = (*mustBeDivide)[i];
      b1 = b0;

      double t=0.5;
      bezier::deCasteljau( b0.bezier, b1.bezier, ByCol(),t );
      b0.box.y1 = b0.box.y0*t + b0.box.y1*(1-t);
      b1.box.y0 = b0.box.y1;

      divided->push_back( b0 );
      divided->push_back( b1 );
    }

    mustBeDivide->clear();
    swap = mustBeDivide;
    mustBeDivide = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  // Division suivant la profondeur
  if ( !minHgt ) {
    
    for ( int i = 0 ; i < mustBeDivide->size() ; ++i ) {
      b0 = (*mustBeDivide)[i];
      b1 = b0;

      double t=0.5;
      bezier::deCasteljau( b0.bezier, b1.bezier, ByHgt(), t);
      b0.box.z1 = b0.box.z0*t + b0.box.z1*(1-t);
      b1.box.z0 = b0.box.z1;

      divided->push_back(b0);
      divided->push_back(b1);
    }

    mustBeDivide->clear();
    swap = mustBeDivide;
    mustBeDivide = divided;
    divided = swap;
    
    if ( !( divided->empty() ) )
      std::cerr << " Error : in split  \n";
  }

  // Si ce n'est pas en 3D => probleme avec l'octree. 
  if ( mustBeDivide->size() != 8 ) {
    std::cout << " Erreur : Taille de la liste "  << mustBeDivide->size() << std::endl;
    exit(0);
  }

  for ( int i = 0 ; i < mustBeDivide->size() ; ++i )
    listbezier3d.push_back( (*mustBeDivide)[i] );

}


//----------------------------------------------------------------------
/** Computation of the root on the edge number 1. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_1(scene<MP,REP3D>& scn) 
{
  // Calculer la racine.
  bezier::rep1d< coeff_t > controlCoeff ( bezier.nbcol() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbcol() ; ++i ) 
    controlCoeff[i] = bezier( 0, i, 0 ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps( sol.rep(), controlCoeff, box.x0, box.x1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_1, + de 1 solution sur l'arete \n";
    exit(0);
  }

  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_1 = new topology::vertex< topology::point< coeff_t > >( sol[0], box.y0, box.z0 );
  insertVertexInList( scn, cube.rootOnEdge_1 );
}


/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_1( Node< cell3d< MP, REP3D > > * node ) 
{
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_1;

  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  // Voisin par l'arete.
  node->getB_NNeighbors( listOfNeighbors );
  
  // Si il n'y a qu'un voisin.
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
    
      topology::box3d< coeff_t > *neighborBox;
      neighborBox = &listOfNeighbors[0]->GetCell().box;
      
      // Le voisin est a un niveau plus grossier ou identique que 
      // le noeud courant et ils ont une arete en commun.
      if ( ( box.y0 == neighborBox->y1 ) &&
	   ( box.x0 >= neighborBox->x0 ) && 
	   ( box.z0 == neighborBox->z1 ) &&  
	   ( box.x1 <= neighborBox->x1 ) ) {

	topology::cube3d< MP,REP3D > *B_NNeighbor;
	B_NNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	// Le voisin a-t-il deja ete calcule ou traiter ?
	if ( !( B_NNeighbor->edge_4_isProcessed )  ) {
	  // si non alors on met a jour le voisin.
	  B_NNeighbor->rootOnEdge_4 = vertex;
	  B_NNeighbor->edge_4_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllNorthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z0 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *northNeighbor;
	northNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if  ( ! northNeighbor->edge_3_isProcessed )  {
	  northNeighbor->rootOnEdge_3 = vertex;
	  northNeighbor->edge_3_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllBackNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z0 == neighborsBox->z1 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *backNeighbor;
	backNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! backNeighbor->edge_2_isProcessed )  {
	  backNeighbor->rootOnEdge_2 = vertex;
	  backNeighbor->edge_2_isProcessed = true;
	}  
      }
      listOfNeighbors.clear();
    }
}

//----------------------------------------------------------------------

/** Computation of the root on the edge number 2. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_2(scene<MP,REP3D>& scn) 
{
  // Calculer la racine.
  bezier::rep1d< coeff_t > controlCoeff ( bezier.nbcol() );
  Seq< coeff_t > sol;

  for ( int i = 0 ; i < bezier.nbcol() ; ++i ) 
    controlCoeff[i] = bezier(0, i, bezier.nbhgt() - 1 ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps( sol.rep(), controlCoeff, box.x0, box.x1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_2, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_2 = new topology::vertex< topology::point< coeff_t > >( sol[0], box.y0, box.z1 );
  insertVertexInList( scn, cube.rootOnEdge_2 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_2( Node< cell3d< MP, REP3D > > * node ) {
   
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_2;
  
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getF_NNeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *F_NNeighbor;
	F_NNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! F_NNeighbor->edge_3_isProcessed )  {
	  F_NNeighbor->rootOnEdge_3 = vertex;
	  F_NNeighbor->edge_3_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllNorthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z1 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *northNeighbor;
	northNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! northNeighbor->edge_4_isProcessed )  {
	  northNeighbor->rootOnEdge_4 = vertex;
	  northNeighbor->edge_4_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllFrontNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *frontNeighbor;
	frontNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! frontNeighbor->edge_1_isProcessed ) {
	  frontNeighbor->rootOnEdge_1 = vertex;
	  frontNeighbor->edge_1_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
}


/** Computation of the root on the edge number 3. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_3(scene<MP,REP3D>& scn) {

 // Calculer la racine.
  bezier::rep1d< coeff_t > controlCoeff ( bezier.nbcol() );
  Seq< coeff_t > sol;
    
  for ( int i = 0 ; i < bezier.nbcol() ; ++i ) 
    controlCoeff[i] = bezier( bezier.nbrow() - 1, i, 0 ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps( sol.rep(), controlCoeff, box.x0, box.x1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_3, + de 1 solution sur l'arete \n";
    exit(0);
  } 

  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_3 = new topology::vertex< topology::point< coeff_t > >( sol[0], box.y1, box.z0 );
  insertVertexInList( scn, cube.rootOnEdge_3 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_3( Node< cell3d< MP, REP3D > > * node ) {

  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_3;

  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getB_SNeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > *neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z0 == neighborsBox->z1 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *B_SNeighbor;
	B_SNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! B_SNeighbor->edge_2_isProcessed ) {
	  B_SNeighbor->rootOnEdge_2 = vertex;
	  B_SNeighbor->edge_2_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
 
  node->getAllBackNeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y1 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z0 == neighborsBox->z1 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *backNeighbor;
	backNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! backNeighbor->edge_4_isProcessed ) {
	  backNeighbor->rootOnEdge_4 = vertex;
	  backNeighbor->edge_4_isProcessed = true;
	}
      }
      listOfNeighbors.clear();  
    }
  
  node->getAllSouthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z0 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *southNeighbor;
	southNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! southNeighbor->edge_1_isProcessed ) {
	  southNeighbor->rootOnEdge_1 = vertex;
	  southNeighbor->edge_1_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
}

/** Computation of the root on the edge number 4. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_4(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbcol() );
  Seq< coeff_t > sol;
 
  for ( int i = 0 ; i < bezier.nbcol() ; ++i ) 
    controlCoeff[i] = bezier( bezier.nbrow()-1, i, bezier.nbhgt() - 1 );
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(),  controlCoeff, box.x0, box.x1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_4, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_4 = new topology::vertex< topology::point< coeff_t > >( sol[0], box.y1, box.z1 );
  insertVertexInList( scn, cube.rootOnEdge_4 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_4( Node< cell3d< MP, REP3D > > * node ) {

  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_4;

   // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getF_SNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *F_SNeighbor;
	F_SNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! F_SNeighbor->edge_1_isProcessed ) {
	  F_SNeighbor->rootOnEdge_1 = vertex;
	  F_SNeighbor->edge_1_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllFrontNeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y1 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z0 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > * frontNeighbor;
	frontNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! frontNeighbor->edge_3_isProcessed ) {
	  frontNeighbor->rootOnEdge_3 = vertex;
	  frontNeighbor->edge_3_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllSouthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 >= neighborsBox->x0 ) && 
	   ( box.z1 == neighborsBox->z1 ) &&  
	   ( box.x1 <= neighborsBox->x1 ) ) {
	
	topology::cube3d< MP,REP3D > *southNeighbor;
	southNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! southNeighbor->edge_2_isProcessed ) {
	  southNeighbor->rootOnEdge_2 = vertex;
	  southNeighbor->edge_2_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }

}
  


/** Computation of the root on the edge number 5. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_5(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbhgt() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbhgt() ; ++i )
    controlCoeff[i] = bezier( 0, 0, i ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(), controlCoeff, box.z0, box.z1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_5, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_5 = new topology::vertex< topology::point< coeff_t > >( box.x0, box.y0, sol[0] );
  insertVertexInList( scn, cube.rootOnEdge_5 );
 }

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_5( Node< cell3d< MP, REP3D > > * node ) { 
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_5;

  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getNWNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x0 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
        
	topology::cube3d< MP,REP3D > * NWNeighbor;
	NWNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! NWNeighbor->edge_8_isProcessed ){
	  NWNeighbor->rootOnEdge_8 = vertex;
	  NWNeighbor->edge_8_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllNorthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x0 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > * northNeighbor;
	northNeighbor = &listOfNeighbors[0]->GetCell().cube; 
	
	if ( ! northNeighbor->edge_7_isProcessed ){
	  northNeighbor->rootOnEdge_7 = vertex;
	  northNeighbor->edge_7_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllWestNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() ) 
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y0 ) &&
	   ( box.x0 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D >	*westNeighbor;
	westNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! westNeighbor->edge_6_isProcessed ) {
	  westNeighbor->rootOnEdge_6 = vertex;
	  westNeighbor->edge_6_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }

}



/** Computation of the root on the edge number 6. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_6(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbhgt() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbhgt() ; ++i ) 
    controlCoeff[i] = bezier( 0, bezier.nbcol()-1 , i ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(), controlCoeff, box.z0, box.z1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_6, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_6 = new topology::vertex< topology::point< coeff_t > >( box.x1, box.y0, sol[0] );
  insertVertexInList( scn, cube.rootOnEdge_6 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_6( Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_6;

  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getNENeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x1 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *NENeighbor;
	NENeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! NENeighbor->edge_7_isProcessed ){
	  NENeighbor->rootOnEdge_7 = vertex;
	  NENeighbor->edge_7_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  
  node->getAllNorthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y1 ) &&
	   ( box.x1 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *northNeighbor; 
	northNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! northNeighbor->edge_8_isProcessed ){
	  northNeighbor->rootOnEdge_8 = vertex;
	  northNeighbor->edge_8_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
 
  
  node->getAllEastNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 == neighborsBox->y0 ) &&
	   ( box.x1 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *eastNeighbor;
	eastNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! eastNeighbor->edge_5_isProcessed ){
	  eastNeighbor->rootOnEdge_5 = vertex;
	  eastNeighbor->edge_5_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
}


/** Computation of the root on the edge number 7. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_7(scene<MP,REP3D>& scn) {

  // Calculer la racine.

  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbhgt() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbhgt() ; ++i ) 
     controlCoeff[i] = bezier( bezier.nbrow()-1, 0 , i );
  
  if( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(),  controlCoeff, box.z0, box.z1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_7, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_7 = new topology::vertex< topology::point< coeff_t > >( box.x0, box.y1, sol[0] );
  insertVertexInList( scn, cube.rootOnEdge_7 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_7(Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_7;
 
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getSWNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
    
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *SWNeighbor;
	SWNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if  ( ! SWNeighbor->edge_6_isProcessed ) {
	  SWNeighbor->rootOnEdge_6 = vertex;
	  SWNeighbor->edge_6_isProcessed = true;
	}
      }
      listOfNeighbors.clear(); 
    }
  
  node->getAllSouthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x0 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *southNeighbor;
	southNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! southNeighbor->edge_5_isProcessed ) {
	  southNeighbor->rootOnEdge_5 = vertex;
	  southNeighbor->edge_5_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllWestNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y1 ) &&
	   ( box.x0 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *westNeighbor;
	westNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! westNeighbor->edge_8_isProcessed ){
	  westNeighbor->rootOnEdge_8 = vertex;
	  westNeighbor->edge_8_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
}


/** Computation of the root on the edge number 8. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_8(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbhgt() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbhgt() ; ++i )
    controlCoeff[i] = bezier( bezier.nbrow()-1, bezier.nbcol()-1 , i );

  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );

  int nbSol = bezier::isole_eps(sol.rep(), controlCoeff, box.z0, box.z1, epsilon);

  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_8, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_8 = new topology::vertex< topology::point< coeff_t > >( box.x1, box.y1, sol[0] );
  insertVertexInList( scn, cube.rootOnEdge_8 );
 }

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_8(Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_8;

  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getSENeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x1 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *SENeighbor;
	SENeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! SENeighbor->edge_5_isProcessed ){
	  SENeighbor->rootOnEdge_5 = vertex;
	  SENeighbor->edge_5_isProcessed = true;
	}
      }
      listOfNeighbors.clear(); 
    }
  
  node->getAllSouthNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y0 ) &&
	   ( box.x1 == neighborsBox->x1 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *southNeighbor;
	southNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! southNeighbor->edge_6_isProcessed ){
	  southNeighbor->rootOnEdge_6 = vertex;
	  southNeighbor->edge_6_isProcessed = true;	
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllEastNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y1 == neighborsBox->y1 ) &&
	   ( box.x1 == neighborsBox->x0 ) && 
	   ( box.z0 >= neighborsBox->z0 ) &&  
	   ( box.z1 <= neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D >	 *eastNeighbor; 
	eastNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! eastNeighbor->edge_7_isProcessed )  {
	  eastNeighbor->rootOnEdge_7 = vertex;
	  eastNeighbor->edge_7_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }

}


/** Computation of the root on the edge number 9. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_9(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbrow() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbrow() ; ++i)
    controlCoeff[i] = bezier( i, 0, 0 );
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(), controlCoeff, box.y0, box.y1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_9, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  cube.rootOnEdge_9 = new topology::vertex< topology::point< coeff_t > >( box.x0, sol[0], box.z0 );
  insertVertexInList( scn, cube.rootOnEdge_9 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_9( Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_9;
 
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getB_WNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() ) 
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x1 ) &&  
	   ( box.z0 == neighborsBox->z1 ) ) {
	
        topology::cube3d< MP,REP3D > *B_WNeighbor;
	B_WNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! B_WNeighbor->edge_12_isProcessed ) {
	  B_WNeighbor->rootOnEdge_12 = vertex;
	  B_WNeighbor->edge_12_isProcessed = true;	
	}
      }
      listOfNeighbors.clear();  
    }
  
  node->getAllBackNeighbors( listOfNeighbors );
  
  if ( !listOfNeighbors.empty() ) 
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x0 ) &&  
	   ( box.z0 == neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > * backNeighbor;
	backNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if  ( ! backNeighbor->edge_11_isProcessed ) {
	  backNeighbor->rootOnEdge_11 = vertex;
	  backNeighbor->edge_11_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
    
  node->getAllWestNeighbors( listOfNeighbors );
  
  if ( !listOfNeighbors.empty() ) 
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x1 ) &&  
	   ( box.z0 == neighborsBox->z0 ) ) {
	
	topology::cube3d< MP,REP3D > *westNeighbor;
	westNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! westNeighbor->edge_10_isProcessed ) {
	  westNeighbor->rootOnEdge_10 = vertex;
	  westNeighbor->edge_10_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
}


/** Computation of the root on the edge number 10. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_10(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbrow() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbrow() ; ++i ) 
    controlCoeff[i] = bezier( i, bezier.nbcol()-1, 0 );
  
  if( controlCoeff[0] == 0 ) sol.push_back( 0 );
  
  int nbSol = bezier::isole_eps(sol.rep(),  controlCoeff, box.y0, box.y1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_10, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_10 = new topology::vertex< topology::point< coeff_t > >( box.x1, sol[0], box.z0 );
  insertVertexInList( scn, cube.rootOnEdge_10 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_10( Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_10;
  
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getB_ENeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x0 ) &&  
	   ( box.z0 == neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *B_ENeighbor; 
	B_ENeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! B_ENeighbor->edge_11_isProcessed ) {
	  B_ENeighbor->rootOnEdge_11 = vertex;
	  B_ENeighbor->edge_11_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllBackNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x1 ) &&  
	   ( box.z0 == neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *backNeighbor;
	backNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! backNeighbor->edge_12_isProcessed ){
	  backNeighbor->rootOnEdge_12 = vertex;
	  backNeighbor->edge_12_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllEastNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x0 ) &&  
	   ( box.z0 == neighborsBox->z0 ) ) {
	
        topology::cube3d< MP,REP3D > *eastNeighbor;
	eastNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! eastNeighbor->edge_9_isProcessed ) {
	  eastNeighbor->rootOnEdge_9 = vertex;
	  eastNeighbor->edge_9_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
}

/** Computation of the root on the edge number 11. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_11(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbrow() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbrow() ; ++i )
    controlCoeff[i] = bezier( i, 0, bezier.nbhgt() - 1 ) ;
  
  if ( controlCoeff[0] == 0 ) sol.push_back( 0 );

  int nbSol = bezier::isole_eps(sol.rep(),  controlCoeff, box.y0, box.y1, epsilon);
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_11, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_11 = new topology::vertex< topology::point< coeff_t > >( box.x0, sol[0] ,box.z1 );
  insertVertexInList( scn, cube.rootOnEdge_11 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_11( Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_11;
 
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getF_WNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x1 ) &&  
	   ( box.z1 == neighborsBox->z0 ) ) {
	
	topology::cube3d< MP,REP3D > *F_WNeighbor;
	F_WNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! F_WNeighbor->edge_10_isProcessed ) {
	  F_WNeighbor->rootOnEdge_10 = vertex;
	  F_WNeighbor->edge_10_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllFrontNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x0 ) &&  
	   ( box.z1 == neighborsBox->z0 ) ) { 
	
	topology::cube3d< MP,REP3D > *frontNeighbor;
	frontNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! frontNeighbor->edge_9_isProcessed ) {
	  frontNeighbor->rootOnEdge_9 = vertex;
	  frontNeighbor->edge_9_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
 
  node->getAllWestNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x0 == neighborsBox->x1 ) &&  
	   ( box.z1 == neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > *westNeighbor;
	westNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! westNeighbor->edge_12_isProcessed ){
	  westNeighbor->rootOnEdge_12 = vertex;
	  westNeighbor->edge_12_isProcessed = true;
	}
      }
      listOfNeighbors.clear(); 
    }
 
}


/** Computation of the root on the edge number 12. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::computeRootOnEdge_12(scene<MP,REP3D>& scn) {

  // Calculer la racine.
  bezier::rep1d< coeff_t >  controlCoeff( bezier.nbrow() );
  Seq< coeff_t > sol;
  
  for ( int i = 0 ; i < bezier.nbrow() ; ++i ) 
    controlCoeff[i] = bezier( i, bezier.nbcol()-1, bezier.nbhgt()-1 ) ;
  
  if( controlCoeff[0] == 0 ) sol.push_back( 0 );

  int nbSol = bezier::isole_eps(sol.rep(), controlCoeff, box.y0, box.y1, epsilon );
  
  // La solution doit etre unique
  if ( nbSol != 1 ) {
    std::cerr << " Erreur : computeRootOnEdge_12, + de 1 solution sur l'arete \n";
    exit(0);
  } 
  
  // Creation et insertion du nouveau sommet.
  cube.rootOnEdge_12 = new topology::vertex< topology::point< coeff_t > >( box.x1, sol[0] ,box.z1 );
  insertVertexInList( scn, cube.rootOnEdge_12 );
}

/** Update of the roots on the close cells. 
    This update is carried out on the cells of same size or coarser level. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::propagationOfEdge_12( Node< cell3d< MP, REP3D > > * node ) {
  
  topology::vertex< topology::point< coeff_t > > * vertex;
  vertex = cube.rootOnEdge_12;
 
  // Mettre les aretes des cellules voisines a jour.
  std::vector< Node< cell3d< MP, REP3D > > * >  listOfNeighbors;

  node->getF_ENeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x0 ) &&  
	   ( box.z1 == neighborsBox->z0 ) ) {
	
	topology::cube3d< MP,REP3D > *F_ENeighbor;
	F_ENeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! F_ENeighbor->edge_9_isProcessed ) {
	  F_ENeighbor->rootOnEdge_9 = vertex;
	  F_ENeighbor->edge_9_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
  
  node->getAllFrontNeighbors( listOfNeighbors );
  
  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x1 ) &&  
	   ( box.z1 == neighborsBox->z0 ) ) {
	
	topology::cube3d< MP,REP3D > *frontNeighbor;
	frontNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! frontNeighbor->edge_10_isProcessed ) {
	  frontNeighbor->rootOnEdge_10 = vertex;
	  frontNeighbor->edge_10_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }
 
  
  node->getAllEastNeighbors( listOfNeighbors );

  if ( ! listOfNeighbors.empty() )
    if ( listOfNeighbors.size() == 1 ) {
      
      topology::box3d< coeff_t > * neighborsBox;
      neighborsBox = &listOfNeighbors[0]->GetCell().box;
      
      if ( ( box.y0 >= neighborsBox->y0 ) &&
	   ( box.y1 <= neighborsBox->y1 ) && 
	   ( box.x1 == neighborsBox->x0 ) &&  
	   ( box.z1 == neighborsBox->z1 ) ) {
	
	topology::cube3d< MP,REP3D > * eastNeighbor;
	eastNeighbor = &listOfNeighbors[0]->GetCell().cube;
	
	if ( ! eastNeighbor->edge_11_isProcessed ) {
	  eastNeighbor->rootOnEdge_11 = vertex;
	  eastNeighbor->edge_11_isProcessed = true;
	}
      }
      listOfNeighbors.clear();
    }

}

//----------------------------------------------------------------------

/** Function being useful has to build a triangle by taking account 
    of its orientation. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::makeOnlyOneTriangle(scene<MP,REP3D>& scn  ){
  
  bool rootOnEdge_1 = false,
    rootOnEdge_2 = false,
    rootOnEdge_3 = false,
    rootOnEdge_4 = false,
    rootOnEdge_5 = false, 
    rootOnEdge_6 = false,
    rootOnEdge_7 = false,
    rootOnEdge_8 = false,
    rootOnEdge_9 = false,  
    rootOnEdge_10 = false,
    rootOnEdge_11 = false,
    rootOnEdge_12 = false;
  
  int nbrow = bezier.nbrow();
  int nbcol = bezier.nbcol();
  int nbhgt = bezier.nbhgt();

  if ( cube.rootOnEdge_1  != NULL )  rootOnEdge_1  = true;
  if ( cube.rootOnEdge_2  != NULL )  rootOnEdge_2  = true;
  if ( cube.rootOnEdge_3  != NULL )  rootOnEdge_3  = true;
  if ( cube.rootOnEdge_4  != NULL )  rootOnEdge_4  = true;
  if ( cube.rootOnEdge_5  != NULL )  rootOnEdge_5  = true;
  if ( cube.rootOnEdge_6  != NULL )  rootOnEdge_6  = true;
  if ( cube.rootOnEdge_7  != NULL )  rootOnEdge_7  = true;
  if ( cube.rootOnEdge_8  != NULL )  rootOnEdge_8  = true;
  if ( cube.rootOnEdge_9  != NULL )  rootOnEdge_9  = true;
  if ( cube.rootOnEdge_10 != NULL )  rootOnEdge_10 = true;
  if ( cube.rootOnEdge_11 != NULL )  rootOnEdge_11 = true;
  if ( cube.rootOnEdge_12 != NULL )  rootOnEdge_12 = true;

  if ( rootOnEdge_1 && rootOnEdge_6 && rootOnEdge_10 ) {
    if ( bezier( 0, nbcol - 1, 0 ) > 0 ) {
      makeTriangle(scn, cube.rootOnEdge_6, cube.rootOnEdge_1,
		    cube.rootOnEdge_10 );
      return;
    }
    makeTriangle(scn, cube.rootOnEdge_6, cube.rootOnEdge_10,
		  cube.rootOnEdge_1 );
    return;
  }
  
  if ( rootOnEdge_1 && rootOnEdge_5 && rootOnEdge_9 ) {
    if ( bezier( 0, 0, 0 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_1, cube.rootOnEdge_5,
		    cube.rootOnEdge_9 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_1, cube.rootOnEdge_9,
		  cube.rootOnEdge_5 );
    return;
  }
  
  if ( rootOnEdge_2 && rootOnEdge_6 && rootOnEdge_12 ) {
    if ( bezier( 0, nbcol - 1, nbhgt - 1 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_2, cube.rootOnEdge_6,
		    cube.rootOnEdge_12 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_2, cube.rootOnEdge_12,
		  cube.rootOnEdge_6 );
    return;
  }
  
  if ( rootOnEdge_2 && rootOnEdge_5 && rootOnEdge_11 ) {
    if ( bezier( 0, 0, nbhgt - 1 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_5, cube.rootOnEdge_2,
		    cube.rootOnEdge_11 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_5, cube.rootOnEdge_11,
		  cube.rootOnEdge_2 );
    return;
  }
  
  if ( rootOnEdge_3 && rootOnEdge_8 && rootOnEdge_10 ) {
    if ( bezier( nbrow - 1, nbcol - 1, 0 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_3, cube.rootOnEdge_8,
		    cube.rootOnEdge_10 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_3, cube.rootOnEdge_10,
		  cube.rootOnEdge_8 );
    return;
  }
  
  if ( rootOnEdge_3 && rootOnEdge_7 && rootOnEdge_9 ) {
    if ( bezier( nbrow - 1, 0, 0 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_3, cube.rootOnEdge_9,
		    cube.rootOnEdge_7 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_3, cube.rootOnEdge_7,
		  cube.rootOnEdge_9 );
    return;
  }
  
  if ( rootOnEdge_4 && rootOnEdge_8 && rootOnEdge_12 ) {
    if ( bezier( nbrow - 1, nbcol - 1, nbhgt - 1 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_4, cube.rootOnEdge_12,
		    cube.rootOnEdge_8 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_4, cube.rootOnEdge_8,
		  cube.rootOnEdge_12 );
    return;
  }
  
  if ( rootOnEdge_4 && rootOnEdge_7 && rootOnEdge_11 ) {
    if ( bezier( nbrow - 1, 0, nbhgt - 1 ) > 0 ) {
      makeTriangle(scn,  cube.rootOnEdge_4, cube.rootOnEdge_7,
		    cube.rootOnEdge_11 );
      return;
    }
    makeTriangle(scn,  cube.rootOnEdge_4, cube.rootOnEdge_11,
		  cube.rootOnEdge_7 );
    return;
  }


}


//----------------------------------------------------------------------
/** Function which for a given cube calculates its mesh. **/
template < class MP, class REP3D >
bool cell3d< MP, REP3D >::buildMeshInCube(scene<MP,REP3D>& scn, Node< cell3d< MP, REP3D > > * node ) {
  
   cell3d< MP, REP3D > b;
   b = *this;
   
   std::vector< Node< cell3d< MP, REP3D > > * >  neighbors;

   std::vector< topology::vertex< topology::point< coeff_t > > * > listRoots;
   cube.getAllRoots( listRoots );

   bool isGood = false;
   
   node->getAllFaceNeighbors( neighbors );

   // Est-ce que le noeud courant est plus petit que ses voisins.
   if ( node->isSmallerThanNeighbors( neighbors ) ) 
     isGood = true;
   
   if ( !listRoots.empty() ) {
     
     // Cas ou il n'y a que trois racines sur le cube.
     if ( ( listRoots.size() == 3 ) && ( isGood ) ) {
       makeOnlyOneTriangle(scn);
       return true;
     }
     
     else 
     
       if ( ( listRoots.size() > 3 ) && ( isGood ) ) {
	 node->GetCell().cube.buildTrianglesOnCube(scn);
	 return true;
       }
     
   }
   else
     return false;


   neighbors.clear();

    // Le sommet qui va permettre la triangulation.
   topology::point< coeff_t >  point(3);

   // Calcul de l'intersection entre la diagonale et la surface.
   computeVertexOnDiagonal( point, node->depth);

   topology::vertex< topology::point< coeff_t > > * vertex = new topology::vertex< topology::point< coeff_t > >( point[0], point[1], point[2] );
   insertVertexInList( scn, vertex );

  
   // Pour la face 1.
   node->getAllBackNeighbors( neighbors );
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_1;
     cube.getLinkRootsOnFace_1( rootsOnFace_1, true );
     if ( !rootsOnFace_1.empty() ) 
       if ( rootsOnFace_1.size()==2 ) {
	 //std::cout << " Pour la face 1, rootsOnFace_1\n";
	 makeTriangle(scn, vertex, rootsOnFace_1[0], rootsOnFace_1[1] );
       }
   }
   else {
     // Si les voisins sont plus petits.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_3;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_3( rootsOnFace_3 ,false );
	 if ( !rootsOnFace_3.empty() ) 
	   if ( rootsOnFace_3.size()==2 ) {
	     allNeighborsAreNotStudied = false;
	     //std::cout << " Pour la face 1, rootsOnFace_3\n";
	     makeTriangle(scn, vertex, rootsOnFace_3[0], rootsOnFace_3[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_1;
	 cube.getLinkRootsOnFace_1( rootsOnFace_1, true );
	 if ( !rootsOnFace_1.empty() ) 
	   if ( rootsOnFace_1.size()==2 ) {
	     //std::cout << " Pour la face 1, rootsOnFace_1\n";
	     makeTriangle(scn, vertex, rootsOnFace_1[0], rootsOnFace_1[1] );
	   }
       }
     }
     // Si le voisin est plus grand.
     else { 
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_1;
       cube.getLinkRootsOnFace_1( rootsOnFace_1 , true);
       if ( !rootsOnFace_1.empty() ) 
	 if ( rootsOnFace_1.size()==2 ) {
	   //std::cout << " Pour la face 1, rootsOnFace_1\n";
	   makeTriangle(scn, vertex, rootsOnFace_1[0], rootsOnFace_1[1] );
	 }
     }
     neighbors.clear();
   }
   

   
   // Pour la face 2.
   node->getAllNorthNeighbors( neighbors );
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_2;
     cube.getLinkRootsOnFace_2( rootsOnFace_2, true );
     if ( !rootsOnFace_2.empty() ) 
       if ( rootsOnFace_2.size()==2 ) {
	 //std::cout << " Pour la face 2, rootsOnFace_2\n";
	 makeTriangle(scn, vertex, rootsOnFace_2[0], rootsOnFace_2[1] );
       }
   }
   else {
     // Si les voisins sont plus petits.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_5;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_5( rootsOnFace_5, false );
	 if ( !rootsOnFace_5.empty() ) 
	   if ( rootsOnFace_5.size()==2 ) {
	     allNeighborsAreNotStudied = false;
	     //std::cout << " Pour la face 2, rootsOnFace_5\n";
	     makeTriangle(scn, vertex, rootsOnFace_5[0], rootsOnFace_5[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_2;
	 cube.getLinkRootsOnFace_2( rootsOnFace_2, true );
	 if ( !rootsOnFace_2.empty() ) 
	   if ( rootsOnFace_2.size()==2 ) {
	     //std::cout << " Pour la face 2, rootsOnFace_2\n";
	     makeTriangle(scn, vertex, rootsOnFace_2[0], rootsOnFace_2[1] );
	   }
       }
     }
     else {
       // Si le voisin est plus grand.
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_2;
       cube.getLinkRootsOnFace_2( rootsOnFace_2, true );
       if ( !rootsOnFace_2.empty() ) 
	 if ( rootsOnFace_2.size()==2 ) {
	   //std::cout << " Pour la face 2, rootsOnFace_2\n";
	   makeTriangle(scn, vertex, rootsOnFace_2[0], rootsOnFace_2[1] );
	 }
     }
     neighbors.clear();
   }

   // Pour la face 3.
   node->getAllFrontNeighbors( neighbors );
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_3;
     cube.getLinkRootsOnFace_3( rootsOnFace_3, true);
     if ( !rootsOnFace_3.empty() )
       if ( rootsOnFace_3.size()==2 ){
	 //std::cout << " Pour la face 3, rootsOnFace_3\n";
	 makeTriangle(scn,  vertex, rootsOnFace_3[0], rootsOnFace_3[1] );
       }
   }
   else {
     // Si les voisins sont plus petite.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_1;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_1( rootsOnFace_1, false );
	 if ( !rootsOnFace_1.empty() ) 
	   if ( rootsOnFace_1.size()==2 ){
	     //std::cout << " Pour la face 3, rootsOnFace_1\n";
	     makeTriangle(scn,  vertex, rootsOnFace_1[0], rootsOnFace_1[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_3;
	 cube.getLinkRootsOnFace_3( rootsOnFace_3, true );
	 if ( !rootsOnFace_3.empty() )
	   if ( rootsOnFace_3.size()==2 ){
	     //std::cout << " Pour la face 3, rootsOnFace_3\n";
	     makeTriangle(scn,  vertex, rootsOnFace_3[0], rootsOnFace_3[1] );
	   }
       }
     }
     // Si le voisin est plus grand.
     else {
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_3;
       cube.getLinkRootsOnFace_3( rootsOnFace_3, true );
       if ( !rootsOnFace_3.empty() )
	 if ( rootsOnFace_3.size()==2 ){
	   //std::cout << " Pour la face 3, rootsOnFace_3\n";
	   makeTriangle(scn,  vertex, rootsOnFace_3[0], rootsOnFace_3[1] );
	 }
     }
     neighbors.clear();
   }

   // Pour la face 4.
   node->getAllEastNeighbors( neighbors );
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_4;
     cube.getLinkRootsOnFace_4( rootsOnFace_4, true );
     if ( !rootsOnFace_4.empty() )
       if ( rootsOnFace_4.size()==2 ) {
	 //std::cout << " Pour la face 4, rootsOnFace_4\n";
	 makeTriangle(scn,  vertex, rootsOnFace_4[0], rootsOnFace_4[1] );
       }
   }  
   else {
     // Si les voisins sont plus petits.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_6;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_6( rootsOnFace_6, false );
	 if ( !rootsOnFace_6.empty() )
	   if ( rootsOnFace_6.size()==2 ){
	     allNeighborsAreNotStudied = false;
	     //std::cout << " Pour la face 4, rootsOnFace_6\n";
	     makeTriangle(scn,  vertex, rootsOnFace_6[0], rootsOnFace_6[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_4;
	 cube.getLinkRootsOnFace_4( rootsOnFace_4, true );
	 if ( !rootsOnFace_4.empty() )
	   if ( rootsOnFace_4.size()==2 ) {
	     //std::cout << " Pour la face 4, rootsOnFace_4\n";
	     makeTriangle(scn,  vertex, rootsOnFace_4[0], rootsOnFace_4[1] );
	   }
       }
     }
     // Si le voisin est plus grand.
     else {
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_4;
       cube.getLinkRootsOnFace_4( rootsOnFace_4, true );
       if ( !rootsOnFace_4.empty() )
	 if ( rootsOnFace_4.size()==2 ) {
	   //std::cout << " Pour la face 4, rootsOnFace_4\n";
	   makeTriangle(scn,  vertex, rootsOnFace_4[0], rootsOnFace_4[1] );
	 }
     }
     neighbors.clear();
   }


   // Pour la face 5.
   node->getAllSouthNeighbors( neighbors );  
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_5;
     cube.getLinkRootsOnFace_5( rootsOnFace_5, true );
     if ( !rootsOnFace_5.empty() )
       if ( rootsOnFace_5.size()==2 ){
	   //std::cout << " Pour la face 5, rootsOnFace_5\n";
	   makeTriangle(scn,  vertex, rootsOnFace_5[0], rootsOnFace_5[1] );
       }
   }
   else {
     // Si les voisins sont plus petits.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_2;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_2( rootsOnFace_2, false );
	 if ( !rootsOnFace_2.empty() )
	   if ( rootsOnFace_2.size()==2 ){
	     allNeighborsAreNotStudied = false;
	     //std::cout << " Pour la face 5, rootsOnFace_2\n";
	     makeTriangle(scn,  vertex, rootsOnFace_2[0], rootsOnFace_2[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_5;
	 cube.getLinkRootsOnFace_5( rootsOnFace_5, true );
	 if ( !rootsOnFace_5.empty() )
	   if ( rootsOnFace_5.size()==2 ){
	     //std::cout << " Pour la face 5, rootsOnFace_5\n";
	     makeTriangle(scn,  vertex, rootsOnFace_5[0], rootsOnFace_5[1] );
	   }
       }
     }
     // Si le voisin est plus grand.
     else {
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_5;
       cube.getLinkRootsOnFace_5( rootsOnFace_5, true );
       if ( !rootsOnFace_5.empty() )
	 if ( rootsOnFace_5.size()==2 ){
	   //std::cout << " Pour la face 5, rootsOnFace_5\n";
	   makeTriangle(scn,  vertex, rootsOnFace_5[0], rootsOnFace_5[1] );
	 }
     }
     neighbors.clear();
   }

   // Pour la face 6.
   node->getAllWestNeighbors( neighbors );
   // Pas de voisins.
   if ( neighbors.empty() ) {
     std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_6;
     cube.getLinkRootsOnFace_6( rootsOnFace_6, true );
     if ( !rootsOnFace_6.empty() )
       if ( rootsOnFace_6.size()==2 ){
	 //std::cout << " Pour la face 6, rootsOnFace_6\n";
	 makeTriangle(scn,  vertex, rootsOnFace_6[0], rootsOnFace_6[1] );
       }
   }
   else {
     // Si les voisins sont plus petits.
     if ( neighbors[0]->depth > node->depth ) {
       bool allNeighborsAreNotStudied = true;
       for ( int i = 0 ; i < neighbors.size() ; ++i ) { 
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_4;
	 neighbors[i]->GetCell().cube.getLinkRootsOnFace_4( rootsOnFace_4, false );
	 if ( !rootsOnFace_4.empty() )
	   if ( rootsOnFace_4.size()==2 ){
	     allNeighborsAreNotStudied = false;
	     //std::cout << " Pour la face 6, rootsOnFace_4\n";
	     makeTriangle(scn,  vertex, rootsOnFace_4[0], rootsOnFace_4[1] );
	   }
       }
       // Si tous les voisins sont des voisins a "probleme".
       if ( allNeighborsAreNotStudied ) {
	 std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_6;
	 cube.getLinkRootsOnFace_6( rootsOnFace_6, true );
	 if ( !rootsOnFace_6.empty() )
	   if ( rootsOnFace_6.size()==2 ){
	     //std::cout << " Pour la face 6, rootsOnFace_6\n";
	     makeTriangle(scn,  vertex, rootsOnFace_6[0], rootsOnFace_6[1] );
	   }
       }
     }
     else {
       // Si le voisin est plus grand.
       std::vector< topology::vertex< topology::point< coeff_t > > * > rootsOnFace_6;
       cube.getLinkRootsOnFace_6( rootsOnFace_6, true );
       if ( !rootsOnFace_6.empty() )
	 if ( rootsOnFace_6.size()==2 ){
	   //std::cout << " Pour la face 6, rootsOnFace_6\n";
	   makeTriangle(scn, vertex, rootsOnFace_6[0], rootsOnFace_6[1] );
	 }  
     }
   }

   return true;
}

/** Calculate a vertex on one of the diagonals of the cube to allow the 
    mesh in the cube. **/
template < class MP, class REP3D >
int cell3d< MP, REP3D >::computeVertexOnDiagonal( topology::point< coeff_t > &point, int depthOfNode ) {
  
  // le numero d'une diagonale qui intersecte la surface.
  int diagonal = findDiagonal( bezier );

  int depth = 13 - depthOfNode;

  switch ( diagonal ) { 
  case 1 : findVertexOnDiagonal_1( depth, point );
    return 1;
    break;
  case 2 : findVertexOnDiagonal_2( depth, point );
    return 2;
    break;
  case 3 : findVertexOnDiagonal_3( depth, point );
    return 3;
    break;
  case 4 : findVertexOnDiagonal_4( depth, point );
    return 4;
    break;
  default :
    std::cerr << " Erreur : computeVertexOnDiagonal, aucune diagonales n'intersectent la surface\n";
    exit(0);
  }
}

/** Calculate a vertex on the diagonal number 1 (in our representation) of the cube
    to allow the mesh in the cube. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::findVertexOnDiagonal_1( int depth, topology::point< coeff_t > &point ) {

  std::vector< cell3d< MP, REP3D > > listbezier3d;
  int nbrow = bezier.nbrow();
  int nbcol = bezier.nbcol();
  int nbhgt = bezier.nbhgt();
  
  depth--;
  splitDiagonal( listbezier3d );
  
  // Comparer sommet 000 B_NW avec sommet 000 F_SE 
  
  // Sommet 000 F_SE.
  coeff_t middlePoint = listbezier3d[7].bezier( 0, 0, 0 );
  
  // Sommet 000 B_NW.
  coeff_t firstPoint  = listbezier3d[0].bezier( 0, 0, 0 );
    
  // Recherche de l'intesection.
  if ( ( ( firstPoint < 0 ) && ( middlePoint > 0) ) || 
       ( ( firstPoint > 0 ) && ( middlePoint < 0) ) ) {
    
    // Le nouveau sommet. 
    // rep(point).resize(3);
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[0].box.x1;
      point[1] = listbezier3d[0].box.y1;
      point[2] = listbezier3d[0].box.z1;
      return;
    }
    
    else
      listbezier3d[0].findVertexOnDiagonal_1( depth, point );
    return; 
  }
  
  else {
    
    // Le nouveau sommet.
    // rep(point).resize(3); 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[7].box.x0;
      point[1] = listbezier3d[7].box.y0;
      point[2] = listbezier3d[7].box.z0;
      return;
    }
    
    else
      listbezier3d[7].findVertexOnDiagonal_1( depth, point );
  }
}

/** Calculate a vertex on the diagonal number 2 (in our representation) of the cube
    to allow the mesh in the cube. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::findVertexOnDiagonal_2( int depth, topology::point< coeff_t > &point ) {

  std::vector< cell3d< MP, REP3D > > listbezier3d;
  int nbrow = bezier.nbrow();
  int nbcol = bezier.nbcol();
  int nbhgt = bezier.nbhgt();
  
  depth--;
  splitDiagonal( listbezier3d );
  
  // comparer sommet 001 F_NW avec sommet 000 F_SE 
  
  // Sommet 000 F_SE.
  coeff_t middlePoint = listbezier3d[7].bezier( 0, 0, 0 );
  
  // Sommet 001 F_NW.
  coeff_t firstPoint  = listbezier3d[1].bezier( 0, 0, nbhgt-1 );
    
  // Recherche de l'intesection.
  if ( ( ( firstPoint < 0 ) && ( middlePoint > 0) ) || 
       ( ( firstPoint > 0 ) && ( middlePoint < 0) ) ) {
    
    // Le nouveau sommet. 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[1].box.x1;
      point[1] = listbezier3d[1].box.y1;
      point[2] = listbezier3d[1].box.z0;
      return;
    }
    
    else
      listbezier3d[1].findVertexOnDiagonal_2( depth, point );
    return; 
  }
  
  else {
    
    // Le nouveau sommet. 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[6].box.x0;
      point[1] = listbezier3d[6].box.y0;
      point[2] = listbezier3d[6].box.z1;
      return;
    }
    
    else
      listbezier3d[6].findVertexOnDiagonal_2( depth, point );
  }
}

/** Calculate a vertex on the diagonal number 3 (in our representation) of the cube
    to allow the mesh in the cube. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::findVertexOnDiagonal_3( int depth, topology::point< coeff_t > &point ) {
  
  std::vector< cell3d< MP, REP3D > > listbezier3d;
  int nbrow = bezier.nbrow();
  int nbcol = bezier.nbcol();
  int nbhgt = bezier.nbhgt();
  
  depth--;
  splitDiagonal( listbezier3d );
  
  // comparer sommet 010 B_NE avec sommet 000 F_SE 
  
  // Sommet 000 F_SE.
  coeff_t middlePoint = listbezier3d[7].bezier( 0, 0, 0 );
  
  // Sommet 010 B_NE.
  coeff_t firstPoint  = listbezier3d[4].bezier( 0, nbcol-1, 0 );
   
  // Recherche de l'intesection.
  if ( ( ( firstPoint < 0 ) && ( middlePoint > 0) ) || 
       ( ( firstPoint > 0 ) && ( middlePoint < 0) ) ) {
    
    // Le nouveau sommet. 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[4].box.x0;
      point[1] = listbezier3d[4].box.y1;
      point[2] = listbezier3d[4].box.z1;
      return;
    }
    
    else
      listbezier3d[4].findVertexOnDiagonal_3( depth, point );
    return; 
  }
  
  else {
    
    // Le nouveau sommet. 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[3].box.x1;
      point[1] = listbezier3d[3].box.y0;
      point[2] = listbezier3d[3].box.z0;
      return;
    }
    
    else
      listbezier3d[3].findVertexOnDiagonal_3( depth, point );
  }  
}

/** Calculate a vertex on the diagonal number 4 (in our representation) of the cube
    to allow the mesh in the cube. **/
template < class MP, class REP3D >
void cell3d< MP, REP3D >::findVertexOnDiagonal_4( int depth, topology::point< coeff_t > &point ) {
  
  std::vector< cell3d< MP, REP3D > > listbezier3d;
  int nbrow = bezier.nbrow();
  int nbcol = bezier.nbcol();
  int nbhgt = bezier.nbhgt();
  
  depth--;
  splitDiagonal( listbezier3d );
  
  // comparer sommet 011 F_NE avec sommet 000 F_SE 
  
  // Sommet 000 F_SE.
  coeff_t middlePoint = listbezier3d[7].bezier( 0, 0, 0 );
  
  // Sommet 011 F_NE.
  coeff_t firstPoint  = listbezier3d[5].bezier( 0, nbcol-1, nbhgt-1 );
  
  // Recherche de l'intesection.
  if ( ( ( firstPoint < 0 ) && ( middlePoint > 0) ) || 
       ( ( firstPoint > 0 ) && ( middlePoint < 0) ) ) {
    
    // Le nouveau sommet.
    if ( depth <= 0 ) {
      point[0] = listbezier3d[5].box.x0;
      point[1] = listbezier3d[5].box.y1;
      point[2] = listbezier3d[5].box.z0;
      return;
    }
    
    else
      listbezier3d[5].findVertexOnDiagonal_4( depth, point );
    return;
  }
  
  else {
    
    // Le nouveau sommet. 
    if ( depth <= 0 ) { 
      point[0] = listbezier3d[2].box.x1;
      point[1] = listbezier3d[2].box.y0;
      point[2] = listbezier3d[2].box.z1;
      return;
    }
    
    else
      listbezier3d[2].findVertexOnDiagonal_4( depth, point );
  }  
}

//----------------------------------------------------------------------

/** Posting of the complete structure. **/
template < class MP, class REP3D >
std::ostream & operator<<( std::ostream & os, const cell3d< MP, REP3D > & b ) {
  
//   os << "Coordonnee de la boite pour l'etude de la courbe :\n " 
//      << b.box << endl << endl;
  os << "Bernstein coefficients:\n";
  TENSOR3D::print( os, b.bezier );
  return os;
}

/*********************************************************************/

__END_NAMESPACE_SYNAPS

#endif // axel_mesh_cell3d_H
