/*********************************************************************
*      This file is part of the source code of SYNAPS library          *
*   Author(s): E. Tobis                                              *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
/** \file                                                           **/
#ifndef axel_mesh_rootsOnEdges_H__
#define axel_mesh_rootsOnEdges_H__
//====================================================================
#include <synaps/init.h>

#include <iostream>
#include <vector>
#include <synaps/topology/mesh/cell3d.h>
#include <synaps/usolve/bezier/Std.h>
//====================================================================

__BEGIN_NAMESPACE_SYNAPS

//////////////////////////////////////////////////////////////
// THIS SHOULD BE REPLACED BY THE PARAMETER OF THE SUBDIVISION
// ALGORITHM
const double epsilon = 1.e-3;
// THIS SHOULD BE REPLACED BY THE PARAMETER OF THE SUBDIVISION
// ALGORITHM
//////////////////////////////////////////////////////////////

// This function translates an edge number into coordinates used
// to traverse the 1-dimensional representation of the 3-dimensional
// cell
// beginning will indicate where in the array the iteration should
// start
// delta will indicate the distance between two consecutive coefficients
// boundOni will indicate the number of coefficients we have to read
template <class REP3D>
void translateToFlatCoordinates(const REP3D & tensor,unsigned edge,
				unsigned & beginning,unsigned & delta,
				unsigned & boundOni)
{
  const int ny = tensor.n1_;
  const int nx = tensor.n2_;
  const int nz = tensor.n3_;

  // The n's represent the number of coefficients stored for each variable
  // That is, the degree in the variable + 1

  // This switch can be optimized, so that less comparisons are performed
  switch (edge) {
  case 1: // [x0,x1] y0 z0
    delta = ny;
    beginning = 0;
    boundOni = nx;
    break;
  case 2: // [x0,x1] y0 z1
    delta = ny;
    beginning = ny*nx*(nz-1);
    boundOni = nx;
    break;
  case 3: // [x0,x1] y1 z0
    delta = ny;
    beginning = ny - 1;
    boundOni = nx;
    break;
  case 4: // [x0,x1] y1 z1
    delta = ny;
    beginning = ny*nx*(nz - 1) + ny - 1;
    boundOni = nx;
    break;
  case 5: // x0 y0 [z0,z1]
    delta = ny*nx;
    beginning = 0;
    boundOni = nz;
    break;
  case 6: // x0 y1 [z0,z1]
    delta = ny*nx;
    beginning = ny - 1;
    boundOni = nz;
    break;
  case 7: // x1 y0 [z0,z1]
    delta = ny*nx;
    beginning = ny * (nx - 1);
    boundOni = nz;
    break;
  case 8: // x1 y1 [z0,z1]
    delta = ny*nx;
    beginning = ny*nx - 1; // ny*(nx - 1) + (ny - 1)
    boundOni = nz;
    break;
  case 9: // x0 [y0,y1] z0
    delta = 1;
    beginning = 0;
    boundOni = ny;
    break;
  case 10: // x1 [y0,y1] z0
    delta = 1;
    beginning = ny * (nx - 1);
    boundOni = ny;
    break;
  case 11: // x0 [y0,y1] z1
    delta = 1;
    beginning = ny*nx*(nz-1);
    boundOni = ny;
    break;
  case 12: // x1 [y0,y1] z1
    delta = 1;
    beginning = ny*nx*nz - ny; // ny*nx*(nz-1) + ny*(nx - 1);
    boundOni = ny;
    break;
  default:
    std::cerr << "Wrong edge number: " << edge 
	      << ". Must be 1-12" << std::endl;
    assert(edge >= 1 && edge <= 12);
  }
}

template <class MP,class REP3D,class C>
void getExtrema(const cell3d<MP,REP3D>& cell,unsigned edge,
		C & from,C & to)
{
  // This function takes a cell and an edge number
  // from and to will hold the extrema of the interval represented by the
  // edge.

  switch (edge) {
  case 1: // All these edges correspond to [x0,x1]
  case 2: 
  case 3: 
  case 4:
    from = cell.box.x0;
    to = cell.box.x1;
    break;
  case 5: // All these edges correspond to [z0,z1]
  case 6:
  case 7:
  case 8:
    from = cell.box.z0;
    to = cell.box.z1;
    break;
  case 9: // All these edges correspond to [y0,y1]
  case 10:
  case 11:
  case 12:
    from = cell.box.y0;
    to = cell.box.y1;
    break;
  default:
    std::cerr << "Wrong edge number " << edge << ". Must be 1-12";
    assert (edge >= 1 && edge <= 12);
  }
}

// This function takes a cell and an edge number, and puts in representation
// the coefficients of the polynomial contained in the cell structure, repre
// sented in the Bersntein basis for the received edge.
template <class REP3D>
void extractRepOfEdgeFromTensor(const REP3D & tensor,unsigned edge,
				bezier::rep1d<double> & representation)
{
  
  if (edge < 1 || edge > 12) {
    std::cerr << "Wrong edge number " << edge << ". Must be 1-12";
    assert (edge >= 1 && edge <= 12);
  }
  
  // To have the edge number as a parameter, we make use of the
  // representation of the 3d matrix.
  
  // Location of the Berstein coefficients we
  // want in the flat representation.
  unsigned beginningOfVector,delta; 
  
  unsigned boundOni; // Number of Bernstein coefficients for the edge

  if (tensor.nbrow() * tensor.nbcol() * tensor.nbhgt() == 0)
    representation.resize(0);
  else {
  
    // First, we translate the edge number into a beginning and a displacement
    translateToFlatCoordinates(tensor,edge,beginningOfVector,delta,boundOni);
    
    // With all this information, we extract the coefficients from the tensor
    
    representation.resize(boundOni);
    
    for (unsigned i = 0;i < boundOni;i++) {
      // We get the coefficients from the linear representation of the tensor
      representation[i] = tensor.tab_[beginningOfVector + i*delta];
    }
  }
}

template <class MP,class REP3D,class C>
void computeRootsOnEdge(const cell3d<MP,REP3D> & cell,
			Seq<Seq<C> > & roots,
			unsigned edge)
{
  // This function computes the roots on the edges of the cell.
  std::vector<C> buffer;
  bezier::rep1d<typename MP::coeff_t> controlCoeff;

  extractRepOfEdgeFromTensor(cell.bezier,edge,controlCoeff);

  C from,to; // Extrema of the interval traversed in the edge

  // Second, we get all the bounds we need
  getExtrema(cell,edge,from,to);
  
  // Finally, we compute the roots
  int nbSol = bezier::isole_eps (buffer,controlCoeff,
			  from,to,epsilon);

  // and we put them in the parameter
  roots[edge - 1].resize(nbSol);
  for (unsigned i = 0;i < roots[edge - 1].size();i++)
    roots[edge - 1][i] = buffer[i];
}

__END_NAMESPACE_SYNAPS

#endif
