/*********************************************************************
*      This file is part of the source code of SYNAPS library          *
*   Author(s): G. Gatellier                                          *
*              B. Mourrain, GALAAD, INRIA                            *
**********************************************************************/
#ifndef synaps_topology_point_complex_H
#define synaps_topology_point_complex_H
/*********************************************************************/
#include <vector>
#include <map>
#include <synaps/base/Seq.h>
#include <synaps/topology/point_graph.h>
#include <synaps/base/io/geomview.h>
/*********************************************************************/
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
namespace topology
{

  //! The point complex structure
  /*!
    The point complexe structure allows us to store the different connections
    in the computation of the topology, that implies we can reuse the result
    of the computation of the topology. \n
    For that, the structure stores the points and an adjacency list.

    \li \c R is the container for the points.\n
    \li \c S is the adjacency list.\n
    \li \c F is the list of faces.\n

  */
  template<class C, 
           class R = Seq<topology::point<C> >, 
           class E = Seq<std::vector<int> >,
           class F = Seq<std::vector<int> >
          >
  struct point_complex: point_graph<C,R,E>  {
    /*! The type for the adjacency list. */
    typedef E Vect;
    /*! The type for storing the vertices. */
    typedef R VectPt;
    /*! The type of point. */
    typedef typename R::value_type point_t;
    /*! The type of the coefficients. */
    typedef C coeff_t;
    /*! The type of the coefficients. */
    typedef F faces_t;
    /*! The type of the coefficients. */
    F Faces;
    /*! The default constructor. */
    point_complex():point_graph<C,R,E>(),Faces() {}

    /*! The copy constructor. */
    point_complex(const point_complex<C,R,E> & G)
      : point_graph<C,R,E>(G),Faces(G.Faces) {}

    /*!
      \param nb_point_begin the number that corresponds to the begin point of
      the edge we want to insert in the map \c Pts_Map.
      \param nb_point_end the number that corresponds to the begin point of 
      the edge we want to insert in the map \c Pts_Map.
      
      This procedure inserts the edge represented by \c nb_point_begin 
      and \c nb_point_end  in the data \c Edges of the point graph structure.
    */
    void insert_face(int  idx_pt0, 
		     int  idx_pt1,
		     int  idx_pt2)
      {
	std::vector<int> t(3);
	t[0]=idx_pt0;
	t[1]=idx_pt1;
	t[2]=idx_pt2;
	Faces.push_back(t);
      }
  };
  //--------------------------------------------------------------------
  template <class C, class R, class E, class F>
  void print( geomview::ostream & out, 
	      const point_complex<C,R,E,F> & G,
	      float r = 0.0, float g = 0.0,
	      float b = 0.9, float a = 1.,
	      float sz= 0.01)
    {
      std::vector< std::pair<unsigned,unsigned> > lines;
      for( unsigned  i = 0;  i <  G.Edges.size(); i++ )
	for( unsigned j = 0; j < G.Edges[i].size(); j++ )
	  lines.push_back( std::pair<unsigned,unsigned>(i,G.Edges[i][j]) );
      std::ofstream f;
      f.open (out.file, std::ofstream::out | std::ofstream::app);
      f << "{=SKEL\n";
      f << G.Vertices.size()  << " " << lines.size() << std::endl;
      for ( unsigned i = 0; i < G.Vertices.size(); i ++ )
	{
	  unsigned m = std::min((unsigned)3,G.Vertices[i].size());
	  for ( unsigned d = 0; d < m; d++ )
	    f << G.Vertices[i][d] << " ";
	  for ( unsigned d = m; d < 3; d++ )
	    f << "0 ";
	  f << std::endl;
	};
      
      if(lines.size())
	f << 2 << " " << lines[0].first << " " << lines[0].second //<< std::endl;
	  << " "<< r << " " << g << " " << b << " " << a << std::endl;
      for ( unsigned i = 1; i < lines.size(); i ++)
	f << 2 << " " << lines[i].first << " " << lines[i].second
	  << " "<< r << " " << g << " " << b << " " << a << std::endl;
      f <<"}\n";



      f << "{=COFF\n";
      // Number of points.
      f << G.Vertices.size() <<  " ";

      // Number of triangles.
      f << G.Faces.size() << std::endl;

      // Edge numbers.
      f << "0"  << std::endl;
      
      // List of points.
      for ( int i = 0 ; i < G.Vertices.size() ; i++ )
	f <<  G.Vertices[i][0] << " "
	  <<  G.Vertices[i][1] << " "
	  <<  G.Vertices[i][2] << " "
	  << r << " " << g << " " << b << " " << a  <<"\n";

    // The triangles and their indices.
      for ( int j = 0 ; j < G.Faces.size() ; j++ )
	f <<  "3" << " "
	  << G.Faces[j][0] << " "
	  << G.Faces[j][1] << " "
	  << G.Faces[j][2] << std::endl;
      
      f <<"}\n";

      /*
	for(unsigned i=0; i< G.Vertices.size(); i++) 
	{ 
	print(f,G.Vertices[i],r,g,b,a,sz);
	}
      */
      f.close(); 
    }

  template <class C, class R, class E, class F>
  geomview::ostream & operator<< (geomview::ostream & os, 
				  const point_complex<C,R,E,F> & G)
  {
    print(os,G,os.color.r,os.color.g,os.color.b,1,os.width.sz); 
    return os;
  }

}

//--------------------------------------------------------------------
__END_NAMESPACE_SYNAPS
/*********************************************************************/
#endif // synaps_topology_point_complex_H
