#ifndef TOPDEG2D_H
#define TOPDEG2D_H

#include <synaps/arithm/sign.h>
#include <synaps/arithm/Interval.h>
#include <synaps/linalg/VectDse.h>
#include <synaps/linalg/MatrDse.h>
#include <synaps/linalg/VECTOR.m>
#include <synaps/upol.h>
#include <synaps/mpol.h>
#include <synaps/upol/bezier/rep1d.h>
#include <synaps/usolve/bezier/IslBzStd.h>
#include <synaps/mpol/mpoldse/bernstein.h>

template<class C>
void deCasteljau(bezier::rep1d<C>& f, bezier::rep1d<C>& left, bezier::rep1d<C>& right)
// algorithme de deCasteljau pour subdiviser (f,[a,b]) en (left,[a,(a+b)/2]), (right,[(a+b)/2,b])
{
  int d = f.degree();
  MatrDse<C> m(d+1,d+1);
  C middle = C(0.5);
  for (int i = 0; i < d+1; i++)
  {
    m(0,i) = f[i];
  }
  for (int i = 1; i < d+1; i++)
  {
    for (int j = 0; j < d+1-i; j++)
    {
      m(i,j) = middle*m(i-1,j)+(1-middle)*m(i-1,j+1);
    }
  }
  for (int i = 0; i < d+1; i++)
  {
    left[i] = m(i,0);
    right[i] = m(d-i,i);
  }
}  

template<class C>
C eval(bezier::rep1d<C>& f, const C& x) 
// evaluation de f en x
{
  int d = f.degree();
  MatrDse<C> m(d+1,d+1);
  C middle = x;
  for (int i = 0; i < d+1; i++)
  {
    m(0,i) = f[i];
  }
  for (int i = 1; i < d+1; i++)
  {
    for (int j = 0; j < d+1-i; j++)
    {
      m(i,j) = (1-middle)*m(i-1,j)+(middle)*m(i-1,j+1);
    }
  }
  return m(d,0);
}  

template<class C>
bool csign(bezier::rep1d<C>& f)
// si les coefficients de f sont de signes constants csign = false, sinon csign = true
{
  bool b = false;
  int d = f.degree();
  int i,j;
  for (i = 0; (i < d) && (f[i] == 0); i++);
  
  if (i < d)
  {
    for (j = i+1; (j < d+1)&&(!b); j++)
    {
      b = (sign(f[i]) * sign(f[j]) < 0);
    }
  }
  return b;
}

template<class C>
void subdiv(bezier::rep1d<C>& bf, bezier::rep1d<C>& bg, C& a, C& b, double eps, std::vector<Interval<C> >& L)
// isolation de racines reelles de l'equation bf*bg = 0 sur [a,b]
{
  std::vector<Interval<C> > L0;
  std::vector<bezier::rep1d<C> > Lf, Lg;
  Interval<C> I, I0;
  bezier::rep1d<C> f,g,f1(bf),g1(bg),f2(bf),g2(bg);
  if (csign<C>(bf) || csign<C>(bg))
  {
    Lf.push_back(bf);
    Lg.push_back(bg);
    I.assign(a,b);
    L0.push_back(I);
  }
  C middle = C(0.5);
  while (L0.size())
  {
    int i = L0.size()-1;
    I = L0[i];
    L0.resize(i);
    f = Lf[i];
    Lf.resize(i);
    g = Lg[i];
    Lg.resize(i);
    
    if (I.upper()-I.lower() < eps)
    {
      L.push_back(I);
    }
    else
    {
      deCasteljau(f,f1,f2);
      deCasteljau(g,g1,g2);

      if (csign<C>(f2) || csign<C>(g2))
      {
        Lf.push_back(f2);
        Lg.push_back(g2);
        I0.assign(middle * I.lower() + (1-middle) * I.upper(),I.upper());
        L0.push_back(I0);
      }
        
      if ((f2[0]==0) || (g2[0]==0))
      {
        Lf.push_back(f2);
        Lg.push_back(g2);
        I0.assign(middle * I.lower() + (1-middle) * I.upper(),middle * I.lower() + (1-middle) * I.upper());
        L0.push_back(I0);
      }
        
      if (csign<C>(f1) || csign<C>(g1))
      {
        Lf.push_back(f1);
        Lg.push_back(g1);
        I0.assign(I.lower(),middle * I.lower() + (1-middle) * I.upper());
        L0.push_back(I0);
      }
    }
  }
}

template<class C>
int topdeg2d(MPol<C>& f, MPol<C>& g, VectDse<C>& box)
// calcul le degree topologique deg((f,g),(0,0),box)
{
  C a = box[0], b = box[1], c = box[2], d = box[3];
  
  mpoldse::bernstein<C> bf(f), bg(g),bfa, bfb, bfc, bfd, bga, bgb, bgc, bgd;;
  
  mpoldse::restrict<C>(bf, 0, a, b);
  mpoldse::restrict<C>(bf, 1, c, d);
  mpoldse::face(bfa, bf, 0, 0);
  mpoldse::face(bfb, bf, 0, 1);
  mpoldse::face(bfc, bf, 1, 0);
  mpoldse::face(bfd, bf, 1, 1);
  
  mpoldse::restrict<C>(bg, 0, a, b);
  mpoldse::restrict<C>(bg, 1, c, d);
  mpoldse::face<C>(bga, bg, 0, 0);
  mpoldse::face<C>(bgb, bg, 0, 1);
  mpoldse::face<C>(bgc, bg, 1, 0);
  mpoldse::face<C>(bgd, bg, 1, 1);

  bezier::rep1d<C> fa(bfa.size()), fb(bfb.size()), fc(bfc.size()), fd(bfd.size()), ga(bga.size()), gb(bgb.size()), gc(bgc.size()), gd(bgd.size());
  // fa(t) = f(a,t), fb(t) = f(b,t), fc(t) = f(t,c), fd(t) = f(t,d)
  // ga(t) = g(a,t), gb(t) = g(b,t), fc(t) = f(t,c), fd(t) = f(t,d)
  
  for (int i=0;i<fa.size();i++)
  {
    fa[i] = bfa[i];
  }
  for (int i=0;i<fb.size();i++)
  {
    fb[i] = bfb[i];
  }
  for (int i=0;i<fc.size();i++)
  {
    fc[i] = bfc[i];
  }
  for (int i=0;i<fd.size();i++)
  {
    fd[i] = bfd[i];
  }
  
  for (int i=0;i<ga.size();i++)
  {
    ga[i] = bga[i];
  }
  for (int i=0;i<gb.size();i++)
  {
    gb[i] = bgb[i];
  }
  for (int i=0;i<gc.size();i++)
  {
    gc[i] = bgc[i];
  }
  for (int i=0;i<gd.size();i++)
  {
    gd[i] = bgd[i];
  }

  std::vector<Interval<C> > sa, sb, sc, sd;
  // sa la liste d'intervalles qui isolent les racines reelles de fa*ga=0
  // sb la liste d'intervalles qui isolent les racines reelles de fb*gb=0
  // sc la liste d'intervalles qui isolent les racines reelles de fc*gc=0
  // sd la liste d'intervalles qui isolent les racines reelles de fd*gd=0
  
  double eps = 1e-8;
  subdiv<C>(fa, ga, c, d, eps, sa);
  subdiv<C>(fb, gb, c, d, eps, sb);
  subdiv<C>(fc, gc, a, b, eps, sc);
  subdiv<C>(fd, gd, a, b, eps, sd);
  
  int ja = sa.size();
  if (ja < 2)
  {
    if (ja == 0) sa.resize(1);
    sa[0].assign(c, d);
  }
  else
  {
    sa[0].assign(c, 0.5*(sa[0].upper() + sa[1].lower()));
    for (int i = 1; i < ja-1; i++)
    {
      sa[i].assign(sa[i-1].upper(), 0.5*(sa[i].upper() + sa[i+1].lower()));
    }
    sa[ja-1].assign(sa[ja-2].upper(), d);
  }

  int jb = sb.size();
  if (jb < 2)
  {
    if (jb == 0) sb.resize(1);
    sb[0].assign(c, d);
  }
  else
  {
    sb[0].assign(c, 0.5*(sb[0].upper() + sb[1].lower()));
    for (int i = 1; i < jb-1; i++)
    {
      sb[i].assign(sb[i-1].upper(), 0.5*(sb[i].upper()+sb[i+1].lower()));
    }
    sb[jb-1].assign(sb[jb-2].upper(), d);
  }
  
  int jc = sc.size();
  if (jc < 2)
  {
    if (jc == 0) sc.resize(1);
    sc[0].assign(a, b);
  }
  else
  {
    sc[0].assign(a, 0.5*(sc[0].upper() + sc[1].lower()));
    for (int i = 1; i < jc-1; i++)
    {
      sc[i].assign(sc[i-1].upper(), 0.5*(sc[i].upper() + sc[i+1].lower()));
    }
    sc[jc-1].assign(sc[jc-2].upper(), b);
  }

  int jd = sd.size();
  if (jd < 2)
  {
    if (jd == 0) sd.resize(1);
    sd[0].assign(a, b);
  }
  else
  {
    sd[0].assign(a, 0.5*(sd[0].upper() + sd[1].lower()));
    for (int i = 1; i < jd-1; i++)
    {
      sd[i].assign(sd[i-1].upper(), 0.5*(sd[i].upper()+sd[i+1].lower()));
    }
    sd[jd-1].assign(sd[jd-2].upper(), b);
  }

  int deg = 0;
  
  for (int i = 0; i < ja; i++)
  {
    C u = sa[i].lower(), v = sa[i].upper();
    int sfu = sign(eval<C>(fa,(u-c)/(d-c))), sfv = sign(eval<C>(fa,(v-c)/(d-c))), sgu = sign(eval<C>(ga,(u-c)/(d-c))), sgv = sign(eval<C>(ga,(v-c)/(d-c)));
    if (sfu == sfv)
      deg += sfu * (sgu - sgv);
    else
      deg += sgu * (sfv - sfu);
  }
  
  for (int i = 0; i < jb; i++)
  {
    C u = sb[i].lower(), v = sb[i].upper();
    int sfu = sign(eval<C>(fb,(u-c)/(d-c))), sfv = sign(eval<C>(fb,(v-c)/(d-c))), sgu = sign(eval<C>(gb,(u-c)/(d-c))), sgv = sign(eval<C>(gb,(v-c)/(d-c)));
    if (sfu == sfv)
      deg += sfu * (sgv - sgu);
    else
      deg += sgu * (sfu - sfv);
  }
  
  for (int i = 0; i < jc; i++)
  {
    C u = sc[i].lower(), v = sc[i].upper();
    int sfu = sign(eval<C>(fc,(u-a)/(b-a))), sfv = sign(eval<C>(fc,(v-a)/(b-a))), sgu = sign(eval<C>(gc,(u-a)/(b-a))), sgv = sign(eval<C>(gc,(v-a)/(b-a)));
    if (sfu == sfv)
      deg += sfu * (sgv - sgu);
    else
      deg += sgu * (sfu - sfv);
  }
  
  for (int i = 0; i < jd; i++)
  {
    C u = sd[i].lower(), v = sd[i].upper();
    int sfu = sign(eval<C>(fd,(u-a)/(b-a))), sfv = sign(eval<C>(fd,(v-a)/(b-a))), sgu = sign(eval<C>(gd,(u-a)/(b-a))), sgv = sign(eval<C>(gd,(v-a)/(b-a)));
    if (sfu == sfv)
      deg += sfu * (sgu - sgv);
    else
      deg += sgu * (sfv - sfu);
  }

  return (int)(deg/8);
}
#endif

