/*********************************************************************
*      This file is part of the source code of SYNAPS library.       *
*   Author(s): B. Mourrain, H. Prieto, GALAAD, INRIA                 *
**********************************************************************
History:
$Id: ModUPol.C,v 1.1 2005/07/11 08:12:26 mourrain Exp $
**********************************************************************/
#include <synaps/init.h>
#include <synaps/linalg/MatrStr.h>
#include <synaps/linalg/hankel.h>
#include <synaps/upol/mult_fft.h>
#include <synaps/upol/resultant.h>
//----------------------------------------------------------------------

__BEGIN_NAMESPACE_SYNAPS

namespace quotient {

/// Addition of @a@ and @b@ in @r@.
template <class C, class R>
void ModUPol<C,R>::add(rep_type & r,const rep_type & a, const rep_type& b)
{
  r.resize(dim());  UPOLDAR::add(r,a,b);
}

/// Inplace addition of @a@ to @r@.
template <class C, class R>
void ModUPol<C,R>::add(rep_type & r,const rep_type & a)
{
  r.resize(dim());  UPOLDAR::add(r,a);
}

/// Susbtraction of @a@ and @b@ in @r@.
template <class C, class R>
void ModUPol<C,R>::sub(rep_type & r,const rep_type & a, const rep_type& b)
{
  r.resize(dim());  UPOLDAR::sub(r,a,b);
}

/// Inplace substraction of @a@ to @r@.
template <class C, class R>
void ModUPol<C,R>::sub(rep_type & r,const rep_type & a)
{
  r.resize(dim());  UPOLDAR::sub(r,a);
}

/// Square @f@ in $\Ac$ and put it @x@.
template <class C, class R>
void ModUPol<C,R>::square(rep_type & x,const rep_type & f)
{
  assert(&x != &f);

  typedef typename R::value_type  value_type;
  typedef typename R::size_type   size_type;

  size_type degree = dim();
  size_type M = 2*degree;

  // Compute the convolution product poly * f,
  // by a product of the FFT of poly and f
  x.resize(M); VECTOR::init(x,0);

#ifdef FFT
  UPOLDAR::mul_fft(x, poly, f);
#else
  UPOLDAR::mul_index(x, poly, f);
  //  VECTOR::print(std::cout,x);std::cout<<std::endl;
#endif

  // Change the signs of the last degree terms of poly*f
  for (size_type i = M-1; i>=degree; i--) x[i]*= value_type(-1);

  // Now, we compute the convolution of this result and f
  linalg::rep1d<value_type> res(M+degree-1);

#ifdef FFT
  UPOLDAR::mul_fft(res,x,f);
#else
  UPOLDAR::mul_index(res,x,f);
  //  VECTOR::print(std::cout,res);std::cout<<std::endl;
#endif

  // We keep only the middle part.
  for (size_type j = 0, i = degree; i<M; i++,j++)
    x[j] = value_type(-1)*(res[i]);

  for(size_type i = degree; i<x.size();++i) x[i]=0;

  UPOLDAR::checkdegree(x);
  //      delete [] res;
}

//----------------------------------------------------------------------
/// Multiply @a@ by @b@ in $\Ac$ and put it @x@.
template <class C,class R>
void ModUPol<C,R>::mul(rep_type & x, const rep_type & a, const rep_type& b)
{
  typedef typename R::size_type   size_type;
  typedef typename R::value_type  value_type;

  size_type degree = dim();
  rep_type res(degree,AsSize()), temp(degree,AsSize());

  add(temp,a,b);
  square(x,temp);
  sub(temp,a,b);
  square(res,temp);
  sub(x,res);
  VECTOR::div_ext(x,value_type(4));
  //      delete [] res.tab_;
}

//----------------------------------------------------------------------
/** Express the polynomial $x$ from a basis (Monomial or Horner)
    in the other basis. From the Horner basis to the monomial basis
    (@MtoH=false@), the
    Bezoutian matrix $B_{1,p}$ is used. For the other case (@MtoH=true@),
    the matrix $H_{\tau_{p}}$ is used.
**/
template <class C, class R>
template<class S>
void ModUPol<C,R>::monomial_to_horner(rep_type &x, const S &  pol)
{
  typedef typename R::value_type  value_type;
  typedef typename R::size_type   size_type;
  typedef typename R::rep_type    rep_type;

  size_type degree = dim();
  rep_type h1(2*degree-1, AsSize());
  h1[degree-1]=1/poly[degree];
  for (size_type i = degree; i < 2*degree-1; i++)
    {
      for(size_type j=0; j<degree; j++)
	h1[i] -= h1[i-degree+j]*poly[j];
      h1[i]/=poly[degree];
    }
  //VECTOR::print(std::cout,h1);std::cout<<std::endl;

  /* If x is expressed in the monomial basis and we want it in
     the Horner basis, we compute x = H1 * x. */
  MatrStr<linalg::hankel<value_type> >  P(degree, degree, h1.begin());
  rep_type tmp(pol.begin(),pol.end());
  tmp.resize(degree);
  x.resize(degree); VECTOR::init(x,0);
  using namespace MATRIX; mul_vect(x,P,tmp);
  //  VECTOR::print(std::cout,x);std::cout<<std::endl;
  VECTOR::reverse(x,degree);
}

//----------------------------------------------------------------------
/** Express the polynomial $x$ from a basis (Monomial or Horner)
    in the other basis. From the Horner basis to the monomial basis
    (@MtoH=false@), the
    Bezoutian matrix $B_{1,p}$ is used. For the other case (@MtoH=true@),
    the matrix $H_{\tau_{p}}$ is used.
**/
template <class C, class R>
template<class S>
void ModUPol<C,R>::horner_to_monomial(S &x, const rep_type &  pol)
{
  typedef typename R::value_type  value_type;
  typedef typename R::size_type   size_type;
  typedef typename R::rep_type    rep_type;

  size_type degree = dim();
  rep_type h1(2*degree,AsSize());
  R r(degree,AsSize());

  /* if x is expressed in the Horner basis and we want it
     in the monomial basis, we compute x = P*x. */
  MatrStr<linalg::hankel<value_type> > P =Bezout(poly);
  //  cout <<P<<endl;
  S tmp(pol.begin(),pol.end()); tmp.resize(degree);
  VECTOR::reverse(tmp,degree);
  x.resize(degree); VECTOR::init(x,0);
  using namespace MATRIX; mul_vect(x,P,tmp);
}
//----------------------------------------------------------------------


} //namespace modulo

__END_NAMESPACE_SYNAPS

