/*********************************************************************
*   This file is part of the source code of the SYNAPS library.
*   Author(s): B. Mourrain, GALAAD, INRIA
*   $Id: bezier/rep.h,v 1.4 2005/09/28 06:40:51 mourrain Exp $
**********************************************************************/
#ifndef synaps_upol_bezier_rep_h_
#define synaps_upol_bezier_rep_h_
/** \ingroup upol **/
/*********************************************************************/
#include <synaps/init.h>
#include <synaps/arithm/binomial.h>
#include "synaps/arithm/let.h"
#include <synaps/linalg/rep1d.h>
#include <synaps/linalg/VECTOR.m>
#include <synaps/upol/bezier/BEZIER.m>
#include <synaps/upol/UPolDse.h>
// --------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
/// Internal representation for Bezier curves and surfaces
namespace bezier
{

  /** Container storing the control coefficients of a polynomial
      in the Bernstein basis $B_d^i=\binomial{d}{i} \, x^i \, (1-x)^i$.
  **/
  template<class C>
  struct rep1d : public linalg::rep1d<C>
  {
    typedef rep1d<C> self_t;
    typedef typename linalg::rep1d<C>::size_type  size_type;

    rep1d<C>():linalg::rep1d<C>(){}
    rep1d<C>(unsigned n):linalg::rep1d<C>(n){}

    rep1d<C>(const rep1d & r):linalg::rep1d<C>(r){}
    rep1d<C>(size_type l, C* t): linalg::rep1d<C>(l,t){}

    int degree(void) const {return this->size()-1;}

    self_t & operator=(const self_t & r){
      this->linalg::rep1d<C>::operator=(r);
      return *this;
    }

  };

  //----------------------------------------------------------------------
  // Univariate Bezier representation with fixed size.
  template<class C, unsigned n>
  struct rep1dfs
  {
    typedef C value_type ;
    typedef unsigned size_type ;
    typedef value_type * iterator;
    typedef const value_type *  const_iterator;

    value_type data[n];
    unsigned size() const { return n;};
    const_iterator begin() const { return data; };
    const_iterator end()   const { return data + n; };
    inline       value_type & operator[] (size_type i)       {return data[i];}
    inline const value_type & operator[] (size_type i) const {return data[i];}

  };

  //--------------------------------------------------------------------
  template<class C, unsigned n, class T>
  void assign(rep1dfs<C,n> & b, const T & p)
  {
    for (unsigned k=0;k<p.size();k++)
      b[k]= p[k]/binomial<C>(p.size()-1,k);
  }
  
  template<class C>
  void checkdegree(rep1d<C> & r) {  }

  template<class C>
  int degree(const rep1d<C> & p) {return p.size()-1;}
  
  template<class C>
  std::ostream & print(std::ostream & os, const rep1d<C> & p, char* x)
  { 
    UPolDse<C> r(p.size(),type::AsSize());
    VECTOR::assign(r,p);
    BEZIER::bezier2monomials<C>(r.rep(),r.size());
    return (os<<r);
  }

  template <class C>
  inline void set_monomial(rep1d<C> & x, const C & c, unsigned n)
  {
    rep1d<C> y; y.resize(n+1);
    VECTOR::init(y,(C)0);
    y[n]=c;
    x.resize(n+1);
    BEZIER::monomials2bezier(x,y,n+1,(C)0,(C)1);
  }

  template<class C>
  void add_cst(rep1d<C> & p, const C & c) 
  { 
    BEZIER::add_cst(p,c);
  }

  template<class C>
  void add(rep1d<C> &  r, const rep1d<C> &  a, const rep1d<C> &  b)
  { 
    BEZIER::add(r,a,b);
  }

  template<class C>
  void add(rep1d<C> &  a, const rep1d<C> &  b)
  { 
    BEZIER::add(a,b);
  }

  template<class C>
  void sub(rep1d<C> &  r, const rep1d<C> &  a, const rep1d<C> &  b)
  { 
    BEZIER::sub(r,a,b);
  }

  template<class C>
  void mul(rep1d<C> &  r, const rep1d<C> &  a, const rep1d<C> &  b)
  { 
    BEZIER::mul(r,a,b);
  }

  template<class C>
  void sub(rep1d<C> &  a, const rep1d<C> &  b)
  { 
    BEZIER::sub(a,b);
  }

  template<class C>
  void diff(rep1d<C> & r, const rep1d<C> & p) 
  { 
    BEZIER::diff(r,p);
  }

  template<class C,class X>
  X eval(const rep1d<C> & p, const X & x) 
  { 
    return BEZIER::eval<X>(p,x);
  }
  //--------------------------------------------------------------------

} //namespace bezier
//====================================================================
namespace let {
  //----------------------------------------------------------------------
  /** Convert a polynomial $p=\sum_{i=0}^d\, p_{i} \,x^{i}$ as
      $b=\sum_{i=0}^{d}\, p_{i}\, t^{i}(1-t)^{d-i} =\sum_{i=0}\,\frac{p_{i}}{binomial(d,i)} B^d_{i}(t)$.
      This corresponds to the change
      of variables $x=\frac{t}{1-t}$ or $t=\frac{x}{x+1}$, which maps
      $[0,\infty[$ to $[0,1[$.
  **/
  template<class C, class T>
  void assign(bezier::rep1d<C> & b, const T & p)
  {
    using let::assign;
    b.resize(p.size());
    for (unsigned k=0;k<p.size();k++)
      {
	assign(b[k],p[k]);
	b[k]/=binomial<C>(p.size()-1,k);
      }
  }

  //--------------------------------------------------------------------
  /** Similar but divides by the largest power of $x$ before. **/
  template<class C, class T>
  void assign(bezier::rep1d<C> & b, const T & p, unsigned vl)
  {
    using let::assign;
    b.resize(p.size()-vl);
    for (unsigned k=vl;k<p.size();k++)
      {
	assign(b[k-vl],p[k]);
	b[k-vl]/=binomial<C>(p.size()-vl-1,k-vl);
      }
  }

  //--------------------------------------------------------------------
  /** Convert to the Bernstein basis representation on [a,b]. **/
  template<class C, class T, class U>
  void assign(bezier::rep1d<C> & bz, const T & p, const U& a, const U& b)
  {
    using let::assign;
    int sz= degree(p)+1;
    bz.resize(sz);
    BEZIER::monomials2bezier(bz,p,sz,a,b);    
  }
} //let
/*********************************************************************/
__END_NAMESPACE_SYNAPS
#endif //synaps_upol_bezier_rep_h_
