#ifndef SYNAPS_ARITHM_ALGEBRAIC_APX_ROOT_OF_H
#define SYNAPS_ARITHM_ALGEBRAIC_APX_ROOT_OF_H


#include <synaps/arithm/algebraic/Algebraic_config.h>


__ALGEBRAIC_BEGIN_NAMESPACE


/** 
	* @brief A class for approximate algebraic numbers
	*
	* The class is parametrized by a Number_traits class, in order
	* to derive the types of the approximate arithmetic. The class
	* has the same functionality as the root_of class. 
	*
	* We can not decide equality of two numbers of this type!
	*
	*
	* \sa Number_traits, root_of
	*/
template < typename K_ >
struct apx_root_of
{
  typedef K_                        K;
  typedef typename K::XT            XT;
  typedef typename K::XIT           XIT;

  typedef typename K::APX_POLY      POLY;

  typedef typename POLY::size_type  size_type;

  // members
  // const static XT eps_;

  POLY         C_;             ///< The polynomials that defines the algebraic number
  mutable XIT  interval_;      ///< The isolating interval
  XT           eps_;           ///< The accuracy

  int          sign_left_;     ///< Sign on the left endpoint of the isolating interval
  int          sign_right_;    ///< Sign on the right endpoint of the isolating interval 

  apx_root_of(const POLY& f, const XIT& I, const XT& eps = 1e-6)
	: C_(f),
      interval_(I),
	  eps_(eps)
		{
			sign_left_  = sign(UPOLDAR::eval_horner(C_, left()));
			sign_right_ = opposite(sign_left_);
		}

  apx_root_of(const POLY& f, const XT& l, const XT& r, const XT& eps = 1e-6)
	: C_(f),
      interval_(l, r),
	  eps_(eps)
  {
	sign_left_  = sign(UPOLDAR::eval_horner(C_, left()));
	sign_right_ = opposite(sign_left_);
  }


  POLY poly() const { return C_; }

  XIT interval() const { return interval_; }
  XT left()  const { return interval_.lower(); }
  XT right() const { return interval_.upper(); }

  XT operator[](const size_type& i) const { return C_[i]; }

  int sign_left()   const { return this->sgn_low_; }
  int sign_right()  const { return this->sgn_upper_; }

  XT eps()  const { return eps_; }

  int improve_once()  const
	{
		std::pair<XIT, XIT> tmp = bisect(interval_);
		int sign_middle =  sign(UPOLDAR::eval_horner(C_, tmp.first.upper()));
		if (sign_middle == sign_right()) {
			interval_.assign(tmp.first);
			return 1;
		} else if (sign_middle == sign_left()) {
			interval_.assign(tmp.second);
			return 1;
		}
		return 0;
	}

  int improve()   const
	{
		int times = 2;
		int res = 0;
		for (int i = 0; ((i < times) && (interval_.width() < eps())); ++i)
			res = improve_once();
		return res;
	}

};

// template < typename POLY_,
//            typename NTR_ >
// const typename apx_root_of<POLY_, NTR_>::XT apx_root_of<POLY_, NTR_>::eps = 1e-6;



template < typename K >
int compare(const apx_root_of<K>& a, const apx_root_of<K>& b)
{
  if (a.interval() > b.interval()) return LARGER;
  if (a.interval() < b.interval()) return SMALLER;

  int ia = a.improve();
  int ib = b.improve();

  if (ia || ib) {
	if (a.interval() > b.interval()) return LARGER;
	if (a.interval() < b.interval()) return SMALLER;
	return UNKNOWN;
  }
  return UNKNOWN;
}



template < typename K >
int compare(const apx_root_of<K>& a, const typename K::RT& x)
{
  if (a.interval() > x) return LARGER;
  if (a.interval() < x) return SMALLER;

  int ia = a.improve();
  if (ia) {
	if (a.interval() > x) return LARGER;
	if (a.interval() < x) return SMALLER;
	return UNKNOWN;
  }
  return UNKNOWN;
}


template < typename K >
int compare(const apx_root_of<K>& a, const typename K::FT& x)
{
  if (a.interval() > x) return LARGER;
  if (a.interval() < x) return SMALLER;

  int ia = a.improve();
  if (ia) {
	if (a.interval() > x) return LARGER;
	if (a.interval() < x) return SMALLER;
	return UNKNOWN;
  }
  return UNKNOWN;
}


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_APX_ROOT_OF_H

