// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)
//
// A routine that isolates the roots of a quartic
// additionally decides the numbers of the roots of the quartic
// based on a special discrimination sytem based on Sturm sequence
//

#ifndef SYNAPS_ARITHM_ALGEBRAIC_ADJUST_POLY_CC
#define SYNAPS_ARITHM_ALGEBRAIC_ADJUST_POLY_CC

#include <synaps/arithm/algebraic/Algebraic_config.h>
#include <synaps/arithm/algebraic/Util.h>
#include <synaps/upol/Bound.h>

__ALGEBRAIC_BEGIN_NAMESPACE



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::redefine(const RT& a, const RT& b, const int& pmu, const int& pos)
{
  typedef Number_traits<RT>   NTR;
  typedef typename NTR::FT    FT;

  construct_poly(b, a, C_);
  make_lcoeff_positive(C_);
  D_ = RT(1);
  mu_ = ROOT_ONE;
  pmu_ = pmu;
  pos_ = pos;
  rootnb_ = 0;

  FT tmp = to_FT(-b, a);
  set_interval(tmp, tmp);
  return;
}


//D1= inv_[0]; D2= inv_[1]; D3= inv_[2]; W1= inv_[3]; W2= inv_[4]; W3= inv_[5]



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::refine_isolating_intervals()
{
  assert(degree() != 1);
  assert(degree() != 2);

  int sgn_a = sign(left_isolating_point());  if (sgn_a == ZERO) return;
  int sgn_b = sign(right_isolating_point()); if (sgn_b == ZERO) return;

  // We have to ask for the rootnb_ !!!!!
  if ((sgn_a < 0) &&  (sgn_b > 0)) {
	if (poly_[0] > RT(0)) {
	  r.set_interval(r.left(), 0);
	  return;
	}
	if ((C_[0]) < RT(0)) {
	  r.set_interval(0, r.right());
	  return;
	}
	//if (poly_[0] == RT(0))
	redefine(RT(1), RT(0), pos_);
	return;
  }
  return;
}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::compute_isolating_intervals_2()
{
  if (degree() != 2) return;
  if (which() == 0) {
	set_interval(-Bound(poly(), Cauchy()), to_FT(-C_[1], C_[2] * 2));
	return;
  }
  if (which() == 1) {
	set_interval(to_FT(-C_[1], C_[2] * 2), Bound(poly(), Cauchy()));
	return;
  }
  return;
}




template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly()
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  assert(degree() >= 1);

  int dg = degree();
  adjust_sign_coeff(C_);
  if (dg == 1) {
	mu_ = ROOT_ONE;
	D_ = RT(1);

	FT temp = to_FT(-C_[0], C_[1]);
	set_interval(temp, temp);
	return;
  }
  if (dg == 2) {
	adjust_poly_2();
	compute_isolating_intervals_2();
	//	refine_isolating_intervals_2(r);
	return;
  }
  if (dg == 3) {
	adjust_poly_3();
	//		refine_isolating_intervals(r);
	return;
  }
  //  if (dg == 4)
  adjust_poly_4();
  //	refine_isolating_intervals(r);
  return;
}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_2()
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  assert(degree() == 2);

  D_ = SQR(C_[1]) - RT(4) * C_[2] * C_[0];

  if (discriminant() > RT(0)) {
	mu_ = ROOT_ONE_ONE;
	pos_ = (rootnb_ == 0 ? POS_FIRST : POS_LAST);
  } else if (discriminant() == RT(0)) {
	mu_ = ROOT_DOUBLE;
	redefine(RT(C_[2]*2), RT(C_[1]), 2, POS_FIRST_LAST);
  } else {
	assert(discriminant() >= RT(0));
	mu_ = ROOT_COMPLEX_COMPLEX;
  }
  return;
}




template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_3()
{
  typedef Number_traits<RT>   NTR;
  typedef typename NTR::FT    FT;

  RT a = RT( 3) * C_[3];
  RT b = RT(-1) * C_[2];
  RT c = RT( 1) * C_[1];
  RT d = RT(-3) * C_[0];
  // D1 = D1, D2 = D2, D3 = D3, W = W
  RT D2x2;
  RT D2 = calc_D2(a, b, c);
  RT D3 = calc_D2(b, c, d);
  RT W = calc_brackets(b, c, a, d);
  RT D2x4 = MUL( RT(4), D2 );
  D_ = calc_D2(D2x4, W, D3);

  D2x2 = MUL( RT(2), D2 );
  if (discriminant() > RT(0)) {
	mu_ = ROOT_ONE_COMPLEX_COMPLEX;
	pos_ = POS_FIRST_LAST;
	if (C_[0] > RT(0)) { //the real root is negative
	  set_interval(-Bound(poly(), Cauchy()), 0);
	  return;
	}
	if (C_[0] < RT(0)) { // the real root is positive
	  set_interval(0, Bound(poly(), Cauchy()));
	  return;
	}
	//else if (r[0] == RT(0))  // the real root is zero
	redefine(RT(1), RT(0), 1, POS_FIRST_LAST);
	return;
  }
  if (discriminant() < RT(0)) {
	// TODO Check if 0 (zero) is a root
	mu_ = ROOT_ONE_ONE_ONE;
	D2x2 = MUL( RT(2), D2 );
	RT f_ba = calc_F3_BA(a, b, W, D2x2);
	if (f_ba > RT(0)) {
	  // W/D is NEVER a root
	  if (rootnb_ == 0) {
		pos_ = POS_FIRST;
		set_interval(-Bound(poly(), Cauchy()), to_FT(b, a));
		return;
	  }
	  if (rootnb_ == 1) {
		pos_ = POS_MIDDLE;
		set_interval(to_FT(b, a), to_FT(W, D2x2));
		return;
	  }
	  // if (r.rootnb_ == 2)
	  pos_ = POS_LAST;
	  set_interval(to_FT(W, D2x2), Bound(poly(), Cauchy()));
	  return;
	}
	if (f_ba < RT(0)) {
	  // W/D is NEVER a root
	  if (rootnb_ == 0) {
		pos_ = POS_FIRST;
		set_interval(-Bound(poly(), Cauchy()), to_FT(W, D2x2));
		return;
	  }
	  if (rootnb_ == 1) {
		pos_ = POS_MIDDLE;
		set_interval(to_FT(W, D2x2), to_FT(b, a));
		return;
	  }
	  // if (rootnb_ == 2)
	  pos_ = POS_LAST;
	  set_interval(to_FT(b, a), Bound(poly(), Cauchy()));
	  return;
	}
	// if b/a is a root then
	// apply half the separation bound to the left and to right
	// in order to have isolating points for the other 2 roots
	// We use Collins CONJECTURE for the bound (so we do a test)
	if (rootnb_ == 1) {
	  redefine(a, -b, 1, POS_MIDDLE);
	  return;
	}
	RT L = std::max( std::max(abs(C_[3]), abs(C_[2])),
		             std::max(abs(C_[1]), abs(C_[0])) );
	L = SQR(L);
	L = MUL(RT(3), L);
	RT q  = MUL(a, L);
	if (rootnb_ == 0) {
	  pos_ = POS_FIRST;
	  RT p1 = calc_brackets(b, L, RT(1), a);
	  set_interval(-Bound(poly(), Cauchy()), to_FT(p1, q));
	  return;
	}
	//		if (rootnb_ == 2)
	pos_ = POS_LAST;
	RT p2 = calc_brackets(b, L, RT(-1), a);
	set_interval(to_FT(p2, q), Bound(poly(), Cauchy()));
	return;
  }
  // if (D1 == RT(0))
  // If the cubic has multiple roots then compute as rationals
  if (D2 == RT(0)) {
	mu_ = ROOT_TRIPLE;
	// The cubic has one triple root
	redefine(a, -b, 3, POS_FIRST_LAST);
	return;
  }
  //	if (D2 != 0)
  RT P = calc_P(a, W, b, D2);
  RT Dr = calc_brackets(RT(2), P, a, W);
  if (Dr > RT(0)) {
	mu_ = ROOT_DOUBLE_ONE;
	if (rootnb_ == 2)
	  redefine(MUL(a, D2), -P, 1, POS_LAST);
	else
	  redefine(D2x2, -W, 2, POS_FIRST);
	return;
  }
  //	 else
  mu_ = ROOT_ONE_DOUBLE;
  if (rootnb_ == 0)
	redefine(MUL(a, D2), -P, 1, POS_FIRST);
  else
	redefine(D2x2, -W, 2, POS_LAST);
  return;
}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::assign_isolating_intervals(
	const RT& p1, const RT& q1,
	const RT& p2, const RT& q2,
	const RT& p3, const RT& q3
	)
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  print(__FUNCTION__);
  if (rootnb_ == 0) {
	set_interval(-Bound(poly(), Cauchy()), to_FT(p1, q1));
	pos_ = POS_FIRST;
	return;
  }
  if (rootnb_ == 1) {
	set_interval(to_FT(p1, q1), to_FT(p2, q2));
	pos_ = POS_MIDDLE;
	return;
  }
  if (rootnb_ == 2) {
	set_interval(to_FT(p2, q2), to_FT(p3, q3));
	pos_ = POS_MIDDLE;
	return;
  }
  //	if (r.rootnb_ == 4)
  set_interval(to_FT(p3, q3), Bound(poly(), Cauchy()));
  pos_ = POS_LAST;
  return;

}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_4()
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

	DEBUG::print(__FUNCTION__);

  RT a = RT(12) * C_[4];
  RT b = RT(-3) * C_[3];
  RT c = RT( 2) * C_[2];
  RT d = RT(-3) * C_[1];
  RT e = RT(12) * C_[0];

  RT D2 = calc_D2(a, b, c); //D2
  RT D3 = calc_D2(b, c, d); //D3
  RT W1 = calc_brackets(a, d, b, c); //W1
  RT W2 = calc_brackets(b, e, c, d); //W2
  RT W3 = calc_brackets(a, e, b, d); //W3

  RT D2_x2 = MUL( RT(2), D2 );
  RT W1_neg  = MUL( RT(-1), W1 );
  RT A  = calc_A(W3, D3);
  RT B  = calc_B(d, W1, e, D2, c, D3);
  RT T1 = calc_T1(W1, D2, D3, W3);
  // calc the discriminant
  D_ = calc_D4(A, B);

  if (discriminant() > RT(0)) {
	if ((T1 <= 0) || (D2 <= 0)) {
	  mu_ = ROOT_ALL_COMPLEX;
	  return;
	}
	adjust_poly_4_ONE(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B
		);
	return;
  }
  if (discriminant() < RT(0)) {
	mu_ = ROOT_ONE_ONE_COMPLEX_COMPLEX;
	adjust_poly_2_ONE_2_COMPLEX(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B
		);
	return;
  }
  //  From now on (r.discriminant() == RT(0))
  if (T1 > RT(0)) {
	adjust_poly_DOUBLE_2_ONE(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, T1
		);
	return;
  }
  if (T1 < RT(0)) {
	//X		mu_ = ROOT_DOUBLE_COMPLEX_COMPLEX;
	// We can compute the root exactly from PRS
	RT p1, q1;
	if (D2 == RT(0)) {
	  p1 = W3;
	  q1 = MUL( RT(3), W1 );
	} else {
	  RT T2 = calc_T2(A, W1, b, B);
	  p1 = MUL( RT(-1), T2 );
	  q1 = RT(3) * T1;
	}
	redefine(q1, -p1, 2, POS_FIRST_LAST);
	return;
  }
  // From now on r.D_ == T1 == 0
  //	if (T1 == RT(0))
  if (D2 > RT(0)) {
	RT R = calc_brackets(b, D2_x2, W1_neg, a);
	if (R == RT(0)) {
	  mu_ = ROOT_DOUBLE_DOUBLE;
	  //TODO: Do it better !!!! No other constructor
	  POLY P;
	  construct_poly(RT(-W3), RT(RT(3) * W2), RT(RT(3) * D2), P);
	  // P[2] = RT(3) * D2;
	  // P[1] = RT(3) * W1;
	  // P[0] = -W3;
	  adjust_sign_coeff(P);
	  RT PD = SQR(P[1]) - RT(4)*P[2]*P[0];
	  FT isol2 = Bound(P, Cauchy());
	  FT isol0 = -isol2;
	  FT isol1 = to_FT(-P[1], RT(2)*P[2]);
	  if ( (rootnb_ == 0) || (rootnb_ == 1) ) {
		*this = root_of<RT, POLY>(P, 0, ROOT_ONE_ONE, 2, PD, isol0, isol1, true, POS_FIRST );
	  }  else {
		*this = root_of<RT, POLY>(P, 1, ROOT_ONE_ONE, 2, PD, isol1, isol2, true, POS_LAST );
	  }
	  return;
	}
	// else R != 0
	// We know the roots exactly.
	// DECIDE if the order is  1  3  or   3  1
	RT tp1 = W1_neg;
	RT tq1 = D2_x2;
	RT tp2 = calc_triple_p(a, W1, b, D2);
	RT tq2 = calc_triple_q(a, D2);
	RT Dp  = calc_brackets(tp1, tq2, tp2, tq1);
	if (Dp > RT(0)) { //3 1
	  mu_ = ROOT_ONE_TRIPLE;
	  if (rootnb_ == 0)
		redefine(tq2, -tp2, 1, POS_FIRST);
	  else
		redefine(tq1, -tp1, 3, POS_LAST);
	  return;
	}
	// else  // 1 3
	mu_  = ROOT_TRIPLE_ONE;
	if (rootnb_ == 3)
	  redefine(tq2, -tp2, 1, POS_LAST);
	else
	  redefine(tq1, -tp1, 3, POS_FIRST);
	return;
  }
  if (D2 < RT(0)) {
	assert(D2 >= RT(0));
	mu_ = ROOT_ALL_COMPLEX;
	return;
  }
  // D2 = 0
  // We know the root exactly, it is  b/a OR - C3_/(4*C4_)
  // r.mu_ = ROOT_QUADRUPLE;
  redefine(a, -b, 4, POS_FIRST_LAST);
  return;
}



// Adjust the poly when it has 4 distinct real roots
template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_4_ONE(
	const RT& a,     const RT& b,      const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,     const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B
	)
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  print(__FUNCTION__);

  mu_ = ROOT_ONE_ONE_ONE_ONE;
  RT f_ba = calc_F4_BA(a, A, D2);
  RT R = calc_R(a, W1, b, D2_x2);
  if (f_ba > RT(0)) {
	adj_4_ONE_fba_pos(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R
		);
	return;
  }  // if (f_ba > RT(0))
  if (f_ba < RT(0)) {
	adj_4_ONE_fba_neg(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R
		);
	return;
  } // if (f_ba < RT(0))
  //if (f_ba == RT(0)) {
  adj_4_ONE_fba_zero(
	  a, b, c, d, e,
	  D2, D3, W1, W2, W3,
	  D2_x2, W1_neg,
	  A, B, R
	  );
  //  } // if (f_ba == RT(0))
  return;
}




template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::test_accuracy(const RT& p, const RT& q, int sgn)
{
  typedef Number_traits<RT>    NTR;

  int test_sgn;
  test_sgn = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
  if (test_sgn != sgn) {
	std::cerr << "We do not have enough bits for sqrt" << std::endl;
	std::cerr << "Nothing to do for the moment" << std::endl;
	exit(1);
  }
  return;
}



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_4_ONE_fba_pos(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R
	)
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  // b/a is in the middle of the 4 roots
  // we calculate the other 2 isolating points
  // using the roots of the isolator polynomial
  print(__FUNCTION__);

  if (rootnb_ == 0) {
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p1 = IB - ID;
	RT q1 = IA;
	test_accuracy(p1, q1, NEGATIVE);

	set_interval(-Bound(poly(), Cauchy()), to_FT(p1, q1));
	pos_ = POS_FIRST;
	return;
  }
  if (rootnb_ == 1) {
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p1 = IB - ID;
	RT q1 = IA;
	test_accuracy(p1, q1, NEGATIVE);

	set_interval(to_FT(p1, q1), to_FT(b, a));
	pos_ = POS_MIDDLE;
	return;
  }
  if (rootnb_ == 2) {
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p3 = IB + ID;
	RT q3 = IA;
	test_accuracy(p3, q3, NEGATIVE);

	set_interval(to_FT(b, a), to_FT(p3, q3));
	pos_ = POS_MIDDLE;
	return;
  }
  //	if (r.rootnb_ == 4)
  RT IB = RT(-3) * W1;
  RT IA = RT( 6) * D2;
  RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
  RT p3 = IB + ID;
  RT q3 = IA;
  test_accuracy(p3, q3, NEGATIVE);

  set_interval(to_FT(p3, q3), Bound(poly(), Cauchy()));
  pos_ = POS_LAST;
  return;
}



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_4_ONE_fba_neg(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R
	)
{
  typedef Number_traits<RT>    NTR;

  // b/a is either the 1st or the 3rd isolating point
  // decide its index and calculate the other 2 isolating points
  // We know for sure that one is -W1_/D2_x2
  // the remaining isolating points is derived as a root
  // of the isolator polynomial
  print(__FUNCTION__);

  RT f_WD = calc_F4_WD(a, B, D2, D3, W1);
  // Decide the order of b/a and -W1/D2_x2
  if (R > RT(0)) {    //  -W1/D2_x2 < b/a
	RT p3 = b;
	RT q3 = a;
	if (f_WD > RT(0)) {
	  RT IB = RT(-3) * W1;
	  RT IA = RT( 6) * D2;
	  RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	  RT p1 = IB - ID;
	  RT q1 = IA;
	  RT p2 = W1_neg;
	  RT q2 = D2_x2;
	  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	  return;
	} // if (f_WD > RT(0))
	if (f_WD < RT(0)) {
	  RT p1 = W1_neg;
	  RT q1 = D2_x2;
	  RT IB = RT(-3) * W1;
	  RT IA = RT( 6) * D2;
	  RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	  RT p2 = IB + ID;
	  RT q2 = IA;
	  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	  return;
	} //if (f_WD < RT(0))
	//	if (f_WD == RT(0))
	if (rootnb_ == 1) {
	  redefine(D2_x2, -W1_neg, 1, POS_MIDDLE);
	  return;
	}
	// TODO Test for Collins
	RT L = std::max( std::max( std::max(abs(a), abs(b)),
		                       std::max(abs(c), abs(d)) ),
		             abs(e) );
	RT L_2x8 = SQR(L);
	L_2x8 = MUL( RT(8), L_2x8);
	RT p1 = calc_brackets(W1_neg, L_2x8, RT(1), D2_x2);
	RT q1 = MUL( L_2x8, D2_x2 );
	RT p2 = calc_brackets(W1_neg, L_2x8, RT(-1), D2_x2);
	RT q2 = MUL( L_2x8, D2_x2 );
	assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	return;
	// if (f_WD == RT(0))
  } // if (R > RT(0))

  // if (R < RT(0))
  RT p1 = b;
  RT q1 = a;
  if (f_WD > RT(0)) {
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p3 = IB + ID;
	RT q3 = IA;
	RT p2 = W1_neg;
	RT q2 = D2_x2;
	assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	return;
  } // if (f_WD > RT(0))
  if (f_WD < RT(0)) {
	RT p3 = W1_neg;
	RT q3 = D2_x2;
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p2 = IB - ID;
	RT q2 = IA;
	assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	return;
  } //if (f_WD < RT(0))
  // if (f_WD == RT(0))
  if (rootnb_ == 2) {
	redefine(D2_x2, -W1_neg, 1, POS_MIDDLE);
	return;
  }
  // TODO Test for Collins
  RT L = std::max( std::max( std::max(abs(a), abs(b)),
		                     std::max(abs(c), abs(d)) ),
	               abs(e) );
  RT L_2x8 = SQR(L);
  L_2x8 = MUL( RT(8), L_2x8);
  RT p2 = calc_brackets(W1_neg, L_2x8, RT(1), D2_x2);
  RT q2 = MUL( L_2x8, D2_x2 );
  RT p3 = calc_brackets(W1_neg, L_2x8, RT(-1), D2_x2);
  RT q3 = MUL( L_2x8, D2_x2 );
  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
  return;
}




template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_4_ONE_fba_zero(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R
	)
{
  typedef Number_traits<RT>    NTR;

  // b/a is a root of the quartic
  // We decide its index
  // So we can have 2 more isolating points from the separation bound
  // We use Collins conjecture for this
  // -W1_/D2_x2 is for sure an isolating point
  // When we need one more isolating point we compute as
  // a root of the isolator polynomial
  //	print(__FUNCTION__);
  print(__FUNCTION__);


  RT f_WD = calc_F4_WD(a, B, D2, D3, W1);
  RT df_ba = calc_cubic_normal(a, b, c, d, b, a);//C4_ * W1_neg - RT(2)*C3_*D2_;
  print("df_ba", df_ba);
  print("f_WD", f_WD);
  RT L = std::max( std::max( std::max(abs(a), abs(b)),
		                     std::max(abs(c), abs(d)) ),
	               abs(e) );
  // b/a is the second root then df(b/a) > 0
  if (df_ba > RT(0)) {
	if (rootnb_ == 1) {
	  redefine(a, -b, 1, POS_MIDDLE);
	  return;
	}
	if (f_WD > RT(0)) {
	  RT L_2x8 = SQR(L);
	  L_2x8 = MUL( RT(8), L_2x8);
	  RT p1 = calc_brackets(b, L_2x8, RT(1), a);
	  RT q1 = MUL(a, L_2x8);
	  RT p2 = W1_neg;
	  RT q2 = D2_x2;
	  RT IB = RT(-3) * W1;
	  RT IA = RT( 6) * D2;
	  RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	  RT p3 = IB + ID;
	  RT q3 = IA;

	  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	  return;
	}
	// if (f_WD < RT(0))
	// TODO Tet for Collins
	RT L_2x8 = SQR(L);
	L_2x8 = MUL( RT(8), L_2x8);
	RT p1 = calc_brackets(b, L_2x8, RT(1), a);
	RT q1 = MUL(a, L_2x8);
	RT p2 = calc_brackets(b, L_2x8, RT(-1), a);
	RT q2 = MUL(a, L_2x8);
	RT p3 = W1_neg;
	RT q3 = D2_x2;

	assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	return;
  } // if (df_ba > RT(0))

  // b/a is the third root then df(b/a) < 0
  //	if (df_ba < RT(0))
  if (rootnb_ == 2) {
	redefine(a, -b, 1, POS_MIDDLE);
	return;
  }
  if (f_WD > RT(0)) {
	RT IB = RT(-3) * W1;
	RT IA = RT( 6) * D2;
	RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	RT p1 = IB - ID;
	RT q1 = IA;
	RT p2 = W1_neg;
	RT q2 = D2_x2;
	RT L_2x8 = SQR(L);
	L_2x8 = MUL( RT(8), L_2x8);
	RT p3 = calc_brackets(b, L_2x8, RT(-1), a);
	RT q3 = MUL(L_2x8, a);

	assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
	return;
  }
  // if (f_WD < RT(0))
  // TODO Test for Collins
  RT p1 = W1_neg;
  RT q1 = D2_x2;
  RT L_2x8 = SQR(L);
  L_2x8 = MUL( RT(8), L_2x8);
  RT p2 = calc_brackets(b, L_2x8, RT(1), a);
  RT q2 = MUL(L_2x8, a);
  RT p3 = calc_brackets(b, L_2x8, RT(-1), a);
  RT q3 = MUL(L_2x8, a);

  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
  return;
}




template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_DOUBLE_2_ONE(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& T1
	)
{
  typedef Number_traits<RT>    NTR;

  print(__FUNCTION__);

  //	print(__FUNCTION__);
  // We know exactly the double root
  // decide if
  //  2 1 1
  //  1 2 1
  //  1 1 2
  RT p, q;
  // find exactly the double root
  if (D2 == RT(0)) {
	p = W3;
	q = MUL(RT(3), W1);
  } else {
	RT T2 = calc_T2(A, W1, b, B);
	print("T2", T2);
	p = MUL( RT(-1), T2 );
	q = MUL( RT(+3), T1 );
  }
  RT Dr1 = calc_brackets(p, a, q, b);
  RT Dr2 = calc_brackets(p, D2_x2, W1_neg, q);
  RT R = calc_brackets(b, D2_x2, a, W1_neg);
  //	RT f_ba = calc_F4_BA(a, A, D2_);
  //case 2 1 1
  if ( (Dr1 < RT(0)) && (Dr2 < RT(0)) ) {
	return adj_Dr1_neg_Dr2_neg(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R,
		p, q);
  }
  // case 1 1 2
  if ((Dr1 > RT(0)) && (Dr2 > RT(0))) {
	return adj_Dr1_pos_Dr2_pos(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R,
		p, q);
  }
  // case 1 2 1
  if ( ((Dr1 < RT(0)) && (Dr2 > RT(0)))  ||
	  ((Dr1 > RT(0)) && (Dr2 < RT(0))) ) {
	return adj_Dr1_diff_Dr2(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R,
		p, q);
  }
  // case 1 2 1
  if ((Dr1 == RT(0)) && (Dr2 == RT(0))) {
	return adj_Dr1_zero_Dr2_zero(
		a, b, c, d, e,
		D2, D3, W1, W2, W3,
		D2_x2, W1_neg,
		A, B, R,
		p, q);
  }
}



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_Dr1_neg_Dr2_neg(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R,
	const RT& p,     const RT& q
	)
{
  typedef Number_traits<RT>    NTR;

  // case 2 1 1
  print(__FUNCTION__);
  mu_ = ROOT_DOUBLE_ONE_ONE;
  if ((rootnb_ == 0) || (rootnb_ == 1)){
	redefine(q, -p, 2, POS_FIRST);
	return;
  }
  RT p1 = p;
  RT q1 = q;
  RT p2, q2;
  RT f_ba = calc_F4_BA(a, A, D2);
  if (f_ba > RT(0)) {
	p2 = b;
	q2 = a;
  } else {
	p2 = W1_neg;
	q2 = D2_x2;
  }

  // TODO: Make them inline functions
  RT D5 = SQR(c) - a * e;
  RT p3 =   A * ( RT(11)* b * D2 + a * W1 )
	- RT(3) * b * SQR(W1)
	+ RT(2) * b * D2 * ( RT(5) * D5 - RT(14) * D3 );
  RT q3 =   SQR(a) * B
	- RT(8) * c * SQR(D2)
	- RT(2) * W1 * ( RT(4) * b * D2 + a * W1 );
  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
//  if (r.rootnb_ == 1) r.pos_ = POS_FIRST;
  return;

}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_Dr1_pos_Dr2_pos(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R,
	const RT& p,     const RT& q
	)
{
  typedef Number_traits<RT>    NTR;

  // case 1 1 2
  print(__FUNCTION__);
  mu_ = ROOT_ONE_ONE_DOUBLE;
  if ((rootnb_ == 2) || (rootnb_ == 3)){
	redefine(q, -p, 2, POS_LAST);
	return;
  }
  // TODO: Make them inline functions
  RT D5 = SQR(c) - a * e;
  RT p1 =   A * ( RT(11)* b * D2 + a * W1 )
	- RT(3) * b * SQR(W1)
	+ RT(2) * b * D2 * ( RT(5) * D5 - RT(14) * D3 );
  RT q1 =   SQR(a) * B
	- RT(8) * c * SQR(D2)
	- RT(2) * W1 * ( RT(4) * b * D2 + a * W1 );

  RT f_ba = calc_F4_BA(a, A, D2);
  RT p2, q2;
  if (f_ba > RT(0)) {
	p2 = b;
	q2 = a;
  } else {
	p2 = W1_neg;
	q2 = D2_x2;
  }
  RT p3 = p;
  RT q3 = q;
  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
//  if (r.rootnb_ == 2) r.pos_ = POS_LAST;
  return;
}



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_Dr1_diff_Dr2(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R,
	const RT& p,     const RT& q
	)
{
  typedef Number_traits<RT>    NTR;

  // case 1 2 1
  print(__FUNCTION__);
  mu_ = ROOT_ONE_DOUBLE_ONE;
  if ((rootnb_ == 1) || (rootnb_ == 2)){
	redefine(q, -p, 2, POS_MIDDLE);
	return;
  }
  RT p2 = p;
  RT q2 = q;
  RT p1, q1, p3, q3;
  if (R > RT(0)) {
	p1 = W1_neg;
	q1 = D2_x2;
	p3 = b;
	q3 = a;
  }
  if (R < RT(0)) {
	p1 = b;
	q1 = a;
	p2 = p;
	q2 = q;
	p3 = W1_neg;
	q3 = D2_x2;
  }
  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
  return;
}


template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adj_Dr1_zero_Dr2_zero(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B,     const RT& R,
	const RT& p,     const RT& q
	)
{
  typedef Number_traits<RT>    NTR;

  // case 1 2 1
  //cout << "The root in the middle" << endl;
  // ALSO We can treat the other roots as
  // roots of  a^2 X^2 - 2 a b X + 6 a c - 5 b^2

  print(__FUNCTION__);
  mu_ = ROOT_ONE_DOUBLE_ONE;
  if ((rootnb_ == 1) || (rootnb_ == 2)){
	redefine(q, -p, 2, POS_MIDDLE);
	return;
  }
  RT p2 = p;
  RT q2 = q;
  RT L = std::max( std::max( std::max(abs(a), abs(b)),
		                     std::max(abs(c), abs(d)) ),
	               abs(e) );
  RT L_2x8 = SQR(L);
  L_2x8 = MUL( RT(8), L_2x8);
  RT p1 = calc_brackets(p, L_2x8, RT(1), q);
  RT q1 = MUL( q, L_2x8 );
  RT p3 = calc_brackets(p, L_2x8, RT(-1), q);
  RT q3 = MUL( q, L_2x8 );
  assign_isolating_intervals(p1, q1, p2, q2, p3, q3);
  return;
}



template < typename RT,
           typename POLY >   inline
void
root_of<RT, POLY>::adjust_poly_2_ONE_2_COMPLEX(
	const RT& a,     const RT& b,     const RT& c,  const RT& d,  const RT& e,
	const RT& D2,    const RT& D3,    const RT& W1, const RT& W2, const RT& W3,
	const RT& D2_x2, const RT& W1_neg,
	const RT& A,     const RT& B
	)
{
  typedef Number_traits<RT>    NTR;
  typedef typename NTR::FT     FT;

  FT left;
  FT right;

  print(__FUNCTION__);
  mu_ = ROOT_ONE_ONE_COMPLEX_COMPLEX;
  assert(rootnb_ <= 1);

  if (rootnb_ == 0)
	left = -Bound(poly(), Cauchy());
  else
	right = Bound(poly(), Cauchy());

  RT f_ba = calc_F4_BA(a, A, D2);
  print("f_ba", f_ba);
  bool first = false;
  bool exact = false;
  RT p, q;
  RT exact_p, exact_q;
  RT f_ed;

  // b/a is an isolating point
  if (C_[0] < RT(0)) {
	p = RT(0);
	q = RT(1);
  } else if (f_ba < RT(0)) {
	p = b;
	q = a;
  } else if (f_ba == RT(0)) {
	exact = true;
	exact_p = b;
	exact_q = a;
	if (rootnb_ == 1) {
	  redefine(exact_q, -exact_p, 1, POS_LAST);
	  return;
	}
	RT L = std::max( std::max( std::max(abs(a), abs(b)),
		                       std::max(abs(c), abs(d)) ),
		             abs(e) );
	RT L_2x8 = SQR(L);
	L_2x8 = MUL( RT(8), L_2x8);
	p = calc_brackets(b, L_2x8, RT(1), a);
	q = MUL( a, L_2x8 );
	int sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
	if (sgn_on_isol == POSITIVE) {
	  p = calc_brackets(b, L_2x8, RT(-1), a);
	  if (rootnb_ == 0) {
		redefine(exact_q, -exact_p, 1, POS_FIRST);
		return;
	  }
	  first = true;
	}
  } else { //1
	f_ed = calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], e, d);
	print("f_ed", f_ed);
	if (f_ed < RT(0)) {
	  p = e;
	  q = d;
	} else if (f_ed == RT(0)) {
	  exact = true;
	  exact_p = e;
	  exact_q = d;
	  if (rootnb_ == 0) {
		redefine(exact_q, -exact_p, 1, POS_LAST);
		return;
	  }
	  RT L = std::max( std::max( std::max(abs(a), abs(b)),
			                     std::max(abs(c), abs(d)) ),
		               abs(e) );
	  RT L_2x8 = SQR(L);
	  L_2x8 = MUL( RT(8), L_2x8);
	  p = calc_brackets(b, L_2x8, RT(1), a);
	  q = MUL( a, L_2x8 );
	  int sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
	  if (sgn_on_isol == POSITIVE) {
		p = calc_brackets(b, L_2x8, RT(-1), a);
		if (rootnb_ == 0) {
		  redefine(exact_q, -exact_p, 1, POS_FIRST);
		  return;
		}
		first = true;
	  }
	} else if (D2 == RT(0)){
	  p = W3;
	  q = RT(3)*W1;
	} else if (W3 == RT(0)){
	  RT D4 = SQR(d) - c * e;
	  p = -D4;
	  q = W2;
	} else {
	  RT IB = RT(-3) * W1;
	  RT IA = RT( 6) * D2;
	  RT ID = isqrt( RT(SQR(IB) + RT(2) * IA * W3) );
	  p = IB - ID;
	  q = IA;
	  //	  cout << "1 = " << QQ(p, q).get_d() << endl;
	  int sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
	  if (sgn_on_isol == POSITIVE) {
		p = IB + ID;
		//		cout << "2 = " << QQ(p, q).get_d() << endl;
		sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
		if (sgn_on_isol == POSITIVE) {
		  p = IB + ID - RT(1);
		  //		  cout << "3 = " << QQ(p, q).get_d() << endl;
		  sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
		  if (sgn_on_isol == POSITIVE) {
			p = IB - ID - RT(1);
			//			cout << "4 = " << QQ(p, q).get_d() << endl;
			sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
			if (sgn_on_isol == POSITIVE) {
			  RT D4 = d * d - c *e;
			  IA = RT(3)*W2;
			  // std::cout << "W3 = " << W3 << std::endl;
			  // std::cout << "ID = " <<  RT( RT(9)*SQR(W2) + RT(12)*D4*W3 )<< std::endl;
			  ID = isqrt( RT( RT(9)*SQR(W2) + RT(12)*D4*W3 ) );
			  p = IA - ID;
			  q = RT(6)*W3;
			  //			  cout << "5 = " << QQ(p, q).get_d() << endl;
			  sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
			  if (sgn_on_isol == POSITIVE) {
				p = IA + ID;
				//				cout << "6 = " << QQ(p, q).get_d() << endl;
				sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
				if (sgn_on_isol == POSITIVE) {
				  p = p - RT(1);
				  //				  cout << "7 = " << QQ(p, q).get_d() << endl;
				  sgn_on_isol = sign( calc_quartic(C_[4], C_[3], C_[2], C_[1], C_[0], p, q) );
				  if (sgn_on_isol == POSITIVE) {
					p = IA - ID - RT(1);
					//					cout << "8 = " << QQ(p, q).get_d() << endl;
				  }
				}
			  }
			}
		  }
		}
	  }

	}
  } //1
  //  cout << "D2 = " << D2 <<  "  W1 = " << W1 << "  W3 = " << W3 << endl;
  //  cout << p << "/" << q << endl;;
  //  cout << " = " << p/q << endl;
  FT isol = to_FT(p, q);
  //  cout << is0 << " " << is1 << " " << is2 << endl;
  //  cout << is0.get_d() << " " << is1.get_d()  << " " << is2.get_d()  << endl;
  if (rootnb_ == 0) {
	pos_ = POS_FIRST;
	right = isol;
	return;
  }
  pos_ = POS_LAST;
  left = isol;

  set_interval(left, right);
  return;
}

__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_ADJUST_POLY_CC

