// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_32_HH
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_32_HH


#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/comparison/Common.h>


__ALGEBRAIC_BEGIN_NAMESPACE

// Forward declaration
template < typename RT, typename P > struct root_of;
//-------------------------------------------------------------------------------


template<class RT>
inline
RT calc_32_S21(const RT& A2, const RT& G, const RT& M1,
		       const RT& B2, const RT& J)
{ return A2 * ( G - RT(2) * M1 ) - RT(4) * B2 * J; }


template<class RT>
inline
RT calc_32_S20(const RT& A2, const RT& M2,
		       const RT& C2, const RT& J)
{ return A2 * M2 + RT(2) * C2 * J; }



template<class RT>
inline
RT calc_G1(const RT& t1, const RT& t2)
{ return t1 - RT(3) * t2; }


template<class RT>
inline
RT calc_G2(const RT& t1, const RT& t2)
{ return t1 + RT(3) * t2; }



template<class RT>
inline
RT calc_J2(const RT& B2, const RT& C1, const RT& B1, const RT& C2)
{ return RT(2) * B2 * C1 + B1 * C2; }



template<class RT>
inline
RT calc_32_L10(const RT& B1, const RT& C2, const RT& C1,
		       const RT& B2, const RT& D1, const RT& A2)
{ return B1 * C2 + D1 * A2 - RT(2) * C1 * B2; }



template<class RT>
inline
RT calc_32_T1(const RT& A1, const RT& B1, const RT& C1,
		      const RT& A2, const RT& B2)
{ return RT(7) * A2 * ( A2 * C1 - B1 * B2) + B2 * ( A2 * B1 + RT(4) * A1 * B2 ); }



template<class RT>
inline
RT calc_32_S3(const RT& S21, const RT& J10, const RT& S20, const RT& J2,
		      const RT& c2, const RT& G1, const RT& G2, const RT& d1,
		      const RT& a2, const RT& L10, const RT& b2, const RT& T1)
{ return RT(-6)*S21*J10 + RT(3)*S20*J2 - c2*G1*G2 + d1* ( a2*a2 *(RT(4)* J2 - L10) - RT(4)*b2*T1); }


//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
eval_full_Sturm_32_pq(
		const RO_t& a,
		const RO_t& b,
		const typename RO_t::FT& isol_1,
		const typename RO_t::FT& isol_2,
		int& where
		)
{
	typedef typename RO_t::NTR    NTR;
	typedef typename RO_t::RT     RT;
	typedef typename RO_t::FT     FT;


	RT a1 = RT( 3) * a[3];
	RT b1 = RT(-1) * a[2];
	RT c1 = RT( 1) * a[1];
	RT d1 = RT(-3) * a[0];
	RT a2 = RT( 2) * b[2];
	RT b2 = RT(-1) * b[1];
	RT c2 = RT( 2) * b[0];
//	RT a1 = a.N3(); RT b1 = a.N2();  RT c1 = a.N1(); RT d1 = a.N0();
//	RT a2 = b.N2(); RT b2 = b.N1();  RT c2 = b.N0();

	RT p1 = numerator(isol_1);
	RT q1 = denominator(isol_1);
	RT p2 = numerator(isol_2);
	RT q2 = denominator(isol_2);

	RT J = calc_brackets(a1, b2, a2, b1);
	RT a1xc2 = MUL(a1, c2);
	RT a2xc1 = MUL(a2, c1);
	RT G = SUBS(a1xc2, a2xc1);
	RT M1 = calc_brackets(a2, c1, b1, b2);
	RT S21 = calc_32_S21(a2, G, M1, b2, J);
	if (S21 == RT(0)) {
		RT M2 = calc_brackets(a2, d1, b1, c2);
		RT S20 = calc_32_S20(a2, M2, c2, J);
		if (S20 > RT(0)) where ^= 16;
		if (S20 == RT(0)) return EQUAL;
		return Static_Result::Result_3.on_S2(where);
	}


	RT M2 = calc_brackets(a2, d1, b1, c2);
	RT S20 = calc_32_S20(a2, M2, c2, J);
	RT S2_p2q2 = calc_one(S21, S20, p2, q2);
	if (S2_p2q2 >= RT(0)) where ^= 16;
	RT S2_p1q1 = calc_one(S21, S20, p1, q1);
	if (S2_p1q1 >= RT(0)) where ^= 32;
	int result = Static_Result::Result_4.on_S2_p1q1(where);
	if (result != UNKNOWN) {
		return result;
	}

	RT G1 = calc_G1(a1xc2, a2xc1);
	RT G2 = calc_G2(a1xc2, a2xc1);
	RT J10 = calc_brackets(b2, d1, c2, c1);
	RT J2 = calc_J2(b2, c1, b1, c2);
	RT L10 = calc_32_L10(b1, c2, c1, b2, d1, a2);
	RT T1 = calc_32_T1(a1, b1, c1, a2, b2);

	RT S3 = calc_32_S3(S21, J10, S20, J2, c2, G1, G2, d1, a2, L10, b2, T1);
	if (S3 == RT(0)) return EQUAL;
	if (S3 > RT(0)) where ^= 64;
	return Static_Result::Result_4.on_S3(where);
}
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_32(const RO_t& a, const RO_t& b)
{
  typedef typename RO_t::NTR    NTR;
  typedef typename NTR::RT      RT;
  typedef typename NTR::FT      FT;
  typedef typename NTR::FIT     FIT;

  int where = 0;
  FIT I;
  int result = decide_common_interval(a, b, I, where);
  if (result != UNKNOWN)
	return result;

  return eval_full_Sturm_32_pq(a, b, I.lower(), I.upper(), where) * b.sign_up();
}
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
compare_23(const RO_t& a, const RO_t& b)
{
  return opposite(compare_32(b, a));
}
//-------------------------------------------------------------------------------


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_32_HH

