// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>

// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARION_COMPARE_24_H
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARION_COMPARE_24_H




#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/comparison/Common.h>


__ALGEBRAIC_BEGIN_NAMESPACE

// Forward declaration
template < typename RT, typename P > struct root_of;
//-------------------------------------------------------------------------------

template < typename RO_t >  inline
int
eval_full_Sturm_42_pq(
		const RO_t& a,
		const RO_t& b,
		const typename RO_t::FT& isol_1,
		const typename RO_t::FT& isol_2,
		int& where
		)
{
	typedef typename RO_t::NTR    NTR;
	typedef typename RO_t::RT     RT;
	typedef typename RO_t::FT     FT;

	int result;
	RT a1 = RT(12) * a[4];
	RT b1 = RT(-3) * a[3];
	RT c1 = RT( 2) * a[2];
	RT d1 = RT(-3) * a[1];
	RT e1 = RT(12) * a[0];
	RT a2 = RT( 2) * b[2];
	RT b2 = RT(-1) * b[1];
	RT c2 = RT( 2) * b[0];
//	RT a2 = b.N2(); RT b2 = b.N1(); RT c2 = b.N0();
//	RT a1 = a.N4(); RT b1 = a.N3(); RT c1 = a.N2(); RT d1 = a.N1(); RT e1 = a.N0();

	RT p1 = numerator(isol_1);
	RT q1 = denominator(isol_1);
	RT p2 = numerator(isol_2);
	RT q2 = denominator(isol_2);

	RT J = calc_brackets(a1, b2, a2, b1);
	RT D22 = b.discriminant();
	RT N4 = calc_brackets(b1, b2, a2, c1);
	RT N1 = calc_brackets(a2, d1, b2, c1);
	RT S21 = calc_24_S21(a2, b2, D22, J, N1, N4);
	RT N2 = calc_brackets(c1, c2, a2, e1);

	RT S2_inf = S21;
	if (S2_inf == RT(0)) {
		RT S20 = calc_24_S20(a1, a2, b2, c2, D22, J, N2, N4);
		if (S20 == RT(0)) return EQUAL;
		if (S20 > RT(0)) where ^= 16;
		return Static_Result::Result_3.on_S2(where);
	}
	RT S20 = calc_24_S20(a1, a2, b2, c2, D22, J, N2, N4);
	RT S2_p2q2 = calc_one(S21, S20, p2, q2);
	if (S2_p2q2 >= RT(0)) where ^= 16;
	RT S2_p1q1 = calc_one(S21, S20, p1, q1);
	if (S2_p1q1 >= RT(0)) where ^= 32;

	result = Static_Result::Result_4.on_S2_p1q1(where);
	if (result != UNKNOWN) {
		return result;
	}
	RT Jp = calc_brackets(b1, c2, b2, c1);
	RT G = calc_brackets(a1, c2, a2, c1);
	RT N3 = calc_brackets(a2, e1, b2, d1);
	RT phi = calc_24_phi(a2, b2, c2, G, Jp, N2, N3);

	RT D12 = calc_D2(a1, b1, c1); //a.D2();
	RT D13 = calc_D2(b1, c1, d1); //a.D3();
	RT D14 = calc_D2(c1, d1, e1);
	RT D15 = calc_D2(a1, c1, e1);
	RT W1 = calc_brackets(a1, d1, b1, c1); //a.W1();
	RT W2 = calc_brackets(b1, e1, c1, d1); //a.W2();

	RT Sigma_p = D22;
	RT Sigma = calc_24_Sigma(a2, b2, c2, D12, D13, D14, D15, W1, W2);

	RT S3 = calc_24_S3(Sigma, Sigma_p, phi);
	if (S3 > RT(0)) where ^= 64;
	if (S3 == RT(0)) return EQUAL;
	return Static_Result::Result_4.on_S3(where);
}
//-------------------------------------------------------------------------------

template < typename RO_t >
int
compare_42(const RO_t& a, const RO_t& b)
{
	typedef typename RO_t::NTR    NTR;
	typedef typename RO_t::RT     RT;
	typedef typename RO_t::FT     FT;
	typedef typename RO_t::FIT    FIT;

	int where = 0;
	FIT I;
	int result = decide_common_interval(a, b, I, where);
	if (result != UNKNOWN)
		return result;

	return  eval_full_Sturm_42_pq(a, b, I.lower(), I.upper(), where) * b.sign_up();
}
//-------------------------------------------------------------------------------

template < typename RO_t >
int
compare_24(const RO_t& a, const RO_t& b)
{
	return opposite(compare_42(b, a));
}
//-------------------------------------------------------------------------------



/// Help functions

template < typename RT >
inline
RT calc_24_S21(
		const RT& a2, const RT& b2,
		const RT& D22, const RT& J, const RT& N1, const RT& N4
		)
{
	return RT(4)*( a2* ( RT(2) * b2 * N4 + a2 * N1 ) - J * ( SQR(b2) + D22 ) );
}


template < typename RT >
inline
RT calc_24_S20(
		const RT& a1, const RT& a2, const RT& b2, const RT& c2,
		const RT& D22, const RT& J, const RT& N2, const RT& N4
		)
{
	return  c2*( a1 * D22 + RT(3) * b2 * J ) + a2 * ( a2 * N2 - RT(5) * c2 * N4 );
}



template < typename RT >
inline
RT calc_24_phi(
		const RT& a2, const RT& b2, const RT& c2,
		const RT& G,  const RT& Jp, const RT& N2, const RT& N3
		)
{
	return  RT(-4) * b2 * Jp  +  c2 * G  +  a2 * ( RT(3) * N2 + RT(4) * N3 );
}


template < typename RT >
inline
RT calc_24_Sigma(
		const RT& a2,  const RT& b2, const RT& c2,
		const RT& D12, const RT& D13, const RT& D14, const RT& D15,
		const RT& W1,  const RT& W2
		)
{
	return  a2 * ( a2 * D14 + RT(2) * b2 * W2  )
		  + b2 * ( b2 * D15 + RT(2) * c2 * W1  )
		  + c2 * ( c2 * D12 + RT(2) * a2 * D13 );
}



template < typename RT >
inline
RT calc_24_S3( const RT& Sigma, const RT& Sigma_p, const RT& phi )
{
	return  RT(16) * Sigma * Sigma_p - SQR(phi);
}


__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARION_COMPARE_24_H

