// Copyright (c) 2003  INRIA Sophia-Antipolis (France) and
//                     Department of Informatics and Telecommunications
//                     University of Athens (Greece).
// All rights reserved.
//
// Authors : Elias P. TSIGARIDAS <et@di.uoa.gr>
// Partially supported by INRIA's project "CALAMATA", a
// bilateral collaboration with National Kapodistrian University of
// Athens.
// Partially supported by the IST Programme of the EU as a Shared-cost
// RTD (FET Open) Project under Contract No  IST-2000-26473
// (ECG - Effective Computational Geometry for Curves and Surfaces)


#ifndef SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_44_HH
#define SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_44_HH


#include <synaps/usolve/algebraic/Algebraic_config.h>
#include <synaps/usolve/algebraic/comparison/Common.h>


__ALGEBRAIC_BEGIN_NAMESPACE

// Forward declaration
template < typename RT, typename P > struct root_of;
//-------------------------------------------------------------------------------

template< typename RT >  inline
int
eval_degJ_Sturm_44_pq(
		      const RT& a1, const RT& b1, const RT& c1, const RT& d1, const RT& e1,
		      const RT& a2, const RT& b2, const RT& c2, const RT& d2, const RT& e2,
		      const RT& p1, const RT& q1, const RT& q1_to_2,
		      const RT& p2, const RT& q2, const RT& q2_to_2,
		      int& where
		      )
{
  int result;
  RT G = calc_brackets(a1, c2, a2, c1);
  RT Gx6 = MUL(RT(6), G);
  RT S2_inf = Gx6;
  //	RT S21 = S2_inf;
  if (S2_inf == (0)) {
    return eval_degG_Sturm_44_pq(
				 a1, b1, c1, d1, e1,
				 a2, b2, c2, d2, e2,
				 p1,  q1, p2, q2,
				 where
				 );
  }
  RT M = calc_brackets(a1, d2, a2, d1);
  RT Mx4 = MUL(RT(4), M);
  RT M3 = calc_brackets(a1, e2, a2, e1);
  RT S2_p2q2 = calc_44_S2_pq(RT(0), Gx6, Mx4, M3, p2, RT(1), q2, q2_to_2);
  if (S2_p2q2 >= RT(0)) where ^= 16;

  RT S2_p1q1 = calc_44_S2_pq(RT(0), Gx6, Mx4, M3, p1, RT(1), q1, q1_to_2);
  if (S2_p1q1 >= 0) where ^= 32;

  RT J1 = calc_brackets(b1, c2, b2, c1);
  RT T32 = calc_44_S3_inf(RT(0), M, J1, G);
  RT M1 = calc_brackets(b1, d2, b2, d1);
  RT M1x16 = MUL(RT(16), M1);
  RT T = ADD(M1x16, M3);
  RT T31 = calc_44_S31(G, M, RT(0), T);
  RT M4 = calc_brackets(b1, e2, b2, e1);
  RT T30 = calc_44_S30(RT(0), M4, G, M3);
  RT M2 = calc_brackets(c1, d2, c2, d1);
  RT M5 = calc_brackets(d1, e2, d2, e1);
  RT Mx8 = MUL(Mx4, RT(2));
  RT Mx16 = MUL(Mx8, RT(2));
  RT S3_inf =  calc_44_S4_inf(T32, T31, T30, RT(0), M,
			      Mx8, Mx16, M1, M1x16, M2, M3, M4, M5);

  RT SS31 = S3_inf;
  if (G == RT(0))
    S3_inf = MUL(RT(1), S3_inf);
  else
    S3_inf = RT(sign(G))*MUL(RT(-1), S3_inf);
  RT S31 = S3_inf;
  //	Degeneracy
  if (S3_inf == RT(0)) {
    RT M6 = calc_brackets(c1, e2, c2, e1);
    RT S30 = calc_44_S40(T32, T30, T, M, Mx8, M3, RT(0), M6);
    if (G == RT(0))
      S30 = MUL(RT(1), S30);
    else
      S30 = RT(sign(G))*MUL(RT(-1), S30);
    S3_inf = S30;

    if (S3_inf == RT(0))
      return Result_deg_3::on_end(where);
    if (S3_inf > RT(0)) { where ^= 64; where ^= 128;}
    return Result_deg_4::on_end(where);

  }
  RT M6 = calc_brackets(c1, e2, c2, e1);
  RT SS30 = calc_44_S40(T32, T30, T, M, Mx8, M3, RT(0), M6);
  RT S30;
  if (G == RT(0))
    S30 = MUL(RT(-1), SS30);
  else
    S30 = RT(sign(G))*MUL(RT(-1), SS30);

  RT S3_p2q2 = calc_44_S4_pq(S31, S30, p2, q2);
  if (S3_p2q2 >= RT(0)) where ^= 64;
  result = Static_Result::Result_5.on_S3_p2q2(where);
  if (result != UNKNOWN) {
    return result;
  }


  RT S3_p1q1 = calc_44_S4_pq(S31, S30, p1, q1);
  if (S3_p1q1 >= 0) where ^= 128;
  result = Static_Result::Result_5.on_S3_p1q1(where);
  if (result != UNKNOWN) {
    return result;
  }

  RT S4 = calc_44_S5(T32, T31, T30, SS31, SS30, RT(0), G, M, Mx16, M3, M4, M5, M6);
  S4 = RT(sign(G))*MUL(RT(-1), S4);
  if (S4 == RT(0)) {
    return EQUAL;
  }
  if (S4 > RT(0)) where ^= 256;
  result = Static_Result::Result_5.on_S4(where);
  if (result != UNKNOWN) {
    return result;
  }
  return EQUAL;
}
//-------------------------------------------------------------------------------


template< typename RO_t >  inline
int
eval_full_Sturm_44_pq(
		      const RO_t& a,
		      const RO_t& b,
		      const typename RO_t::FT& isol_1,
		      const typename RO_t::FT& isol_2,
		      int& where
		      )
{
  typedef typename RO_t::NTR   NTR;
  typedef typename RO_t::RT   RT;
  typedef typename RO_t::FT   FT;

  int result;

  RT a1 = RT(12) * a[4];
  RT b1 = RT(-3) * a[3];
  RT c1 = RT( 2) * a[2];
  RT d1 = RT(-3) * a[1];
  RT e1 = RT(12) * a[0];
  RT a2 = RT(12) * b[4];
  RT b2 = RT(-3) * b[3];
  RT c2 = RT( 2) * b[2];
  RT d2 = RT(-3) * b[1];
  RT e2 = RT(12) * b[0];
  //	RT a1 = a.N4(); RT b1 = a.N3();  RT c1 = a.N2(); RT d1 = a.N1();  RT e1 = a.N0();
  //	RT a2 = b.N4(); RT b2 = b.N3();  RT c2 = b.N2(); RT d2 = b.N1();  RT e2 = b.N0();

  RT p1 = numerator(isol_1);
  RT q1 = denominator(isol_1);
  RT p2 = numerator(isol_2);
  RT q2 = denominator(isol_2);
  RT q1_to_2 = SQR(q1);
  RT q1_to_3 = MUL(q1, q1_to_2);
  RT q2_to_2 = SQR(q2);
  RT q2_to_3 = MUL(q2, q2_to_2);

  RT J = calc_brackets(a1, b2, a2, b1);
  if (J == RT(0))
    return eval_degJ_Sturm_44_pq<RT>(
				     a1, b1, c1, d1, e1,
				     a2, b2, c2, d2, e2,
				     p1,  q1, q1_to_2,
				     p2,  q2, q2_to_2,
				     where
				     );
  RT Jx4 = MUL(RT(4), J);
  RT S2_inf = calc_44_S2_inf(Jx4);

  RT G = calc_brackets(a1, c2, a2, c1);
  RT Gx6 = MUL(RT(6), G);
  RT M = calc_brackets(a1, d2, a2, d1);
  RT Mx4 = MUL(RT(4), M);
  RT M3 = calc_brackets(a1, e2, a2, e1);

  RT S2_p2q2 = calc_44_S2_pq(S2_inf, Gx6, Mx4, M3, p2, q2, q2_to_2, q2_to_3);
  if (S2_p2q2 >= RT(0)) where ^= 16;


  RT S2_p1q1 = calc_44_S2_pq(S2_inf, Gx6, Mx4, M3, p1, q1, q1_to_2, q1_to_3);
  if (S2_p1q1 >= 0) where ^= 32;

  RT J1 = calc_brackets(b1, c2, b2, c1);
  RT S3_inf = calc_44_S3_inf(J, M, J1, G);
  RT S32 = S3_inf;
  //	Degeneracy
  if (S3_inf == RT(0)) {
    RT M1 = calc_brackets(b1, d2, b2, d1);
    RT M1x16 = MUL(RT(16), M1);
    RT T = ADD(M1x16, M3);
    RT S31 = calc_44_S31(G, M, J, T);
    S3_inf = S31;
    if (S3_inf == RT(0)) {
      RT M4 = calc_brackets(b1, e2, b2, e1);
      RT S30 = calc_44_S30(J, M4, G, M3);
      S3_inf = S30;
      if (S3_inf == RT(0))
	return Result_deg_3::on_end(where);
      if (S3_inf > RT(0)) { where ^= 64; where ^= 128;}
      return Result_deg_5::on_end(where);
    }
  }
  RT M1 = calc_brackets(b1, d2, b2, d1);
  RT M1x16 = MUL(RT(16), M1);
  RT T = ADD(M1x16, M3);
  RT S31 = calc_44_S31(G, M, J, T);
  RT M4 = calc_brackets(b1, e2, b2, e1);
  RT S30 = calc_44_S30(J, M4, G, M3);
  RT S3_p2q2 = calc_44_S3_pq(S32, S31, S30, p2, q2, q2_to_2);
  if (S3_p2q2 >= RT(0)) where ^= 64;

  RT S3_p1q1 = calc_44_S3_pq(S32, S31, S30, p1, q1, q1_to_2);
  if (S3_p1q1 >= 0) where ^= 128;
  result = Static_Result::Result_6.on_S3_p1q1(where);
  if (result != UNKNOWN) {
    return result;
  }

  RT M2 = calc_brackets(c1, d2, c2, d1);
  RT M5 = calc_brackets(d1, e2, d2, e1);
  RT Mx8 = MUL(Mx4, RT(2));
  RT Mx16 = MUL(Mx8, RT(2));
  RT S4_inf =  calc_44_S4_inf(S32, S31, S30, J, M,
			      Mx8, Mx16, M1, M1x16, M2, M3, M4, M5);

  RT S41 = S4_inf;
  //	Degeneracy
  if (S4_inf == RT(0)) {
    RT M6 = calc_brackets(c1, e2, c2, e1);
    RT S40 = calc_44_S40(S32, S30, T, M, Mx8, M3, J, M6);
    S4_inf = S40;
    if (S4_inf == RT(0))
      return Result_deg_4::on_end(where);
    if (S4_inf > RT(0)) { where ^= 256; where ^= 512;}
    return Result_deg_5::on_end(where);
  }

  RT M6 = calc_brackets(c1, e2, c2, e1);
  RT S40 = calc_44_S40(S32, S30, T, M, Mx8, M3, J, M6);
  RT S4_p2q2 = calc_44_S4_pq(S41, S40, p2, q2);
  if (S4_p2q2 >= RT(0)) where ^= 256;
  result = Static_Result::Result_6.on_S4_p2q2(where);
  if (result != UNKNOWN) {
    return result;
  }

  RT S4_p1q1 = calc_44_S4_pq(S41, S40, p1, q1);
  if (S4_p1q1 >= RT(0)) where ^= 512;
  result = Static_Result::Result_6.on_S4_p1q1(where);
  if (result != UNKNOWN) {
    return result;
  }

  RT S5 = calc_44_S5(S32, S31, S30, S41, S40, J, G, M, Mx16, M3, M4, M5, M6);
  if (S5 == RT(0)) return EQUAL;
  if (S5 > RT(0)) where ^= 1024;
  result = Static_Result::Result_6.on_S5(where);
  if (result != UNKNOWN) {
    return result;
  }
  return EQUAL;
}
//-------------------------------------------------------------------------------

template< typename RT >  inline
int
eval_degG_Sturm_44_pq(
		      const RT& a1, const RT& b1, const RT& c1, const RT& d1, const RT& e1,
		      const RT& a2, const RT& b2, const RT& c2, const RT& d2, const RT& e2,
		      const RT& p1, const RT& q1,
		      const RT& p2, const RT& q2,
		      int& where
		      )
{
  int result;

  RT M = calc_brackets(a1, d2, a2, d1);
  RT S2_inf = MUL(RT(-4), M);
  if (M == RT(0)) {
    RT M3 = calc_brackets(a1, e2, a2, e1);
    RT S2_inf = M3;
    if (S2_inf == RT(0)) return EQUAL;
    if (S2_inf > RT(0)) { where ^= 16; where ^= 32; }
    return Result_deg_3::on_end(where);
  }

  RT M3 = calc_brackets(a1, e2, a2, e1);
  RT S2_p2q2 = calc_44_S4_pq(S2_inf, M3, p2, q2);
  if (S2_p2q2 > RT(0)) where ^= 16;

  RT S2_p1q1 = calc_44_S4_pq(S2_inf, M3, p1, q1);
  if (S2_p1q1 >= RT(0)) where ^= 32;
  result = Static_Result::Result_4.on_S2_p1q1(where);
  if (result != UNKNOWN) {
    return result;
  }

  //	RT M1 = calc_brackets(b1, d2, b2, d1);
  RT M4 = calc_brackets(b1, e2, b2, e1);
  RT M5 = calc_brackets(d1, e2, d2, e1);
  RT M6 = calc_brackets(c1, e2, c2, e1);
  RT S3 = calc_44_deg_G_S3(M, M3, M4, M5, M6);
  S3 = MUL(RT(-1),S3);
  if (S3 == RT(0)) {
    return EQUAL;
  }
  if (S3 > RT(0)) where ^= 64;
  if (S3 == RT(0)) return EQUAL;
  result = Static_Result::Result_4.on_S3(where);
  if (result != UNKNOWN) {
    return result;
  }
  return EQUAL;
}
//-------------------------------------------------------------------------------

template < class RO_t >
int
compare_44(const RO_t& a, const RO_t& b)
{
  typedef typename RO_t::RT    RT;
  typedef typename RO_t::FIT   FIT;

  int where = 0;
  FIT I;
  int result = decide_common_interval(a, b, I, where);
  if (result != UNKNOWN)
    return result;

  return eval_full_Sturm_44_pq(a, b, I.lower(), I.upper(), where) * b.sign_up();
}
//-------------------------------------------------------------------------------


template<class RT>
inline
RT calc_44_S2_inf(const RT& J)
{ return RT(-1) * J; }


template<class RT>
inline
RT calc_44_S2_pq(const RT& A, const RT& B, const RT& C, const RT& D,
		 const RT& p, const RT& q, const RT& q_to_2, const RT q_to_3 )
{
  return ((A*p + B*q)*p - C*q_to_2)*p + D*q_to_3;
}


template<class RT>
inline
RT calc_44_S3_inf(const RT& J, const RT& M, const RT& J1, const RT G)
{
  RT G_to_2 = SQR(G);
  return RT(8)*J*(M + RT(6)*J1) - RT(18)*G_to_2;
}



template<class RT>
inline
RT calc_44_S31(const RT& G, const RT& M, const RT& J, const RT& T)
{
  return RT(12)*G*M - RT(2)*J*T;
}

template<class RT>
inline
RT calc_44_S30(const RT& J, const RT& M4, const RT& G, const RT& M3)
{
  return RT(8)*J*M4 - RT(3)*G*M3;
}

template<class RT>
inline
RT calc_44_S3_pq(const RT& S32, const RT& S31, const RT& S30,
		 const RT& p,   const RT& q,   const RT& q_to_2)
{
  return (S32*p + S31*q)*p + S30*q_to_2;
}



template<class RT>
inline
RT calc_44_S4_inf( const RT& S32, const RT& S31,   const RT& S30,
		   const RT& J,
		   const RT& M,   const RT& Mx8,   const RT& Mx16,
		   const RT& M1,  const RT& M1x16,
		   const RT& M2,  const RT& M3,    const RT& M4,     const RT& M5)
{
  RT M_to_2 = SQR(M);
  RT J_to_2 = SQR(J);
  return    RT(-4)*S32*(RT(6)*M2 + M4) - M1x16*S31 + Mx8*S30
    + RT(-2)*J*M3*(M1x16 - M3)
    + RT(2)*Mx16*(M_to_2 - RT(6)*J*M2) - RT(64)*J_to_2*M5;
}

template<class RT>
inline
RT calc_44_S40(const RT& S32, const RT& S30,
	       const RT& T,   const RT& M,   const RT& Mx8,
	       const RT& M3,  const RT& J,   const RT& M6)
{
  return RT(6)*M6*S32 - T*S30 - Mx8*(M*M3 - RT(6)*J*M6);
}



template<class RT>
inline
RT calc_44_S4_pq(const RT& S41, const RT& S40, const RT& p, const RT& q)
{
  return S41*p + S40*q;
}

template<class RT>
inline
RT calc_44_S5(
	      const RT&  S32, const RT&  S31, const RT&  S30, const RT&  S41,  const RT&  S40,
	      const RT&  J,   const RT&  G,   const RT&  M,   const RT&  Mx16,
	      const RT&  M3,  const RT&  M4,  const RT&  M5,  const RT&  M6)
{
  RT M3_to_2 = SQR(M3);
  RT K1 = RT(-8) * M5 * ( S41 - M3 * S31 );
  RT K2 = RT(32) * M4 * ( M5 * S32 - M4 * S30 );
  RT K3 =	RT(-12) * M6 * ( S40 - RT(2) * M3 * S30);
  RT K4 =	M3_to_2 * (M3_to_2 - Mx16 * M4 - RT(16) * J * M5 + RT(36) * G * M6);
  RT K = K1 + K2 + K3 + K4;

  return K;
}


template<class RT>
inline
RT calc_44_deg_G_S3(const RT& M, const RT& M3, const RT& M4, const RT& M5, const RT& M6)
{
  RT M3_to_2 = SQR(M3);
  RT M_to_2 = SQR(M);
  return M3_to_2 * ( M3_to_2 - RT(16) * M * M4 ) + M_to_2*(RT(-256) * M * M5 + RT(96) * M3 * M6);
}




__ALGEBRAIC_END_NAMESPACE

#endif // SYNAPS_ARITHM_ALGEBRAIC_COMPARISON_COMPARE_44_HH

