/********************************************************************
 *   This file is part of the source code of the SYNAPS library
 *   Author(s): B. Mourrain, GALAAD, INRIA
 *   History:
 *     The Rockwood method for computing the first positive real root added.
 *   $Id: Std.h,v 1.1 2005/09/28 06:30:41 mourrain Exp $
 *********************************************************************/
/** \file                                                           **/
//--------------------------------------------------------------------
#ifndef synaps_solve_bezier_Rockwood_H
#define synaps_solve_bezier_Rockwood_H
//--------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/arithm/binomial.h>
#include <synaps/base/Seq.h>
#include "synaps/arithm/sign.h"
#include <synaps/upol/bezier/rep1d.h>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//====================================================================
namespace bezier 
{

  //--------------------------------------------------------------------
  /** Compute the first real positive root of a polynomial with coefficients
      @bz@ in the Bernstein basis for the interval $]u,v[$. It returns
      @v@ when no root is found.
  **/
  template<class C,class B>
  C solve_rockwood(const B & bz,const C & u, const C& v, const C & eps)
  {
    unsigned sz =bz.size();
    C* tab= new C[2*sz];
    C* t0=tab, *t1= tab+sz, *t2;
    for(unsigned i=0;i<sz;++i) t1[i]=bz[i];


    C x=u, a=u, b=v, c=u/*, h*/;
    unsigned i=0;
    bool s;

    while ( (t1[0]>eps || t1[0]<-eps) && (b-a >eps) )
    {
      //  for(unsigned k=0;k<sz;++k) std::cout<<t1[k]<<" "; std::cout<<std::endl;
      i=1;
      s = (t1[0]>0);
      while( s == (t1[i]>0) && i <sz) i++;
      if(i==sz) return v;
      x = ((i-1) - t1[i-1]/(t1[i]-t1[i-1]))/(sz-1);
      if(i==(sz-1) && t1[sz-1]<eps && t1[sz-1]>-eps) x=(C(sz)-C(2))/(sz-1);
         //return b;
      BEZIER::subdivide_size(t0,t1,sz,x);
      s = (t0[0]>0); i=1;
      while( s == (t0[i]>0)  && i <sz) ++i;
      if(i<sz)
	{ t2=t1;t1=t0;t0=t2; b-=(1-x)*(b-a); c=b;}
      else
	{ a+=x*(b-a); c=a;}
    }
    delete [] tab;
    return c;
  }

  
  //--------------------------------------------------------------------
  /** Assuming there is one root in $[a,b[$, it computes it within the
      precision e. **/
  template<class C,class T>
  C dichotom(const T & p, const C&  a, const C&  b, const C&  e)
  {
    C u0=0, u1=1, re=e/(b-a);
    int s0=sign(p[0]), s1=sign(p[p.degree()]);
    for(unsigned i=1; i < p.size() && s0==0; i++)  s0=sign(p[i]);
    while((u1-u0)>re)
      {
	int sm= sign(eval(p,C((u0+u1)/2)));
	if(sm != s0)
	  { s1=sm; u1=(u0+u1)/2; }
	else if(sm != s1)
	  { s0=sm; u0=(u0+u1)/2; }
	else if(sm!=0)
	  {
	    std::cerr <<"Error in sign evaluation"<< std::endl;
	    return 100000;//	  exit(1);
	  }
      }
    return (a+(u0+u1)*(b-a)/2);
  }
  //--------------------------------------------------------------------
  /** Compute the first root of a polynomial @v1@ in the interval $[a,b]$.
      Put the result in @x@ and return @true@, if a root is found.  Else if
      returns @false@. Step the subdivision when the size of the interval is
      small than @e@.
  **/
  template<class C, class T>
  bool firstroot(C & x, T& v1, const C& a, const C& b, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var = sign_variation(v1,2);
    if(b-a<e && b-a>-e && var>0)
      {
	x= (a+b)/C(2); return true;
      }
    else if(v1[0]< e && v1[0]>-e)
      {x= a; return true;}
    else if (var==1)
      {
	x = bezier::dichotom(v1,a,b,e); return true;
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	unsigned nv0 = sign_variation(v0,2);
	if(nv0 >0 && bezier::firstroot(x,v0,a,m,e))
	  return true;
	else
	  return bezier::firstroot(x,v1,m,b,e);
      }
    else
      return false;
  }

  //---------------------------------------------------------------------
  /**  Compute the $r^{th}$ first root on the interval $[a,b]$.
       If it founds it, it returns @true@ and put the isolating  interval
       in @x@. If it does not found it, it decrease @r@ by the number of
       roots found and return false.
       Roots are counted with multiplicity.
  **/
  template<class C, class T>
  bool isole_nth_first_root(std::pair<C,C> & x, T& v1,
			    const C& a, const C& b,
			    int & r, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var=Variation2(v1);
    if(v1[0]< e && v1[0]>-e)
      {
	unsigned mu=0;
	while((int)mu<d && v1[mu]< e && v1[mu]>-e) mu++;
	if(r<=(int)mu)
	  { x = std::pair<C,C>(a,b); r=0; return true; }
	else
	  { r-=mu;  return false; }
      }
    else if(var>0 && b-a<e && b-a>-e)
      {
	if ((int)var>=r)
	  {  x = std::pair<C,C>(a,b); r-=var; return true;}
	else
	  {r-=var; return false;}
      }
    else if (var==1)
      {
	if(r==1)
	  {
	    C u= shrinkinf(v1,a,b);
	    x = std::pair<C,C>(a,u);
	    r=0;
	    return true;
	  }
	else
	  { r--; return false; }
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	bool nth_root_found = bezier::isole_nth_first_root(x,v0,a,m,r,e);
	if(nth_root_found)
	  return true;
	else
	  return bezier::isole_nth_first_root(x,v1,m,b,r,e);
      }
    else
      return false;
  }

//---------------------------------------------------------------------
/**  Compute the $r^{th}$ first root on the interval $[a,b]$.
       If it founds it, it returns @true@ and put the isolating  interval
       in @x@. If it does not found it, it decrease @r@ by the number of
       roots found and return false.
       Roots are counted with multiplicity.
**/
  template<class C, class T>
  bool isole_nth_last_root(std::pair<C,C> & x, T& v1,
			    const C& a, const C& b,
			    int & r, const C& e)
  {
//    unsigned nbr=0;
    int d=(v1.size())-1;
    unsigned var=Variation2(v1);
    if(v1[d]< e && v1[d]>-e)
      {
	unsigned mu=d;
	while(mu>=0 && v1[mu]< e && v1[mu]>-e) mu++;
	if(r<=(int)mu)
	  { x = std::pair<C,C>(a,b); r=0; return true; }
	else
	  { r-=mu;  return false; }
      }
    else if(var>0 && b-a<e && b-a>-e)
      {
	if ((int)var>=r)
	  { x = std::pair<C,C>(a,b); r-=var; return true;}
	else
	  { r-=var; return false;}
      }
    else if (var==1)
      {
	if(r==1)
	  {
	    C u= shrinkinf(v1,a,b);
	    x = std::pair<C,C>(a,u);
	    r=0;
	    return true;
	  }
	else
	  { r--; return false; }
      }
    else if (var>1)
      {
	T v0(d+1);
	BEZIER::subdivide(v0,v1);
	C m=(a+b)/2;
	bool nth_root_found = bezier::isole_nth_last_root(x,v1,m,b,r,e);
	if(nth_root_found)
	  return true;
	else
	  return bezier::isole_nth_last_root(x,v0,a,m,r,e);
      }
    else
      return false;
  }
} // namespace  bezier
//====================================================================
template<class O, class X=double>
struct Rockwood 
{
  typedef X precision_t;
  X eps;
  Rockwood(): eps(1.e-6) {}
  Rockwood(const X& e):eps(e) {}
  Rockwood(const Rockwood & mth): eps(mth.eps) {}

};
//----------------------------------------------------------------------
/** Compute the first real positive roots of p, by subdivision of the control
    polygon, within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
C RootOf(const UPOL & p, const Rockwood<C> & mth)
{
  bezier::rep1d<C> bz(p.size()); 
  let::assign(bz,p);
  C x = bezier::solve_rockwood(bz,C(0),C(1),mth.eps);
  if (x==1)
    {
      std::cerr<< "No real positive root"<<std::endl;
    }
  return x/(1-x);
}
//====================================================================
template<class O, class X=double>
struct SlvBzFst 
{
  typedef X precision_t;
  X eps;
  SlvBzFst(): eps(1.e-6) {}
  SlvBzFst(const X& e):eps(e) {}
  SlvBzFst(const SlvBzFst & mth): eps(mth.eps) {}

};
//----------------------------------------------------------------------
/** Compute the first real positive roots of p, by subdivision of the control
    polygon, within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class C, class UPOL>
C RootOf(const UPOL & p, const SlvBzFst<C> & mth)
{
  bezier::rep1d<C> bz(p.size()); 
  let::assign(bz,p);
  C x;
  bool b=bezier::firstroot(x,bz,C(0),C(1),mth.eps);
  if (!b)
    {
      std::cerr<< "No real positive root found"<<std::endl;
    }
  return x/(1-x);
}
//====================================================================
__END_NAMESPACE_SYNAPS
/********************************************************************/
#endif // synaps_solve_bezier_Rockwood_H

