/********************************************************************
 *   This file is part of the source code of the SYNAPS library
 *   Author(s): B. Mourrain, GALAAD, INRIA
 *   History:
 *     The Rockwood method for computing the first positive real root added.
 *********************************************************************/
/** \file                                                           **/
//--------------------------------------------------------------------
#ifndef synaps_usolve_bezier_SlvRkw_h_
#define synaps_usolve_bezier_SlvRkw_h_
//--------------------------------------------------------------------
#include <synaps/init.h>
#include <synaps/upol/bezier/rep1d.h>
#include <synaps/upol/bezier/BEZIER.m>
//--------------------------------------------------------------------
__BEGIN_NAMESPACE_SYNAPS
//--------------------------------------------------------------------
/** Class specifying the parameters of the Rockwood subdivision algorithm.
    The default precision is \c 1e-6. The default type of coefficients \c C
    is \c double.
**/
template<class O=double, class C = double, class X = double>
struct SlvRkw
{
  X eps;
  SlvRkw():eps(1.e-6) {}
  SlvRkw(const X& e):eps(e) {}
  SlvRkw(const SlvRkw & mth): eps(mth.eps) {}
};
//====================================================================
namespace bezier 
{
  using BEZIER::subdivide_size;

  /** Compute the first real positive root of a polynomial with coefficients
      @bz@ in the Bernstein basis for the interval $]u,v[$. It returns
      @v@ when no root is found.
  **/
  template<class B, class C, class X>
  C solve_rockwood(const B & bz,const C & u, const C& v, const X & eps)
  {
    unsigned sz =bz.size();
    C* tab= new C[2*sz];
    C* t0=tab, *t1= tab+sz, *t2;
    for(unsigned i=0;i<sz;++i) t1[i]=bz[i];


    C x=u, a=u, b=v, c=u/*, h*/;
    unsigned i=0;
    bool s;

    while ( (t1[0]>eps || t1[0]<-eps) && (b-a >eps) )
    {
      //  for(unsigned k=0;k<sz;++k) std::cout<<t1[k]<<" "; std::cout<<std::endl;
      i=1;
      s = (t1[0]>0);
      while( s == (t1[i]>0) && i <sz) i++;
      if(i==sz) return v;
      x = ((i-1) - t1[i-1]/(t1[i]-t1[i-1]))/(sz-1);
      if(i==(sz-1) && t1[sz-1]<eps && t1[sz-1]>-eps) x=(C(sz)-C(2))/(sz-1);
         //return b;
      subdivide_size(t0,t1,sz,x);
      s = (t0[0]>0); i=1;
      while( s == (t0[i]>0)  && i <sz) ++i;
      if(i<sz)
	{ t2=t1;t1=t0;t0=t2; b-=(1-x)*(b-a); c=b;}
      else
	{ a+=x*(b-a); c=a;}
    }
    delete [] tab;
    return c;
  }

} // namespace  bezier
//----------------------------------------------------------------------
/** Compute the first real positive roots of p, by subdivision of the control
    polygon, within a precision @eps@ after the change of coordinnates
    $x=\frac{t}{1-t}$.
**/
template<class O, class C, class X, class UPOL>
  O solve(const UPOL & p, const SlvRkw<O,C,X> & mth)
{
  unsigned d1=p.size();
  bezier::rep1d<C> bz(d1); 
  let::assign(bz,p);
  O x = bezier::solve_rockwood(bz,C(0),C(1),mth.eps);
  if (x==1)
    {
      std::cerr<< "No real positive root"<<std::endl;
    }
  assert(x!=1);
  return x/(1-x);
}
/********************************************************************/
__END_NAMESPACE_SYNAPS 
#endif //

