
#ifndef _SYNAPS_SOLVE_STURM_STURM_ABERTH_SOLVER_H_
#define _SYNAPS_SOLVE_STURM_STURM_ABERTH_SOLVER_H_


#include <synaps/init.h>
#include <synaps/usolve.h>
#include <synaps/usolve/Aberth.h>
#include <synaps/usolve/sturm/common.h>
#include <synaps/usolve/sturm/solve_sturm.h>
#include <synaps/usolve/sturm/construct_root_of.h>


__ALGEBRAIC_BEGIN_NAMESPACE




template < typename K,
	   typename OutputIterator > inline
OutputIterator
solve_sturm_aberth( const typename K::Poly& h, OutputIterator sol, K)
{
    typedef typename K::RT           RT;
    typedef typename K::FT           FT;
    typedef typename K::XT           XT;
    typedef typename K::RL           RL;

    typedef typename K::FIT          FIT;
    typedef typename K::XIT          XIT;
    typedef typename K::RIT          RIT;
  
    typedef typename K::Poly         Poly;
    typedef typename K::RO_t         RO_t;
    typedef typename K::solution_type         solution_type;
  
  
    if (UPOLDAR::degree(h) < 5) {
	solve_small_degree(h, 1, sol, K());
	return sol;
    }

    // Check if x = 0, is a root
    int mult = 0;    
    if ( h[0] == RT(0)) {
	mult = 1;
	RO_t r = construct_root_of( RT(0), RT(1), K());
	
	while (h[mult] == RT(0)) {
	    r.set_multiplicity( ++mult); 
	}
	*sol++ = r;
    }
    int dg = UPOLDAR::degree( h) - mult;
    
    Poly f( dg+1, AsSize());
    for (int i = 0; i <= dg; ++i) { f[i]  = h[i+mult]; }


    SturmSeq<RT> sth;
    Seq<Poly> PL = SquareFreeFactorization( f, sth);

    // Compute the approximate solutions
    Seq<RL> apxsol;
    for (int i = 0; i < PL.size(); ++i) {
	if (UPOLDAR::degree( PL[i]) < 1) continue;

	if (UPOLDAR::degree( PL[i]) < 5) {
	    solve_small_degree( PL[i], i+1, sol, K());
	    continue;
	}
	
	Seq<RL> apxsol = solve( PL[i], Aberth<RR>(20, 30));
	std::sort( apxsol.begin(), apxsol.end());
	
	// Compute rationals between the approximate solutions
	Seq<FT> cand;    
	if (apxsol.size() == 0) {
	    cand.resize( 2);
	    FT b = Bound( f, Cauchy());
	    cand[0] = -b;
	    cand[1] = b;
	} else {
	    cand.resize( apxsol.size() + 1);
	    cand[0] =  simplest_rational_in_interval<K>( apxsol[0] - 1, apxsol[0]);
	    cand[cand.size()-1] = simplest_rational_in_interval<K>( 
		apxsol[ apxsol.size()-1], apxsol[apxsol.size() - 1] + 1);
//	    cand[0] = FT(apxsol[0]) - 1;
//	    cand[cand.size()-1] = FT(apxsol[apxsol.size()-1]) + 1;
	    
	    for (int j = 0; j < apxsol.size()-1; ++j) {
		cand[j+1] = simplest_rational_in_interval<K>( apxsol[j], apxsol[j+1]);
		//cand[j+1] = FT( apxsol[j] + apxsol[j+1]) /2;
	    }
	}
	
	// Do we need to recompute the Sturm sequence ?
	if (PL.size() == 1) {
	    ;
	} else {
	    sth = SturmSeq<RT>( PL[i], UPOLDAR::diff( PL[i]), SUBRES());
	}
	std::unique( cand.begin(), cand.end());
	
	Seq<FT> isol;
	isol.push_back( cand[0]);
	
	// check if an endpoint is a root
	for (int j = 1; j < cand.size() - 1; ++j) {
	  Poly x(2, AsSize()); x[1] = RT(1);
	  FT p = cand[j];
	  
	  int sgn = sign( homogeneous_eval_horner<K>( PL[i], p));
	  if (sgn == ZERO) {
	    RO_t r = construct_root_of( RT(-numerator(p)), denominator(p), K());
	    r.set_multiplicity( i);
	    *sol++ = r;
	    
	    Poly f = translate(sth[0], p)/x;
	    FT M = min_bound(f, Cauchy());
	    isol.push_back( p - M);
	    isol.push_back( p + M);
	  } else {
	    isol.push_back( cand[j]);
	    isol.push_back( cand[j]);
	  }
	  
	}
	isol.push_back( cand[ cand.size() - 1]);	
	
	// Isolate the real roots, using Sturm sequeneces
	for (int j = 0; j < isol.size() - 1; j+= 2) {
	  if (isol[j] > isol[j+1]) continue;
	  FIT I( isol[j], isol[j+1]);
	  sturm_rootof( sth, I,  j+1, sol, K());
	}
    }

    
    return sol;
}


__ALGEBRAIC_END_NAMESPACE

#endif // _SYNAPS_SOLVE_STURMSTURM_ABERTH_SOLVER_H_

